# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

from abc import ABC
from abc import abstractmethod
from datetime import datetime
import logging

from iotcs.client import ExternalObject
from iotcs.client.model import DeviceModelAttribute
from iotcs.exception import ArgumentException

# com.oracle.iot.client


class VirtualDeviceAttribute(ABC):
    """
    VirtualDeviceAttribute is an attribute in a VirtualDevice model.

    The :meth:`.AbstractVirtualDevice.get` and
    :meth:`.AbstractVirtualDevice`set` methods are wrappers around the
    :class:`.VirtualDeviceAttribute` instances of a
    :class:`.VirtualDevice`.
    """

    __slots__ = ()

    @abstractmethod
    def get(self):
        """
        Return the value of an attribute.

        This method returns the current value held by the virtual device model.
        For an enterprise client, no REST API call is made to the server. For a
        device client, no call is made to the physical device.
        :return: the attribute value
        :raises: `ValueError` if the attribute value is not the correct
        return type.
        """
        pass

    @abstractmethod
    def getLastKnown(self):
        """
        Return the last known value of an attribute.

        This method returns the last known value of the attribute, which
        is not to be interpreted as the "previous" value.
        For a device client, the last known value is the device value.
        For an enterprise client, the last known value is the last value
        retrieved from the server for a specific attribute, irrespective of the
        local 'set' value.
        :return: the last known attribute value
        :raises: `ValueError` if the attribute value is not the correct
        return type.
        """
        pass

    @abstractmethod
    def set(self, value):
        """
        Set the value of an attribute.

        This method is used by the application
        to synchronize the `VirtualDevice` state with the physical device.
        This results in a REST API call to the server. For an enterprise
        client, an endpoint resource REST API call is made to the server. For
        a device client, an endpoint :class:`iotcs.shared.message.Type`
        message REST API call is made to the server.
        The value is validated according to the constraints in the device
        model. If the value is not valid, an `ValueError` is
        raised.
        :param value: the value to set
        :raises ValueError: if the value is not valid
        """
        pass

    @abstractmethod
    def setOnChange(self, callback):
        """
        Set a :class:`.ChangeCallback` callback instance.

        The callback is invoked when the value of an attribute changes on the
        remote virtual device.
        See :meth:`.AbstractVirtualDevice.setOnChange`
        :param callback: a :class:`.ChangeCallback` instance to invoke when an
        attribute value changes. If `callback` is `None`,  the existing
        callback isremoved

        """
        pass

    @abstractmethod
    def setOnError(self, callback):
        """
        Set an :class:`<.AbstractVirtualDevice.ErrorCallback>` instance.

        The callback is invoked when there is an error in setting the
        value of an attribute. An error may arise because of a network
        exception. In addition, on an enterprise client, an error indicates
        that a request timed out or was rejected by the device client.

        See :meth:`setOnError <.AbstractVirtualDevice.setOnError>`
        :param callback: a
        :class:`ErrorCallback <.AbstractVirtualDevice.ErrorCallback>`
        callback instance to invoke
        when there is an error. If `callback` is `None`, the existing callback
        is removed.
        """
        pass

# com.oracle.iot.client.impl


class VirtualDeviceAttributeBase(VirtualDeviceAttribute):
    """
    VirtualDeviceAttributeBase is an attribute in the virtualDevice model.

    It is the virtualization of :class:`.DeviceModelAttribute`
    """

    __slots__ = (
        '_virtualDevice',
        '_modelAttr',
        '_value',
        '_lastKnownValue',
        '_changeCallback',
        '_errorCallback',
    )

    def __init__(self, virtualDevice, modelAttr):
        """Create a :class:`.VirtualDeviceAttributeBase` instance."""
        super(VirtualDeviceAttributeBase, self).__init__()
        self._virtualDevice = virtualDevice
        self._modelAttr = modelAttr
        self._value = None
        self._lastKnownValue = None
        self._errorCallback = None
        self._changeCallback = None

    def getDeviceModelAttribute(self):
        """Return the :class:`.DeviceModelAttribute` instance."""
        return self._modelAttr

    def update(self, value):
        """
        Set the attribute's value with `value`.

        Sets the attribute's value but does not generate an event.
        :param value: the new attribute value
        :return: `True` if the value was updated
        """
        pass

    def isSettable(self):
        """
        Return `True` if  the attribute value may be set.

        Return `True` if the attribute can be set. On the device client, an
        attribute should always be settable. On the enterprise client, an
        attribute is settable if the model defines the attribute as writable.
        :return: `True` if the attribute's value can be set
        """
        pass

    def get(self):
        """
        Return the attribute's value.

        If the attribute's value has not been set return the
        :class:`.DeviceModelAttribute`'s default value.
        See :meth:`iotcs.client.VirtualDeviceAttribute.get`
        :return: the attribute's value or the default value.
        """
        return self._value if self._value is not None \
            else self.getDeviceModelAttribute().getDefaultValue()

    def getLastKnown(self):
        """
        Return the last known attribute value.

        See :meth:`iotcs.client.VirtualDeviceAttribute.getLastKnown`.
        """
        return self._lastKnownValue

    def setOnChange(self, callback):
        """
        Assign the :class:`.ChangeCallback`.

        See :meth:`iotcs.client.VirtualDeviceAttribute.setOnChange`.
        :param callack: a :class:`.ChangeCallbake` instance, if `None`,
        the existing callback is removed.
        """
        self._changeCallback = callback

    def getOnChange(self):
        """
        Return the :class:`.ChangeCallback` callback for this attribute.

        :return: the on change callback :class:`.ChangeCallback` or `None`
        """
        return self._changeCallback

    def setOnError(self, callback):
        """
        Assign the :class:`.ErrorCallback`.

        See :meth:`iotcs.client.VirtualDeviceAttribute.setOnError`.
        :param callback: a :class:`.ErrorCallback` instance, if `None`,
        the existing callback will be removed
        """
        self._errorCallback = callback

    def getOnError(self):
        """
        Return the :class:`.ErrorCallback` callback for this attribute.

        :return: the on error callback :class:`.ErrorCallback or`None`
        """
        return self._errorCallback

    def __str__(self):
        """Return a string representation of `.VirtualDeviceAttribute`."""
        m = self._modelAttr
        s0 = f'''name : {m.__name__}\n
            description : {str(m.getDescription())}\n
            alias : {str(m.getName())}\n
            model : {str(m.getModel())}\n
            access : {str(m.getAccess())}\n
            type : {str(m.getType())}\n
            '''
        s1 = ""
        modelType = m.getType()
        if modelType == DeviceModelAttribute.Type.NUMBER or modelType == \
                DeviceModelAttribute.Type.INTEGER:
            s1 = f'''lowerBound : {str(m.getLowerBound())}\n
                upperBound : {str(m.getUpperBound())}\n
                value : {str(self._value)}\n
                lastKnownValue : {str(self._lastKnownValue)}\n
                defaultValue : {str(self.getDeviceModelAttribute().getDefaultValue())}\n
                '''

        return f'{s0}{s1}'

    def validate(self, attribute, value):
        """
        Validate `value` of :class:`.DeviceModelAttribute.Type`.

        Ensure that the `value` native type is valid for the
        :class:`iotcs.shared.DeviceModelAttribute` `attribute`'s
        :class:`iotcs.shared.DeviceModelattribute.Type`.
        :param attribute: a :class:`.DeviceModelAttribute`
        :param value: a the attribute value
        :raises: `ArgumentException` if `value` is not a valid type for `attribute`
        """
        if value is None:
            return
        attrType = attribute.getType()
        if attrType == DeviceModelAttribute.Type.INTEGER:
            if not (isinstance(value, (int, ))):
                raise ArgumentException("value is not INTEGER")
        elif attrType == DeviceModelAttribute.Type.NUMBER:
            if not (isinstance(value, (float, ))):
                raise ArgumentException("value is not NUMBER")
        elif attrType == DeviceModelAttribute.Type.STRING:
            if not (isinstance(value, (str, ))):
                raise ArgumentException("value is not STRING")
        elif attrType == DeviceModelAttribute.Type.BOOLEAN:
            if not (isinstance(value, (bool, ))):
                raise ArgumentException("value is not BOOLEAN")
        elif attrType == DeviceModelAttribute.Type.DATETIME:
            if not (isinstance(value, (datetime, ))) and \
                    not (isinstance(value, (int, ))):
                raise ArgumentException("value is not DATETIME")
        elif attrType == DeviceModelAttribute.Type.URI:
            if not (isinstance(value, (ExternalObject, ))):
                raise ArgumentException("value is not an ExternalObject")
        if ((attrType == DeviceModelAttribute.Type.INTEGER) or
                (attrType == DeviceModelAttribute.Type.NUMBER)):
            val = value
            if attribute.getUpperBound() is not None:
                upper = attribute.getUpperBound()
                if val > upper:
                    raise ArgumentException(f'validate(): {val} > {upper}')
            if attribute.getLowerBound() is not None:
                lower = attribute.getLowerBound()
                if val < lower:
                    raise ArgumentException(f'validate(): {val} < {lower}')


# com.oracle.iot.client.impl.device

class VirtualDeviceAttributeImpl(VirtualDeviceAttributeBase):
    """VirtualDeviceAttribute is an attribute in the device model."""

    __slots__ = (
        '_virtualDevice'
    )

    def __init__(self, virtualDevice, deviceModelAttribute):
        """
        Create a :class:`.VirtualDeviceAttribute` instance.

        :param virtualDevice: the `VirtualDevice` with this attribute
        :param deviceModelAttribute: a `DeviceModelAttribute` represented
            by this `VirtualDeviceAttribute`
        """
        super(VirtualDeviceAttributeImpl, self).__init__(virtualDevice,
                                                         deviceModelAttribute)
        self._virtualDevice = virtualDevice

    def set(self, value):
        """
        Set the value of an attribute.

        See :meth:`.VirtualDeviceAttribute.set`
        :param value: the new value for the attribute

        """
        #  validate throws IllegalArgumentException if value is not valid
        self.validate(self.getDeviceModelAttribute(), value)
        # TODO: None value occurring somewhere
        # self.getLogger().debug("\nVirtualDevice: " +
        #   self.__virtualDevice.__str__() + "\n\t attributeName " +
        #   self.getDeviceModelAttribute().getName() + "\n\t attributeValue " +
        #   self.get() + "\n\t newValue " +
        #       (value if value is not None else "None") + "\n")
        self._lastKnownValue = self._value = value
        #  This may set up an infinite loop!
        self._virtualDevice.processOnChange(self, self._value)

    def update(self, value):
        """
        Set the attribute's value with `value`.

        See :meth:`.VirtualDeviceAttributeBase.update`
        :return: `True` if if the value was updated.
        """
        #  validate throws IllegalArgumentException if value is not valid
        self.validate(self._modelAttr, value)
        vdevstr = self._virtualDevice.__str__()
        attnamestr = self.getDeviceModelAttribute().getName()
        attvaluestr = f'{self.get()}' if self.get() is not None else "None"
        valuestr = f'{value}'

        self.getLogger().debug(
            msg=f'VirtualDevice:\n'
                f'{vdevstr}:'
                f'\tattributeName: {attnamestr},'
                f'\tattributeValue: {attvaluestr},'
                f'\tnewValue: {valuestr}')
        self._lastKnownValue = self._value = value
        return True

    def __eq__(self, obj):
        """Return `True` if this instance is equal to `obj`."""
        if obj is None:
            return False
        if id(self) == id(obj):
            return True
        if obj.__class__ != self.__class__:
            return False
        if not self._value != obj.__value if self._value is not None \
                else obj._value is not None:
            return False
        return self.getDeviceModelAttribute() == obj.getDeviceModelAttribute()

    def isSettable(self):
        """
        Return `True` if the attribute value may be set.

        See :meth:`.VirtualDeviceAttributeBase.isSettable`
        :return: `True` if the attribute's value can be set
        """
        return True

    # TODO: Use pyhon hash technique
    def __hash__(self):
        """Return the hash of this instance."""
        result = 37

        if self._value is not None:
            result = 37 * result + hash(self._value)
        else:
            result = 37 * result

        result = 37 * hash(self.getDeviceModelAttribute())
        return result

    @classmethod
    def getLogger(cls):
        """Return a logger for :class:`.VirtualDeviceAttributeImpl`."""
        return logging.getLogger("iotcs.client.attributes.VirtualDeviceAttributeImpl")
