# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

from abc import ABC
from abc import abstractmethod
from datetime import date, datetime
import logging
import json
import numbers
from threading import Thread
import threading
#import traceback

from iotcs.client import AbstractVirtualDevice
from iotcs.client import Client
from iotcs.client import ExternalObject
from iotcs.client import StorageObject
from iotcs.client.attributes import VirtualDeviceAttributeImpl
from iotcs.client.device.storage import StorageObjectImpl
from iotcs.client.device.util import MessageDispatcher
from iotcs.client.device.util import MessageDispatcherImpl
from iotcs.client.model import DeviceModelFormat
from iotcs.client.model import DeviceModelImpl
from iotcs.client.model import DeviceModelAttribute
from iotcs.common import Pair
from iotcs.common import StatusCode
from iotcs.common import TimeManager
from iotcs.diagnostics import Diagnostics
from iotcs.exception import ArgumentException
from iotcs.exception import NetworkException
from iotcs.exception import StateException
from iotcs.message import AlertMessage
from iotcs.message import DataItem
from iotcs.message import DataMessage
from iotcs.message import Message
from iotcs.message import ResponseMessage
from iotcs.messaging.client.device import DirectlyConnectedDeviceImpl
from iotcs.messaging.client.device import GatewayDeviceImpl
from iotcs.messaging.client.device.policy import DevicePolicy
from iotcs.messaging.client.device.policy import DevicePolicyManager
from iotcs.messaging.client.device.policy import ScheduledPolicyData
from iotcs.messaging.client.device.policy import  TimedPolicyThread
from iotcs.messaging.client.device.util import RequestHandler
from iotcs.messaging.client.storage import StorageConnectionBase
from iotcs.messaging.policy import Formula
from iotcs.messaging.client.device.policy import DeviceAnalog
from iotcs.messaging.policyfunctions import DeviceFunction

# oracle.iot.client.device

class Alert(ABC):
    """
    An Alert to be sent to the server.

    The alert is sent by calling the
    :meth:`.raise_` method. The time of the alert is set when :meth:`.raise_`
    is called, allowing :meth:`.raise_` to be called more than once.

    The :meth:`.set` method returns the :class:`Alert` instance to allow
    the fields of the alert to be set in fluent style.
    See :meth:`iotcs.client.device.VirtualDevice.createAlert`
    """

    __slots__ = ()

    @abstractmethod
    def set(self, field, value):
        """
        Set the value of a field in the `Alert`.

        The fields are determined by the format given when the `Alert` is
        created. The value is validated according to the constraints in the
        format. If the value is not valid, an `IllegalArgumentException` is
        raised.

        All fields defined in the format that are 'optional", `True`
        must be set before calling :meth:`.raise_`.

        :param field the name of a field from the alert format
        :param value the value to set
        :return: this Alert instance
        :raises ValueError: if the value is not valid,
        or `field` is `None`
        See :meth:`iotcs.client.device.VirtualDevice.createAlert`
        """
        pass

    # TODO: Need better solution for naming symbols that match python keywords
    @abstractmethod
    def raise_(self):
        """
        Send the alert. The event time is set when this method is called.

        The `onError` handler of the parent virtual device will be called
        if there is error sending the alert.

        All fields defined in the format that are "optional",  `True`
        must be set before calling :meth:`.raise_`. If :meth:`.raise_` is
        called before setting all "optional" `True` fields,
        `IllegalStateException` is raised.

        :raises ValueError: if :meth:`.raise_` is called
            if the alert message cannot be queued
            before setting all "optional", `True` fields
        """
        pass

class AlertImpl(Alert):

    __slots__ = (
        '__virtualDevice',
        '__deviceModelFormat',
        '__fieldValues'
    )

    def __init__(self, virtualDevice, alertUrn, deviceModel=None, msgDispatcher=None, endpointId=None):
        """
        Create an `AlertImpl` instance.

        :param virtualDevice: the virtual device creating this `Alert`,
            must not be `None`
        :param alertUrn: the alert format urn, must not be `None`
        :raises: :class:`.ArgumentException` if a required argument is not set
        """
        if alertUrn is None:
            raise ArgumentException(f'alertUrn must not be None')

        # TODO: JCK tests for VirtualDeviceImpl type
        if virtualDevice is None:
            raise ArgumentException(f'virtualDevice must not be None')
        self.__virtualDevice = virtualDevice

        self.__deviceModelFormat = \
            AlertImpl.getDeviceModelFormat(virtualDevice, alertUrn)
        if self.__deviceModelFormat is None:
            raise ArgumentException(f'{alertUrn} not found')

        if self.__deviceModelFormat.getType() != DeviceModelFormat.Type.ALERT:
            raise ArgumentException(f'{alertUrn} is not an alert format')

        self.__fieldValues = dict()

    def set(self, fieldName, value):
        """

        :param fieldName:
        :param value:
        :raises: :class:`.ArgumentException` if `fieldName` is `None`,
            `fieldName`'s type is not a :class:`.DeviceModelAttribute.Type`,
        :return: this `Alert`
        """
        field = self._getField(fieldName)
        if field is None:
            raise ArgumentException(f'{fieldName} is not in device model')

        fieldType = field.getType()
        if fieldType not in list(DeviceModelAttribute.Type):
            raise ArgumentException(
                f'{fieldName} has an invalid field type {fieldType}')

        if (fieldType is DeviceModelAttribute.Type.NUMBER and
                    isinstance(value, float)) or \
                (fieldType is DeviceModelAttribute.Type.INTEGER and
                    isinstance(value, int)) or \
                (fieldType is DeviceModelAttribute.Type.DATETIME and
                    isinstance(value, datetime)) or \
                (fieldType is DeviceModelAttribute.Type.BOOLEAN and
                    isinstance(value, bool)) or \
                (fieldType is DeviceModelAttribute.Type.STRING and
                    isinstance(value, str)) or \
                (fieldType is DeviceModelAttribute.Type.URI and
                    isinstance(value, ExternalObject)):
            self.__fieldValues[fieldName] = value
        else:
            raise ArgumentException(
                f'value for "{fieldName}" is not a {fieldType.name}')

        return self

    def raise_(self):
        """
        Raise an alert.
        :raises: :class:`.StateException` if a non optional field's value
            is not set
        """
        alertMessage = \
            AlertMessage(format=self.__deviceModelFormat.getURN(),
                         description=self.__deviceModelFormat.getDescription(),
                         source=self.__virtualDevice.getEndpointId(),
                         reliability=Message.Reliability.GUARANTEED_DELIVERY)
        fieldList = self.__deviceModelFormat.getFields()
        storageObject = None
        for field in fieldList:
            key = field.getName()
            value = None
            if key in self.__fieldValues:
                value = self.__fieldValues.pop(key)
            if value is not None:
                if field.getType() == DeviceModelAttribute.Type.NUMBER:
                    alertMessage.dataItem(key, float(value))
                elif field.getType() == DeviceModelAttribute.Type.INTEGER:
                    alertMessage.dataItem(key, int(value))
                elif field.getType() == DeviceModelAttribute.Type.DATETIME:
                    if isinstance(value, (datetime, )):
                        alertMessage.dataItem(key, (value).timestamp())
                    elif isinstance(value, (int, )):
                        alertMessage.dataItem(key, int(value))
                elif field.getType() == DeviceModelAttribute.Type.BOOLEAN:
                    alertMessage.dataItem(key, bool(value))
                elif field.getType() == DeviceModelAttribute.Type.STRING:
                    alertMessage.dataItem(key, str(value))
                elif field.getType() == DeviceModelAttribute.Type.URI:
                    if isinstance(value, (StorageObjectImpl, )):
                        storageObject = value
                        # TODO: See IOT-46453
                        # if (storageObject.getSyncStatus() ==
                        #         StorageObject.SyncStatus.NOT_IN_SYNC) or \
                        #         (storageObject.getSyncStatus() ==
                        #             StorageObject.SyncStatus.SYNC_PENDING):
                        #     pass
                        storageObject.setSyncEventInfo(
                            self.__virtualDevice, key)
                    alertMessage.dataItem(key, value.getURI())
                else:
                    self.__getLogger().info(
                        f'unknown type: {field.getType()}')
            elif not field.isOptional():
                self.__fieldValues.clear()
                raise StateException(f'non-optional field \'{key}\' not set')
        self.__fieldValues.clear()



        try:
            self.__virtualDevice.queueMessage(alertMessage, [storageObject])
        except Exception as e:
            msg = f'AlertImpl.raise_(): queueMessage() threw exception {e}'
            # traceback.print_exception(Exception, e)
            self.__getLogger().critical(msg)

    def _getField(self, fieldName):
        """
        Return the the `DeviceModelFormat.Field` with the name `fieldName` or
        None if the `fieldName` is not a field in the `DeviceModelFormat` for
        this `Alert`.

        :param fieldName: name of a `DeviceModelFormat.Field` for this `Alert`.
        :return: the `DeviceModelField.Format` with name `fieldName` else None
        """
        if self.__deviceModelFormat is None:
            return None
        fieldList = self.__deviceModelFormat.getFields()
        for field in fieldList:
            if field.getName() == fieldName:
                return field
        return None

    @staticmethod
    def getDeviceModelFormat(virtualDevice, alertUrn):

        dm = virtualDevice.getDeviceModel()
        if not (isinstance(dm, (DeviceModelImpl, ))):
            return None
        deviceModel = dm
        return deviceModel.getDeviceModelFormats().get(alertUrn)

    def __getLogger(self):
        return logging.getLogger("iotcs.client.device.AlertMessageImpl")

# oracle.iot.client.device

class Data(ABC):
    """
    Represents a set of custom data fields.

    The custom data fields are key, value pairs to be sent to the
    server. The custom data is sent by calling the
    :meth:`.submit` method. The time of the data fields are set when
    :meth:`.submit` is called, allowing :meth:`.submit` to be called more
    than once.

    The :meth:`.set` method returns the `Data` instance to allow
    fields of data to be set in fluent style.
    See :meth:`iotcs.client.device.VirtualDevice.createData`
    """

    __slots__ = ()

    @abstractmethod
    def set(self, field, value):
        """
        Set the value of a field in the `Data`.

        The fields are determined by the format given when the Data is created.
        The value is validated according to the constraints in the format.
        If the value is not valid, an IllegalArgumentException is raised.

        All fields defined in the format that are *"optional" : false*
        must be set before calling :meth:`.submit`.

        See :meth:`iotcs.client.device.VirtualDevice.createData`

        :param field: the name of a field from the custom data format
        :param value: the value to set
        :return: this :class:`.Data` instance
        :raises ValueError: if the value is not valid,
            or `field` is `None`
        """
        pass

    @abstractmethod
    def submit(self):
        """
        Submit the data.

        The event time is set when this method is called.
        The `onError` handler of the parent virtual device will be called
        if there is error submiting the custom data fields.
        See :meth:`iotcs.client.device.VirtualDevice.setOnError`

        All fields defined in the format that are *"optional" : false*
        must be set before calling :meth:`.submit`. If `submit` is
        called before setting all non *optional* fields, an
        IllegalStateException is raised.

        :raises ValueError: if `submit` is called
            before setting all `optional` fields
        """
        pass

class DataImpl(Data):
    """
    # TODO: inherit doc ?
    """
    # TODO: Make members consistent, use '_' or '__' but must be consistent
    # TODO: across classes
    __slots__ = (
        'virtualDevice', # `iotcs.device.impl.VirtualDeviceImpl`
        'dataUrn',
            # the format urn
        'deviceModelFormat',
            # `iotcs.device.impl.DeviceModelFormat` (in DeviceModelImpl.py)
        'fieldValues' # the custom data field values
    )

    def __init__(self, virtualDevice, dataUrn):
        """

        :param virtualDevice: the `iotcs.client.device.VirtualDevice` that
            owns the data
        :param dataUrn: the `iotcs.client.DeviceModelFormat` urn
        :raises: :class:`.ArgumentException` if `dataUrn` does not identify a
            known `DeviceModelFormat`
        """
        super(DataImpl, self).__init__()
        # TODO: JCL Asserts VirtualDeviceImpl ? Why what else could it be ?
        self.virtualDevice = virtualDevice
        self.dataUrn = dataUrn
        self.deviceModelFormat = self.__getDeviceModelFormat()
        if self.deviceModelFormat is None:
            raise ArgumentException(f'"{self.dataUrn}" not found.')
        if self.deviceModelFormat.getType() != DeviceModelFormat.Type.DATA:
            raise ArgumentException(
                f'"{dataUrn}" is not a DeviceModelFormat.Type.DATA format')
        self.fieldValues = dict()

    def set(self, fieldName, value):
        """

        :param fieldName:
        :param value:
        :return: this :class:`.DataImpl` instance
        :raises: :class:`.ArgumentException` if there is a type mismatch,
            between `fieldName` and `value`, `fieldName` is `None` or
            if `fieldName`'s type is not a :class:`.DeviceModelAttribute.Type`
        """
        field = self.__getField(fieldName)
        if field is None:
            raise ArgumentException(
                f'"{fieldName}" not in device model, {self.dataUrn}')

        fieldType = field.getType()
        if fieldType not in list(DeviceModelAttribute.Type):
            raise ArgumentException(
                f'{fieldName} has an invalid field type {fieldType}')

        if (fieldType is DeviceModelAttribute.Type.NUMBER and
                    isinstance(value, float)) or \
                (fieldType is DeviceModelAttribute.Type.INTEGER and
                    isinstance(value, int)) or \
                (fieldType is DeviceModelAttribute.Type.DATETIME and
                    isinstance(value, (int,datetime,))) or \
                (fieldType is DeviceModelAttribute.Type.BOOLEAN and
                    isinstance(value, bool)) or \
                (fieldType is DeviceModelAttribute.Type.STRING and
                    isinstance(value, str)) or \
                (fieldType is DeviceModelAttribute.Type.URI and
                    isinstance(value, ExternalObject)):
            self.fieldValues[fieldName] = value
        else:
            raise ArgumentException(
                f'value for "{fieldName}" is not a {fieldType.name}')

        return self

    def submit(self):
        """

        :raises: :class:`.StateException` if a non optional field's value is
            not set
        """
        fieldList = self.deviceModelFormat.getFields()
        storageObjects = []
        dataMessage = DataMessage(source=self.virtualDevice.getEndpointId(),
                                  format=self.deviceModelFormat.getURN())
        for field in fieldList:
            key = field.getName()
            if key not in self.fieldValues and not field.isOptional():
                self.fieldValues.clear()
                raise StateException(f'non-optional field "{key}" not set')

            value = self.fieldValues.pop(key)
            fieldType = field.getType()
            if fieldType == DeviceModelAttribute.Type.URI:
                if isinstance(value, (StorageObjectImpl)):
                    syncStatus = value.getSyncStatus()
                    if (syncStatus ==
                            StorageObject.SyncStatus.NOT_IN_SYNC) or \
                            (syncStatus ==
                                StorageObject.SyncStatus.SYNC_PENDING):
                        storageObjects.append(value)
                    value.setSyncEventInfo(self.virtualDevice, key)
                # value isa ExternalObject
                dataMessage.addItem(key, value.getURI())
            elif DeviceModelAttribute.Type(fieldType) is None:
                self._getLogger().info(f'unknown type: {fieldType.name}')
            else:
                dataMessage.addItem(key, value)

        self.fieldValues.clear()

        try:
            self.virtualDevice.queueMessage(dataMessage, storageObjects)
        except Exception as e:
            self._getLogger().critical(f'{e}')


    def __getField(self, fieldName):
        # TODO: Can't happen
        if self.deviceModelFormat is None:
            return None
        fieldList = self.deviceModelFormat.getFields()
        for field in fieldList:
            if field.getName() == fieldName:
                return field
        return None

    def __getDeviceModelFormat(self):
        """

        :return: The device model formats for `dataUrn`
        """
        dm = self.virtualDevice.getDeviceModel()
        # TODO: JCL asserts "isa" DeviceModelImpl. What else could it be ?
        return dm.getDeviceModelFormats().get(self.dataUrn)

    def _getLogger(self):
        return logging.getLogger("iotcs.client.device.DataImpl")

# oracle.iot.client.device

class DirectlyConnectedDevice(Client):
    """
    A `DirectlyConnnectedDevice` can send and receive messages.

    A `DirectlyConnectedDevice` sends and receives messages to and from an IoT
    server. The directly-connected device is assigned an
    *activation identifier* when the device is registered on the server.
    When the directly-connected device is activated, the server assigns an
    *endpoint identifer* that is used by the `DirectlyConnectedDevice`
    for sending messages to, and receiving messages from, the server.

    See :meth:`.activate`, :meth:`.getEndpointId`
    """

    # TODO: distinguish the messaging (low level) packaging
    # TODO: Currently only "iotcs.device.*Impl" is distinguishing this.
    __slots__ = (
        '_dcdImpl'  # com.oracle.iot.client.device.DirectlyConnectedDevice
    )

    # Support the no argument constructor, but this is partialy supports
    # it since both or none is the real api, no "context" support
    # There is also an internal single arg constructor that takes a low level
    # dcd that is used by the GatewayDevice, that passes the dcdImpl aggregate

    def __init__(self, path=None, passphrase=None, dcdImpl=None):
        """
        Create a `DirectlyConnectedDevice` instance.

        Load the device configuration from the given `path` and `passphrase`.
        See :ref:`Configuration` for details.

        :param path: configFilePath the path of the configuration file
        :param passphrase: configFilePassword the configuration file password,
            or `None` if the configurationFile is not encrypted
        :raises  ValueError: if the configuration could not be
            loaded, or the server scheme is not supported.
        """
        super(DirectlyConnectedDevice, self).__init__()

        # If path and passphrase are `None`, there is nothing to do at this
        # time, just return, otherwise try and create a dcdImpl.
        # When path and passphrase are None, it is a call to super from
        # GatewayDevice to quiet python when the super call is not made.
        if dcdImpl is not None:
            self._dcdImpl = dcdImpl
        else:
            self._dcdImpl = DirectlyConnectedDeviceImpl(path, passphrase)

        # TODO: Implement message persistence
        # bootstrap message persistence
        # messagePersistence = MessagePersistence.initMessagePersistence()
        # bootstrap BatchByPersistence
        # batchByPersistence = BatchByPersistence.initBatchByPersistence()

        self._bootstrapMessageDispatcher()

    # TODO: SHould this be *deviceModels ?
    def activate(self, deviceModels):
        """
        Activate the device.

        The device will be activated on the server if
        necessary. When the device is activated on the server, the server
        assigns an endpoint identifier to the device.

        If the device is already activated, this method will raise an
        `IllegalArgumentException`. The user should call the
        :meth:`isActivated()` method prior to calling activate.

        :param deviceModels: should contain the device model type URNs of this
            directlyconnected device.
            The device is activated with the given device models.
            The `deviceModels` parameter is zero or more, comma separated,
            device model URNs.
        :raises ValueError: if there is an I/O exception, when key or signature
            algorithm class cannot be loaded, or the key is not in
            the trusted assets store, or the private key is invalid or
            if the device has already been activated

        See :meth:`.isActivated`
        """
        # Make sure to add diagnostics and message_dispatcher capability URNs.
        # DCD impl activate takes the array and stuffs the elements into a Set,
        # so there is no need to worry about duplicates in the array.
        dmset = set()
        dmset.update([
            Diagnostics.DIAGNOSTICS_URN,
            DevicePolicyManager.CAPABILITY_URN,
            MessageDispatcher.MESSAGE_DISPATCHER_URN
        ])
        if deviceModels is not None:
            for m in deviceModels:
                dmset.add(m)

        self._dcdImpl.activate(dmset)

        self._bootstrapMessageDispatcher()

    def isActivated(self):
        """
        Return `True` if the device has been activated.

        :return: `True` if activated else `False`
        """
        return self._dcdImpl.isActivated()

    def getEndpointId(self):
        """
        Return the endpoint identifier of this directly connected device.

        The endpoint identifier is assigned by the server
        as part of the activation process.

        :return: the endpoint identifier of this directly-connected device.

        See :meth:`.activate`
        """
        return self._dcdImpl.getEndpointId()

    # TODO: inherit doc
    def createVirtualDevice(self, endpointId, deviceModel):
        """
        Create a :class:`iotlic.clien.device.VirtualDevice` instance.

        The virtual device is created with the given device model
        and given device identifier.

        This method creates a new :class:`iotcs.client.device.VirtualDevice`
        instance for the given parameters. The client library
        does not cache previously created VirtualDevice objects.

        :param endpointId: The device identifier of the device being modeled
        :param deviceModel: The device model URN that this device implements
        :raises ValueError: if the device does not implement the
            model, if deviceId or deviceModel are `None`
        :return: a new VirtualDevice
        """
        if endpointId is None:
            raise(ValueError("enpodintId cannot be None"))

        if deviceModel is None:
            raise(ValueError("deviceModel cannot be None"))

        return VirtualDeviceImpl(self._dcdImpl, endpointId, deviceModel)

    def getDeviceModel(self, urn):
        """
        Get the `DeviceModel` for the device model URN.

        This method may return `None` if there is no device model for the URN.
        `None` may also be returned if the device model is a "draft" and the
        property `allow_draft_device_models` is
        set to `False`, which is the default.

        :param urn: the URN of the device model
        :return: A representation of the device model
            :class:`iotcs.client.DeviceModel` or `None` if it does not exist
        :raises ValueError: if deviceModel is `None`,
            if there is an I/O error when communicating
            with the server or when key or signature algorithm class
            cannot be loaded, or the key is not in the trusted assets store,
            or the private key is invalid
        """
        return self._dcdImpl.getDeviceModel(urn)

    def close(self):
        """
        Close the directly connected device and its resources.

        :raises IOEerror: if an I/O error occurs
        """
        try:
            # if not activated, 'close()' call produce Exception, and not
            # needed in fact
            if self.isActivated():
                messageDispatcher = \
                    MessageDispatcher.removeMessageDispatcher(self._dcdImpl)
                if messageDispatcher is not None:
                    messageDispatcher.close()
                    self._getLogger().info("messageDispatcher closed")

        finally:
            # If MessageDispatcher close throws an exception, we still call
            # messagingDeviceClient.close(). If MessageDispatcher close throws
            # an exception AND messagingDeviceClient.close() throws an
            # exception, then the messagingDeviceClient.close() exception is
            # thrown. If MessageDispatcher close throws an exception and
            # messagingDeviceClient.close() exits normally, then the exception
            # from MessageDispatcher close is thrown.
            self._dcdImpl.close()
            self._getLogger().info("_dcdImpl closed")

    # JCL has two methods, only use one here with a default value, similar in JCL
    # but here there is no backward compatibility issue.

    def createStorageObject(self, name, contentType):
        """
        Create a new `StorageObject`.

        The created storage object that will have a name with the given
        object name prefixed with the device's endpoint ID and a directory
        separator. The prefix addition can be disabled by setting the
        :mod:`.config` property *disable_storage_object_prefix* in section
        *media_storage* to `True`.

        if `contentType` is `None`, the mime-type defaults to
        *application/octet-stream*.

        :param name the unique name to be used to reference the content in
        storage
        :param contentType The mime-type of the content or `None`
        :return a :class:`StorageObject <.messaging.client.device.StorageObjectDelegate`  # noqa: E501
        :raises TransportError: if exception raised by the
        runtime, or an abnormal response from the storage cloud
        :raises SecurityException: if there is an exception establishing
        a secure connection to the storage cloud
        """
        # delegate "isa" iotcs.messaging.client.StorageObject
        delegate = self._dcdImpl.createStorageObject(None, name, contentType)
        return StorageObjectImpl(self._dcdImpl, delegate)


    def _bootstrapMessageDispatcher(self):
        """Bootstrap the :class:`.MessageDispatcher` if activated."""
        if not self.isActivated():
            return

        messageDispatcher = MessageDispatcherImpl(self._dcdImpl)
        MessageDispatcher.setMessageDispatcher(self._dcdImpl,
                                               messageDispatcher)

    def _getLogger(self):
        return logging.getLogger("iotcs.client.device.DirectlyConnectedDevice")

class GatewayDevice(DirectlyConnectedDevice):
    """
    A GatewayDevice is a directly connected device.

    A GatewayDevice differs from other directly connected devices in that
    it is capable of registering indirectly connected devices
    and proxies messages for indirectly connected devices.
    """

    MANUFACTURER = "manufacturer"
    """The `manufacturer` attribute. This attribute is to be used
    when setting the `manufacturer` value in the meta-data."""

    MODEL_NUMBER = "modelNumber"
    """The `modelNumber` attribute. This attribute is to be used
    when setting the `modelNumber` value in the meta-data."""

    SERIAL_NUMBER = "serialNumber"
    """The `serialNumber` attribute. This attribute is to be used
    when setting the `serialNumber` value in the meta-data."""

    DEVICE_CLASS = "deviceClass"
    """The `deviceClass` attribute. This attribute is to be used
    when setting the `deviceClass` value in the meta-data."""

    PROTOCOL = "protocol"
    """The `protocol` attribute. This attribute is to be used
    when setting the `protocol` value in the meta-data."""

    PROTOCOL_DEVICE_CLASS = "protocolDeviceClass"
    """The `protocolDeviceClass` attribute. This attribute is to be used
    when setting the `protocolDeviceClass` value in the meta-data."""

    PROTOCOL_DEVICE_ID = "protocolDeviceId"
    """The `protocolDeviceId` attribute. This attribute is to be used
    when setting the `protocolDeviceId` value in the meta-data."""

    __slots__ = (
    )

    def __init__(self, path, passphrase):
        """
        Create a `GatewayDevice` instance.

        The device configuration is loaded from the given file `path` and
        `passphrase`.
        See :ref:`Configuration` for details.

        :param path: the path of the configuration file
        :param passphrase: the configuration file passphrase,
            or `None` if the configurationFile is not encrypted
        :raises ValueError: if the configuration file could not be loaded
        """
        # TODO: Need to refactor class hierarchy to properly initialize and
        # TODO: bootrstap. There are inconsistencies with JCL with respect
        # TODO: inheritance hierarchy and code sequence
        gd = GatewayDeviceImpl(path, passphrase)
        super(GatewayDevice, self).__init__(dcdImpl=gd)

    def registerDevice(self, hardwareId, metaData, deviceModels,
                       restricted=True):
        """
        Register an indirectly-connected device with the cloud service.

        The `restricted` parameter controls whether or not the client
        library is *required* to supply credentials for activating
        the indirectly-connected device. The client library will
        *always* supply credentials for an indirectly-connected
        device whose trusted assets have been provisioned to the client.
        If however, the trusted assets of the indirectly-connected device
        have not been provisioned to the client, the client library can
        create credentials that attempt to restrict the indirectly connected
        device to this gateway device.

        If the `restricted` parameter is `True` the indirectly-connected
        device cannot be activated by this gateway device without presenting
        credentials. If `restricted` is `True`, the client library will provide
        credentials to the server. The server will reject the activation
        request if the indirectly connected device is not allowed to roam to
        this gateway device.

        If the `restricted` parameter is `False` the indirectly-connected
        can be activated without presenting credentials if the trusted assets
        of the indirectly-connected device have not been provisioned to the
        client. If `restricted` is `False`, the client library will provide
        credentials if, and only if, the credentials have been provisioned to
        the client. The server will reject the activation if credentials are
        required but not supplied, or if the provisioned credentials do not
        allow the indirectly connected device to roam to this gateway device.

        The `hardwareId` is a unique identifier within the cloud service
        instance and may not be `None`. If one is not present for the device,
        it should be generated based on other metadata such as,
        *model niumber*, *manufacturer*, *serial number*, etc.

        The `metaData` `dict()` should typically contain all the standard
        metadata (the constants documented in this class) along with any other
        vendor defined metadata.

        :param restricted: indicate whether or not credentials are required for
            activating the indirectly connected device
        :param hardwareId: an identifier unique within the Cloud Service
            instance
        :param metaData: The metadata of the device
        :param deviceModels: should contain the device model type URNs
            supported by the indirectly connected device. The `deviceModels`
            parameter is one or more, comma separated, device model URNs.
        :return: The endpoint id of the indirectly-connected device
        :raises IOError: if the message could not be sent
        :raises ValueError: when key or signature algorithm class cannot be
            loaded, or the key is not in the trusted assets store, or the
            private key is invalid or if `hardwareId` is null,
            or `deviceModels` is null or zero length
        """
        return self._dcdImpl.registerDevice(hardwareId, metaData, deviceModels,
                                            restricted)


# oracle.iot.client.device


class VirtualDevice(AbstractVirtualDevice):
    """
    The `VirtualDevice` to represent a device-client.

    The `VirtualDevice` provides methods to handle write-only and
    executable actions.
    """

    __slots__ = ()

    class Callable(ABC):
        """
        Callback interface for device model actions.

        This class is implemented to handle
        :class:`iotcs.client.device.DeviceModel` actions.
        :deprecated:
        """

        __slots__ = ()

        @abstractmethod
        def call(self, virtualDevice, data):
            """
            Handle a device model action.

            For an execute action the client library will pass `None` as the
            `data` parameter. For a write-only action, the type of `data`
            should match the expected data type of the action.

            :param  virtualDevice: the
                :class:`iotcs.client.device.VirtualDevice`
                on which the action is being invoked
            :param  data: the data or `None`
            """
            pass

    class ActionEvent(AbstractVirtualDevice.Event):
        """
        An :class:`.Event` class that is passed to :class:`.ActionCallback` callbacks.
        """
        __slots__ = ()
        @abstractmethod
        def getName(self):
            """
            Return the action name.
            :return: the action name
            """
            pass

    class ActionCallback(ABC):
        """
        A callback interface for receiving notification of a :class:`.DeviceModelAction`.
        """
        __slots__ = ()
        def onAction(self, actionEvent):
            """
            Called to perform a :class:`.DeviceModelAction` action.

            :param actionEvent: the :class:`.ActionEvent`
            """
            pass

    def setOnAction(self, callback, actionName=None):
        """
        Set a callback for a specific action or all actions.

        If `actionName` is `None`, `callback` is invoked any all requested actions.
        If `actionName is not `None` then `callback` is invoked for that specific `actionName`

        If there is a callback for a specific action and for all actions,
        both callbacks will be invoked, with the specific action invoked first.
        :param actionName: The action name from the device model, or `None`
        :param callback: the :class:`.ActionCallback` that is inkoked. If `None` an
        existing callback is removed.
        """
        pass


    def setCallable(self, actionName, callable):
        """
        Set a :class:`.Callable` instance for handling an action, `actionName`.

        :param actionName: The action name
        :param callable: The :class:`.Callbable` instance to invoke
        :deprecated:
        """
        pass

    def createAlert(self, alertformaturn):
        """
        Create an Alert for this VirtualDevice.

        The alert will be created with the given format URN.
        :param alertformaturn: the alert format URN.
        :return: an `iotcs.client.Device.Alert`
        """
        pass

    def createData(self, messageformaturn):
        """
        Create a custom data object for this VirtualDevice.

        The custom data object will be created with the given message format
        URN.
        :param messageformaturn: the custom data message format URN.
        :return: a custom data object :class:`iotcs.client.device.Data`
        """
        pass

    def offer(self, attributeName, value):
        """
        Offer to set the value of an attribute.

        The attribute value is set depending upon any policy that may have been
        configured for the attribute. If there is no policy for the
        given attribute, offer behaves as if the
        `AbstractVirtualDevice.set(String, Object)` method were called.
        The value is validated according to the constraints in the device
        model.
        If the value is not valid, an IllegalArgumentException is raised.
        :param attributeName: the name of an attribute from the device type
        model
        :param value: the value to set
        :return: this VirtualDevice instance
        :raises ArgumentException: if the value is not valid,
        the attribute is not in the model,
        or `attributeName` is `None`
        """
        pass

    # def __init__(self):
    #     '''
    #     Compiler-required constructor for derived classes.
    #     '''
    #     super(VirtualDevice, self).__init__()


# com.oracle.iot.client.impl

class VirtualDeviceBase():
    """Implementation of VirtualDevice."""

    class Adapter(ABC):
        """
        An interface for adapting to the device or enterprise client.

        Different implementations of the virtual device interfaces
        :class:`.AbstractVirtualDevice` are required for device clients and
        enterprise clients.
        """

        __slots__ = ()

        @abstractmethod
        def setValue(self, deviceAttribute, value):
            """
            Set the the value for attribute.

            For enterprise client make a resource request to set the
            the attribute. For device client calls the attribute's set method.
            Invokes the on error handlers on error.
            :param deviceAttribute: the device attribute name
            :param value: the device attribute value
            """
            pass

        @abstractmethod
        def getAttribute(self, name):
            """
            Return a :class:.`VirtualDeviceAttribute` instance.

            The virtual device attribute is appropriate for either the
            enterprise or device client.
            :param name: the name of the attribute
            :return: The VirtualDeviceAttribute for the given name
            """
            pass

        @abstractmethod
        def updateFields(self, updatedAttributes):
            """
            Set all the attributes in an update batch.

            Errors are handled in the set call, including calling the on error
            handler.
            :param updatedAttributes: The attributes modified during an update
            """
            pass

    __slots__ = (
        'adapter',
        'endpointId',
        'deviceModel',
        'updateMode',
        'UPDATE_LOCK',
        'SYNC_LOCK',
        'changeCallback',
        'errorCallback',
        'updatedAttributes'
    )

    def __init__(self, adapter, endpointId, deviceModel):
        """
        Create a :class:`.VirtualDeviceBase` instance.

        :param adapter the device or enterprise specific implementation
        of the Adapter interface
        :param endpointId: the server's endpoint id for the device
        :param deviceModel: the device model the device implements
        """
        self.adapter = adapter
        self.endpointId = endpointId
        self.deviceModel = deviceModel
        self.updateMode = False
        #  The attributes modified during an update
        self.updatedAttributes = []
        #  Lock for update changes
        self.UPDATE_LOCK = threading.Lock()
        self.SYNC_LOCK = threading.Lock()
        self.changeCallback = None
        self.errorCallback = None

    def getDeviceModel(self):
        """Return the :class:`.DeviceModel` instance for the virtual device."""
        return self.deviceModel

    def getEndpointId(self):
        """Return the device client identifier."""
        return self.endpointId

    def set(self, attributeName, value):
        """
        Set the value of `attributeName` to `value`.

        If :meth:`.isUpdateMode` is `True`, set the value on the instance,
        and do not send any messages or change events.
        If :meth:`.isUpdateMode` is `False`, set the value and send messages
        and change events.
        :param attributeName: the name of the attibutes to change
        :param value: the new attribute value.
        """
        # Consider moving validation from adapter.setValue and pass
        # VirtualDeviceAttribute instance.
        # Validate here for updateSetField.
        # getAttribute throws IllegalArgumentException if the attribute
        # is not in the model.
        df = self.adapter.getAttribute(attributeName)
        if not df.isSettable():
            # TODO: Create error classes UnsupportedOperationException
            raise Exception(
                f"attempt to modify read-only attribute '{attributeName}")

        #  Will throw IllegalArgumentException
        df.validate(df.getDeviceModelAttribute(), value)
        if not self.isUpdateMode():
            self.adapter.setValue(df, value)
        else:
            self.updateSetField(df, value)

    def update(self):
        """Open an update transaction."""
        with self.UPDATE_LOCK:
            if self.updatedAttributes is None:
                self.updatedAttributes = []
            self.updateMode = True
        return self

    def finish(self):
        """End the :meth:`.update` mode and send the changes to the server."""
        with self.UPDATE_LOCK:
            if self.updatedAttributes is not None and \
                    len(self.updatedAttributes) > 0:
                workingCopy = list(self.updatedAttributes)
                self.adapter.updateFields(workingCopy)
                self.updatedAttributes.clear()

            self.updateMode = False

    def setOnChange(self, callback):
        """
        Set the :class:`.ChangeCallback` callback instance.

        See :class:`.AbstractVirtualDevice.ChangeCallback`
        :param callback: a :class:`.ChangeCallback` instance or `None`
        :return:
        """
        with self.SYNC_LOCK:
            self.changeCallback = callback

    def getOnChangeCallback(self):
        """
        Return the :class:`.ChangeCallback` callback instance or `None`.

        See :class:`.AbstractVirtualDevice.ChangeCallback`
        :return: the :class:`.ChangeCallback` callback instance or `None`
        """
        with self.SYNC_LOCK:
            return self.changeCallback

    def setOnError(self, callback):
        """
        Set the :class:`.ErrorCallback` callback instance.

        See :class:`.AbstractVirtualDevice.ErrorCallback`
        :param callback: a :class:`.ErrorCallback` instance or `None`
        :return:
        """
        with self.SYNC_LOCK:
            self.errorCallback = callback

    def getOnErrorCallback(self):
        """
        Return the :class:`.ErrorCallback` callback instance or `None`.

        See :class:`.AbstractVirtualDevice.ErrorCallback`
        :return: the :class:`.ErrorCallback` instance or `None`
        """
        with self.SYNC_LOCK:
            return self.errorCallback

    def isUpdateMode(self):
        """
        Return `True` if the device is in update mode.

        :return: `True` if in update mode else `False`
        """
        with self.UPDATE_LOCK:
            return self.updateMode

    def updateSetField(self, attribute, value):
        """
        Add attribute `attribute` to the set of attributes for batch update.

        :param attribute: the attribute being set
        :param value: the attribute value
        """
        with self.UPDATE_LOCK:
            self.updatedAttributes.append(Pair(attribute, value))

    def __str__(self):
        """Return the string representation of this virtual device."""
        return "{" + self.endpointId + " , " + self.deviceModel.getURN() + "}"

    class NamedValueImpl(AbstractVirtualDevice.NamedValue):
        """The implementation of :class:`.AbstractVirtualDevice.NamedValue`."""

        __slots__ = (
            '__name',
            '__value',
            '__next'
        )

        def __init__(self, name, value):
            """
            Create a :class:`.NamedValue` instance.

            :param name:  the name of the value
            :param value: the value
            """
            super(VirtualDeviceBase.NamedValueImpl, self).__init__()
            self.__name = name
            self.__value = value
            self.__next = None

        def getName(self):
            """Return the value name."""
            return self.__name

        def getValue(self):
            """Return the value."""
            return self.__value

        def next(self):
            """Return the next :class:`.NamedValue` instance."""
            return self.__next

        def setNext(self, next):
            """Set the next :class:`.NamedValue` instance."""
            self.__next = next

    class ChangeEvent(AbstractVirtualDevice.ChangeEvent):
        """An event that describes a change to the virtual device."""

        __slots__ = (
            'virtualDevice',
            'value'
        )

        def __init__(self, virtualDevice, value):
            """
            Create a :class:`.ChangeEvent` instance.

            :param virtualDevice: a :class:`.VirtualDevice` instance
            :param value: the :class:`.NamedValue` related to this change
            """
            super(VirtualDeviceBase.ChangeEvent, self).__init__()
            self.virtualDevice = virtualDevice
            self.value = value

        def getVirtualDevice(self):
            """Return the :class:`.VirtualDevice` instance."""
            return self.virtualDevice

        def getNamedValue(self):
            """Return the :class:`.NamedValue` instance."""
            return self.value

    class ErrorEvent(AbstractVirtualDevice.ErrorEvent):
        """An event that describes an error."""

        __slots__ = (
            'virtualDevice',
            'value',
            'message'
        )

        def __init__(self, virtualDevice, value, message):
            """
            Create a :class:`.VirtualDeviceBase` instance.

            :param virtualDevice: a :class:`.VirtualDevice` instance
            :param value: the :class:`.NamedValue` related to this error
            :param message: the error event message
            """
            super(VirtualDeviceBase.ErrorEvent, self).__init__()
            self.virtualDevice = virtualDevice
            self.value = value
            self.message = message

        def getVirtualDevice(self):
            """Return the :class:`.VirtualDevice` instance."""
            return self.virtualDevice

        def getNamedValue(self):
            """Return the :class:`.NamedValue` instance."""
            return self.value

        def getMessage(self):
            """Return the error message."""
            return self.message


class VirtualDeviceImpl(VirtualDevice, VirtualDeviceBase.Adapter,
                        RequestHandler, DeviceAnalog,
                        DevicePolicyManager.ChangeListener):

    __slots__ = (
        '__base',
        # Used from AlertImpl
        '__directlyConnectedDevice',
        '__attributeMap',
        '__UPDATE_LOCK',
        '__ACTIONMAP_LOCK'
        '__messageDispatcher',
        '__requestDispatcher',
        '__devicePolicyManager',
        ## Data pertaining to this virtual device and its attributes for
        ## computing policies.
        ## The key is attribute name (or null for the device model policies),
        ## value is a list. Each element in the list corresponds to a function
        ## in the pipeline for the attribute, and the map is used by the
        ## function to hold data between calls. Note that there is a 1:1
        ## correspondence between the configuration
        ## data for the attribute, and the pipeline data for the attribute.
        '__pipelineDataCache',
        '__computedMetricTriggerMap',
        '__computedMetricTriggerMapLock',
        '__scheduledPolicies',
        # Window based policy support (as in "window", not Windows OS).
        # Have one scheduled task for each policy "slide" value. The slide
        # value is how much to move the window, so we want to run the policy
        # when the slide expires. When the slide expires, the runnable will
        # call back each VirtualDeviceAttribute that has a policy for that
        # slide value. Window and slide are the key.
        # { {window,slide} : ScheduledPolicyData }
        '__scheduledPoliciesLock',
        '__timedPolicyThread',
        # action maps
        '__actionCallbackMap',
        '__callableMap'

    )

    def __init__(self, dcd,  endpointId, deviceModel):

        # TODO: How should the relationship between VirtualDeviceImpl and
        # TODO: and VirtualDeviceBase really work ? Implements and
        # TODO: delegates ?

        # TODO: Serious investigation is needed to understand how
        # TODO: multiple inheritance works in the absence of "interface"
        # TODO: and "implements"
        #super(VirtualDeviceImpl,self).__init__()

        self.__base = VirtualDeviceBase(self, endpointId, deviceModel)
        self.__UPDATE_LOCK = threading.RLock()
        self.__ACTIONMAP_LOCK = threading.RLock()
        self.__directlyConnectedDevice = dcd
        self.__attributeMap = self._createAttributeMap(self, deviceModel)
        self.__messageDispatcher = MessageDispatcher.getMessageDispatcher(dcd)
        self.__requestDispatcher = self.__messageDispatcher.getRequestDispatcher()
        self.__computedMetricTriggerMap = []
        self.__computedMetricTriggerMapLock = threading.RLock()
        self.__scheduledPolicies = dict()
        self.__scheduledPoliciesLock = threading.RLock()
        self.__timedPolicyThread = TimedPolicyThread(self)

        self.__actionCallbackMap = dict()
        self.__callableMap = dict()

        # handler for all attribute and action requests
        self.__requestDispatcher.registerRequestHandler(
            endpointId, f'deviceModels/{deviceModel.getURN()}', self)

        _ERROR_CALLBACK_BRIDGE.add(self)
        self.__messageDispatcher.setOnError(_ERROR_CALLBACK_BRIDGE)

        ## device policy related stuff
        self.__pipelineDataCache = dict()
        self.__devicePolicyManager = \
            DevicePolicyManager.getDevicePolicyManager(dcd)
        self.__devicePolicyManager.addChangeListener(self)


    def _createAttributeMap(self, virtualDevice, deviceModel):
        """
        Return a device model attribute map.
        :param virtualDevice:
        :param deviceModel:
        :return: a device model attribute map
        """
        map = dict()
        if isinstance(deviceModel, DeviceModelImpl):
            deviceModelImpl = deviceModel
            for attribute in deviceModelImpl.getDeviceModelAttributes().values():
                vda = VirtualDeviceAttributeImpl(virtualDevice, attribute)
                map[attribute.getName()] =  vda
                alias = attribute.getName()
                if alias != None and alias != "":
                    map[alias] = vda
        return map

    # VirtualDeviceAttributeBase<VirtualDevice, Object>
    def getAttribute(self, attributeName):
        """
        VirtualDeviceBase.Adapter API
        {@inheritDoc}
        :param attributeName:
        :return: the `VirtualDeviceAttributeBase` attribute for `attributeName`
        :raises: :class:`.ArgumentException` if `attributeName` is not an
            attribute in the model.
        """
        if attributeName in self.__attributeMap and \
                self.__attributeMap[attributeName] is not None:
            return self.__attributeMap[attributeName]
        else:
            raise(ArgumentException("no such attribute '" + attributeName +
                "'.\n\tVerify that the URN for the device model you created " +
                "matches the URN that you use when activating the device in " +
                "the PCL application.\n\tVerify that the attribute name " +
                "(and spelling) you chose for your device model matches the " +
                "attribute you are setting in the PCL application."))

    def setValue(self,  attribute, value):
        """
        VirtualDeviceBase.Adapter API.

        #{@inheritDoc}
        :param attribute: VirtualDeviceAttributeBase
        :param value: obejct
        :raises: :class:`.ArgumentException` if `attribute` is `None`
        """
        if attribute is None:
            raise(ValueError("attribute may not be null"))

        attribute.set(value)

    def updateFields(self, updatedAttributes):
        """
        Set all the attributes in an update batch. Errors are handled in the
        set call, including calling the on error handler.
        {@inheritDoc}
        :param updatedAttributes: List<Pair<VirtualDeviceAttributeBase<VirtualDevice, Object>, Object>>
        """
        updatedAttributeSet = set()
        with self.__UPDATE_LOCK:
            for pair in updatedAttributes:
                attribute = pair.getKey()
                value = pair.getValue()
                attributeName = attribute.getDeviceModelAttribute().getName()
                try:
                    # Here, we assume
                    # 1. That attribute is not null. If the attribute were not found
                    #    an exception would have been thrown from the VirtualDevice
                    #    set(String attributeName, T value) method.
                    # 2. That the set method validates the value. The call to
                    #    update here should not throw an exception because the
                    #    value is bad.
                    # 3. The author of this code knew what he was doing.
                    if not attribute.update(value):
                        #iterator.remove();
                        pass
                    else:
                        updatedAttributeSet.add(attributeName)

                except Exception as e:
                    self._getLogger().warning(f'updateFields(): {e}')

                finally:
                    DeviceFunction.removeInProcessValue(self.getEndpointId(),
                       self.getDeviceModel().getURN(), attributeName)

                # Here is the check to see if the updated attributes will trigger computedMetrics.
                # There returned set are the attributes whose computedMetrics resulted in an
                # attribute.update(value). Those attributes are added to the list of updated attributes
                # so that they are included in the resulting data message.
                computedMetrics = self.checkComputedMetrics(updatedAttributeSet)
                if len(computedMetrics) > 0:
                   for attr in computedMetrics:
                       attributeBase = self.getAttribute(attr)
                       value = attributeBase.get()
                       # Pair<VirtualDeviceAttributeBase<VirtualDevice, Object>, Object>
                       pair = Pair(attributeBase, value)
                       updatedAttributes.append(pair)

        self._processOnChange(updatedAttributes)

    def getDeviceModel(self):
        return self.__base.getDeviceModel()

    def getEndpointId(self):
        return self.__base.getEndpointId()

    def get(self, attributeName):
        attribute = self.getAttribute(attributeName)
        return attribute.get()

    def getLastKnown(self, attributeName):
        #VirtualDeviceAttributeBase<VirtualDevice, Object> \
        attribute = self.getAttribute(attributeName)
        return attribute.getLastKnown()

    def set(self, attributeName,value):
        self.__base.set(attributeName, value)
        return self

    def update(self):
        self.__base.update()
        return self

    def finish(self):
        self.__base.finish()

    def offer(self, attributeName, value):
        """

        :param attributeName:
        :param value:
        :return:
        """
        attribute = self.getAttribute(attributeName)
        if not attribute.isSettable():
            raise StateException(f'attempt to modify read-only attribute "{attributeName}"')

        devicePolicy = self.__devicePolicyManager.getPolicy(
            self.getDeviceModel().getURN(), self.getEndpointId())

        if devicePolicy is None:
            return self.set(attributeName, value)

        pipeline = devicePolicy.getPipeline(attributeName)
        if pipeline is None or len(pipeline) == 0:
            return self.set(attributeName, value)

        pipelineData = self.getPipelineData(attributeName)

        policyValue = self.offer0(attribute.getDeviceModelAttribute(), value,
                                  pipeline, pipelineData)
        if policyValue is not None:
            # FINE
            self._getLogger().debug(f'offer(): {self.getEndpointId()}'
                f' : Set   : "{attributeName}"={policyValue}')
            if self.__base.isUpdateMode():
                self.set(attributeName, policyValue)
            else:
                #//
                #// Handle calling offer outside of an update when there are computed metrics involved.
                #//
                #// Call updateFields ito ensure the computed metrics get run,
                #// and will put this attribute and computed attributes into
                #// one data message.
                #//
                #ArrayList<Pair<VirtualDeviceAttributeBase<VirtualDevice, Object>, Object>>(1)
                updatedAttributes =[]
                updatedAttributes.append(Pair(attribute, policyValue))
                self.updateFields(updatedAttributes)
        return self

    #// This is the main logic for handling a policy pipeline.
    #// Returns true if attribute is set.
    def offer0(self, attribute, value, pipeline, pipelineData):
        attributeName = attribute.getName()
        policyValue = value
        if pipeline is not None and len(pipeline) > 0:
            # offer0 can be entered by different threads.
            # make sure only one at a time thread can modify
            with self.__UPDATE_LOCK:
                # // InProcessValue helps resolve $(<attribute-name>) in formulas and conditions
                DeviceFunction.putInProcessValue(self.getEndpointId(),
                                                 self.getDeviceModel().getURN(),
                                                 attributeName, policyValue)
                index = 0
                maxIndex = len(pipeline)
                while index < maxIndex:
                    plfunction = pipeline[index]
                    functionData = None
                    if index < len(pipelineData):
                        functionData = pipelineData[index]
                    else:
                        functionData = dict()
                        pipelineData.append(functionData)

                    index += 1

                    key = plfunction.getId()
                    parameters = plfunction.getParameters()
                    deviceFunction = DeviceFunction.getDeviceFunction(key)
                    if deviceFunction is None:
                        continue
                    if deviceFunction.apply(self, attributeName, parameters,
                                            functionData, policyValue):
                        valueFromPolicy = deviceFunction.get(self,
                                                             attributeName,
                                                             parameters,
                                                             functionData)
                        if valueFromPolicy is not None:
                            policyValue = DeviceModelAttribute.Type.cast(
                                attribute.getType(), valueFromPolicy,
                                convert=True)
                            DeviceFunction.putInProcessValue(
                                self.getEndpointId(),
                                self.getDeviceModel().getURN(),
                                attributeName, policyValue)
                        else:
                            # FINEST
                            self._getLogger().debug(f'offer0(): {attributeName} '
                                f'" got None value from policy {deviceFunction.getDetails(parameters)}')
                            return None

                    else:
                        # FINEST (all in "if logging FINEST"
                        if deviceFunction.getId().startswith("filter"):
                            self._getLogger().debug(f'offer0(): '
                                                   f'VirtualDevice: {self.getEndpointId()} : offer0 '
                                                   f'"{attributeName}"  = "{policyValue}" rejected by policy '
                                                   f'"{deviceFunction.getDetails(parameters)}"')
                        return None

        return policyValue

    def setOnChange(self, callback, attributeName=None):
        if attributeName is None:
            self.__base.setOnChange(callback)
        else:
            attribute = self.getAttribute(attributeName)
            attribute.setOnChange(callback)

    def getErrorCallback(self):
        return self.__base.getOnErrorCallback()

    def setOnError(self, callback, attributeName=None):
        if attributeName is None:
            self.__base.setOnError(callback)
        else:
            attribute = self.getAttribute(attributeName)
            attribute.setOnError(callback)

    def handleRequest(self, requestMessage):
        """
        Dispatch the request message to the appropriate handler for the method

        :param requestMessage: request for device
        :return: `ResponseMessage`
        """
        method = requestMessage.getMethod().upper()
        responseStatus = StatusCode.BAD_REQUEST
        if "POST" == method:
            methodOverride = requestMessage.getMethodOverride()
            if "PATCH" == methodOverride:
                responseStatus = self.handlePatch(requestMessage)
            else:
                responseStatus = self.handlePost(requestMessage)

        elif "PUT" == method:
            responseStatus = self.handlePut(requestMessage)

        elif "PATCH" == method:
            responseStatus = self.handlePatch(requestMessage)
        #elif "GET" == method:
        #    For a GET, the server should return the value
        else:
            self._getLogger().critical("unexpected method: " + method)

        return ResponseMessage(requestMessage=requestMessage, statusCode=responseStatus)

    def handlePatch(self, requestMessage):
        """
        A 'PATCH' request is a multi attribute change request
        :param requestMessage: iotcs.shared.message.RequestMessage identifies the attributes to be changed
        :return: iotcs.shared.message.StatusCode
        """
        #NamedValue<?>
        root = None
        #VirtualDeviceBase.NamedValueImpl<?>
        last = None
        try:
            # bytearray
            rawData = requestMessage.getBody()
            jsonObject = json.loads(rawData)

            if self.__base.getOnChangeCallback() == None:
                # If there is no device level onChange callback then all
                # attributes in the message must have handlers.
                for key in jsonObject.keys():
                    attributeName = key
                    jsonValue = jsonObject[attributeName]
                    # Throws IllegalArgumentException if attribute is not
                    # in model.
                    #VirtualDeviceAttributeImpl
                    attribute = self.getAttribute(attributeName)
                    if attribute.getOnChange() == None:
                        self._getLogger().info("No handler for: '" +
                                         requestMessage.getMethod().upper() + " " +
                                         requestMessage.getURL())
                        return StatusCode.NOT_FOUND

            for key in jsonObject.keys():
                attributeName = key
                jsonValue = jsonObject[attributeName]

                # Throws IllegalArgumentException if attribute is not in model
                #VirtualDeviceAttributeImpl
                attribute = self.getAttribute(attributeName)

                #DeviceModelAttribute
                dma = attribute.getDeviceModelAttribute()
                newValue = self._getValue(dma.getType(), jsonValue,
                                          attributeName)

                # Throws IllegalArgumentException if newValue isn't right type
                attribute.validate(dma, newValue)

                oldValue = attribute.get()
                # Do not update the value, or call onChange, if the value
                # hasn't changed.
                if (oldValue == newValue if oldValue is not None else
                        newValue == None):
                    continue

                #VirtualDeviceBase.NamedValueImpl<?>
                nameValue = VirtualDeviceBase.NamedValueImpl(attributeName,
                                                             newValue)

                if attribute.getOnChange() != None:
                    attribute.getOnChange().onChange(
                        VirtualDeviceBase.ChangeEvent(self, nameValue))

                if last != None:
                    last.setNext(nameValue)
                    last = nameValue
                else:
                    root = last = nameValue

        except Exception as e:
            # The onChange method may throw exceptions. Since the CL has no
            # knowledge of what these might be, Exception is caught here.
            #
            # Possible exceptions from CL internals are JSONException,
            # IllegalArgumentException, NullPointerException, and
            # ClassCastException.
            self._getLogger().critical(str(e))
            return StatusCode.BAD_REQUEST

        # root == None just means that no values were updated,
        # which can happen if the new values equal the old values.
        if root == None:
            return StatusCode.ACCEPTED

        # patch is handled by change callback set on VirtualDevice
        #ChangeCallback<VirtualDevice>
        changeCallback = self.__base.getOnChangeCallback()
        try:
            if changeCallback != None:
                changeCallback.onChange(VirtualDeviceBase.ChangeEvent(self, root))

            #List<DataItem<?>>
            dataItems = [] # Was set([])

            namedValue = root
            while namedValue is not None:
                attributeName = namedValue.getName()
                #VirtualDeviceAttributeImpl
                attribute = self.getAttribute(attributeName)
                newValue = namedValue.getValue()

                # Call update, not set! Update sets the value but does not
                # cause a DataMessage to be sent. We want to send one
                # DataMessage with all the updated data items.
                #
                # There should be no issues with calling update here. All the
                # checking was done while building the event chain, and
                # newValue should be the converted data from the
                # RequestMessage.
                attribute.update(newValue)

                #DeviceModelAttribute
                dma = attribute.getDeviceModelAttribute()
                dataItems.append(DataItem(attributeName, newValue))
                namedValue = namedValue.next()

            # format, source, data
            dataMessage = DataMessage(format=self.getDeviceModel().getURN() + ":attributes",
                                      items=dataItems,
                                      source=self.getEndpointId())

            # TODO: why not use "local message dispatcher self.__messageDispatcher ?
            messageDispatcher = MessageDispatcher.getMessageDispatcher(self.__directlyConnectedDevice)
            messageDispatcher.queue([dataMessage])

            return StatusCode.ACCEPTED

        except Exception as e:
            # The onChange method may throw exceptions. Since the CL has no
            # knowledge of what these might be, Exception is caught here.
            self._getLogger().critical(str(e))
            return StatusCode.INTERNAL_SERVER_ERROR

    # POST is for actions
    def handlePost(self, requestMessage):
        path = requestMessage.getURL()
        dmURN = "deviceModels/" + self.getDeviceModel().getURN() + "/actions"
        if not path.startswith(dmURN):
            return StatusCode.NOT_FOUND

        actionname = None
        index = path.find("/actions/")
        actionlen = len("/actions/")
        status = StatusCode.NOT_FOUND
        if index > 0:
            actionname = path[index + actionlen:]

        body = requestMessage.getBody()
        jsonObject = None
        try:
            jsonObject = json.loads(body)
        except Exception as e:
            raise NetworkException(f'VirtualDeviceImpl.handlePost(): '
                f'json.loads(body) raised exception {e}')

        argumentvalues = dict()
        for (k,v) in jsonObject.items():
            argumentvalues[k] = v

        statusCode = self.callImpl(actionname, argumentvalues)
        return statusCode

    # PUT is for attributes or for resources
    def handlePut(self, requestMessage):
        try:
            path = requestMessage.getURL()
            dmURN = "deviceModels/" + self.getDeviceModel().getURN() + "/attributes"
            if not path.startswith(dmURN):
                return StatusCode.NOT_FOUND

            attributeName = None
            index = path.find("/attributes/")
            attributeLen = len("/attributes/")
            status = StatusCode.NOT_FOUND
            if index > 0:
                attributename = path[index + attributeLen:]

            #VirtualDeviceAttributeBase<VirtualDevice, Object>
            virtualDeviceAttribute = self.getAttribute(attributeName)
            #DeviceModelAttribute
            deviceModelAttribute = virtualDeviceAttribute.getDeviceModelAttribute()
            #DeviceModelAttribute.Type
            attributeType = deviceModelAttribute.getType()

            # bytearray
            data = requestMessage.getBody()
            jsonObject = json.loads(data)
            newValue = None
            if "value" in jsonObject:
                newValue = jsonObject["value"]
            newValue = self._getValue(attributeType, newValue, "value")
            oldValue = virtualDeviceAttribute.get()
            # Do not update the value, or call onChange, if the value hasn't changed.
            if (oldValue == newValue if oldValue is not None else newValue == None):
                return StatusCode.ACCEPTED

            namedValue = VirtualDeviceBase.NamedValueImpl(attributeName, newValue)
            attrOnChangeCalled = False

            if virtualDeviceAttribute.getOnChange() is not None:
                virtualDeviceAttribute.getOnChange().onChange(VirtualDeviceBase.ChangeEvent(self, namedValue))
                attrOnChangeCalled = True

            if self.__base.getOnChangeCallback() is not None:
                self.__base.getOnChangeCallback().onChange(VirtualDeviceBase.ChangeEvent(self, namedValue))
                attrOnChangeCalled = True

            if attrOnChangeCalled:
                # Call set, not update! Update sets the value but does not
                # cause a DataMessage to be sent. Since the put is for a
                # single attribute only, we want the DataMessage to be sent.
                # This is different than handlePatch.
                virtualDeviceAttribute.set(newValue)
                return StatusCode.ACCEPTED
            else:
                self._getLogger().info("No handler for: '" +
                                 requestMessage.getMethod().upper() + " " +
                                 requestMessage.getURL())
                return StatusCode.NOT_FOUND
        except Exception as e:
            # The call method may throw exceptions. Since the CL has no
            # knowledge of what these might be, Exception is caught here.
            #
            # Possible exceptions from CL internals are JSONException,
            # IllegalArgumentException, NullPointerException, and
            # ClassCastException.
            #
            self._getLogger().debug(f'VirtualDeviceImpl.handlePut(): {e}')
            return StatusCode.BAD_REQUEST

    # Why was this static ?
    # DeviceModelAction
    @staticmethod
    def getDeviceModelAction(deviceModel, actionName):
        if deviceModel is None:
            return None

        #Map<String, DeviceModelAction>
        actions = deviceModel.getDeviceModelActions()
        if actions is None or len(actions) == 0 or actionName not in actions:
            return None

        #DeviceModelAction
        action = actions[actionName]
        if action is not None:
            return action

    def setCallable(self, actionName, callable):
        deviceModelAction = \
            VirtualDeviceImpl.getDeviceModelAction(self.__base.getDeviceModel(),
                                                   actionName)
        if deviceModelAction is None:
            raise ArgumentException(f'VirtualDeviceImpl.setCallable(): '
                f'{actionName} not found in model {self.__base.getDeviceModel().getURN()}')

        with self.__ACTIONMAP_LOCK:
            # if callable is None remove any existing callback
            if callable is None:
                if actionName in self.__callableMap:
                    del(self.__callableMap[actionName])
            else:
                self.__callableMap[actionName] = callable

    def setOnAction(self, actionCallback, actionName=None):
        if actionName is not None:
            deviceModelAction = \
                VirtualDeviceImpl.getDeviceModelAction(self.__base.getDeviceModel(),
                                                       actionName)
            if deviceModelAction is None:
                raise ArgumentException(f'VirtualDeviceImpl.setCallable(): '
                    f'{actionName} not found in model {self.__base.getDeviceModel().getURN()}')
        else:
            actionName = "*"

        with self.__ACTIONMAP_LOCK:
            if actionCallback is None and actionName in self.__actionCallbackMap:
                del(self.__actionCallbackMap[actionName])
            else:
                self.__actionCallbackMap[actionName] = actionCallback

    def createAlert(self, format):
        return AlertImpl(self, format)

    def createData(self, format):
        return DataImpl(self, format)

    # called from VirtualDeviceAttributeImpl, protected in JCL
    def processOnChange(self, virtualDeviceAttribute, newValue):

        storageObjectContainer = []
        # JCL implements processOnChange(builder,..) by modifying the
        # build in the call. Instead create and return a DataMessage
        dataMessage = None
        try:
            dataMessage = self.__processOnChange(virtualDeviceAttribute,
                                                 newValue,
                                                 storageObjectContainer)
        except Exception as e:
            self._getLogger().debug(str(e))
            return

        try:
            so = storageObjectContainer[0] \
                if len(storageObjectContainer) != 0 else None
            self.queueMessage(dataMessage, [so])
        except Exception as e:
            # MessageDispatcher queue is full.
            attributes = set([])
            attributes.add(virtualDeviceAttribute)
            self.notifyException(attributes, e)

    # Also called from AlertImpl and DataImpl
    # This is also the DeviceAnalog overload.
    def queueMessage(self, message, storageObjects=None):
        """

        :param message:
        :param storageObjects: list of :class:`.client.device.StorageObectImpl`
        :return:
        """
        pair = Pair(message, storageObjects)
        pairs = [pair]
        deviceModelURN = self.getDeviceModel().getURN()
        devicePolicy = \
            self.__devicePolicyManager.getPolicy(deviceModelURN,
                                                 self.getEndpointId())
        # Handling of device model level policies here...
        if devicePolicy is not None and \
                devicePolicy.getPipeline(DevicePolicy.ALL_ATTRIBUTES) is not None:
            # Some policies are negated by an alert of a given severity
            # (batching policies, for example
            alertSeverity = None
            if message is AlertMessage:
                alertSeverity = message.getSeverity()

            pipeline = devicePolicy.getPipeline(DevicePolicy.ALL_ATTRIBUTES)
            if len(pipeline) > 1:
                # TODO: handle more than one function in a pipeline for
                # TODO: all attributes
                # print out a warning message about too many functions
                # for all attributes pipeline
                first = True
                logmessage = ""
                for plfunction in pipeline:
                    id = plfunction.getId()
                    parameters = plfunction.getParameters()
                    deviceFunction = DeviceFunction.getDeviceFunction(id)
                    if first:
                        logmessage = f'queueMessage(): ' \
                             f'Only one function allowed for all-attribute pipeline.' \
                             f'\n\tApplying: {deviceFunction.getDetails(parameters)}'
                        first = False
                    else:
                        logmessage = f'{logmessage}' \
                             f'\n\tIgnoring: {deviceFunction.getDetails(parameters)}'

                self._getLogger().warning(logmessage)

            pipelineData = self.getPipelineData(DevicePolicy.ALL_ATTRIBUTES)

            # TODO: handle more than one function in a pipeline for all-attributes
            # Note: A warning is printed out, several lines above this, if there
            # is more than one function for an all-attributes policy.
            # Leaving original line here, commented-out, for future reference.
            # for (int index = 0,
            #      maxIndex = pipeline.size(); index < maxIndex; index++)
            index = 0
            maxIndex = 1
            while index < maxIndex:
                plfunction = pipeline[index]
                id = plfunction.getId()
                parameters = plfunction.getParameters()
                deviceFunction = DeviceFunction.getDeviceFunction(id)
                if deviceFunction is None:
                    index += 1
                    continue

                # if "batchByTime".equals(deviceFunction.getId()):
                #   index += 1
                #   continue

                alertOverridesPolicy = False
                if alertSeverity is not None:
                    configuredSeverity = AlertMessage.Severity.CRITICAL
                    criterion = None
                    if "alertSeverity" in parameters:
                        criterion = str(parameters["alertSeverity"])
                    if criterion is not None:
                        try:
                            criterion = criterion.upper()
                            configuredSeverity = AlertMessage.Severity[criterion]
                        except Exception as e:
                            configuredSeverity = AlertMessage.Severity.CRITICAL
                    alertOverridesPolicy = \
                        configuredSeverity.value <= alertSeverity.value
                else:
                    alertOverridesPolicy = False

                functionData = None
                if index < len(pipelineData):
                    functionData = pipelineData[index]
                else:
                    functionData = dict()
                    pipelineData.append(functionData)

                window = DeviceFunction.getWindow(parameters)
                if deviceFunction.apply(self, None, parameters, functionData,
                                        pair) or alertOverridesPolicy:
                    if window > 0:
                        slide = DeviceFunction.getSlide(parameters, window)
                        key = ScheduledPolicyData.Key(window, slide)
                        timeZero = TimeManager.currentTimeMillis()
                        if key in self.__scheduledPolicies:
                            scheduledPolicyData = self.__scheduledPolicies[key]
                            updatedAttributes = []
                            scheduledPolicyData.processExpiredFunction(self,
                                updatedAttributes, timeZero)
                            if len(updatedAttributes) > 0:
                                self.updateFields(updatedAttributes)
                            return
                    pairs = deviceFunction.get(self, None, parameters,
                                               functionData)
                    # FINE
                    self._getLogger().debug(f'queueMessage(): '
                        f'VirtualDevice: {self.getEndpointId()} '
                        f'dispatching {len(pairs)} messages per policy '
                        f'"{deviceFunction.getDetails(parameters)}"')
                    index += 1
                else:
                    return

        try:
            # TODO: Why not use local __messagedispatcher ?
            messageDispatcher = MessageDispatcher.getMessageDispatcher(
                    self.__directlyConnectedDevice)
            messages = []
            for pair in pairs:
                messages.append(pair.getKey())
                solist = pair.getValue()
                if solist is not None:
                    for so in solist:
                        if so is None:
                            continue
                        messageDispatcher.addStorageObjectDependency(
                            so, message.getClientId())
                        so.sync()

            messageDispatcher.queue(messages)
        except Exception as e:
            # raise(e) if ArrayStorageException
            # SEVERE
            self._getLogger().critical(f'queueMessage(): {e}')

    # called from updateFields, private in JCL
    #List<Pair<VirtualDeviceAttributeBase<VirtualDevice, Object>, Object>>
    def _processOnChange(self, updatedAttributes):

        if len(updatedAttributes) == 0:
            return

        #Set<VirtualDeviceAttributeBase<VirtualDevice, Object>>
        keySet = set()
        storageObjectContainer = []

        #Pair<VirtualDeviceAttributeBase<VirtualDevice, Object>, Object>
        dataMessage = None
        for entry in updatedAttributes:

            #VirtualDeviceAttributeBase<VirtualDevice, Object>
            attribute = entry.getKey()
            keySet.add(attribute)
            newValue = entry.getValue()

            try:
                dataMessage = self.__processOnChange(
                    attribute, newValue,
                    storageObjectContainer=storageObjectContainer,
                    dataMessage=dataMessage)
            except Exception as e:
                self._getLogger().critical(str(e))
                return

        # There is an assumption that there will only ever be one
        # storage object attribute among list of updated attributes
        try:
            so = storageObjectContainer[0] \
                if len(storageObjectContainer) != 0 else None
            self.queueMessage(dataMessage, [so])
        except Exception as e:
            # MessageDispatcher queue is full.
            self.notifyException(keySet, e)
            self._getLogger().critical(str(e))

    def handleStorageObjectStateChange(self, storageObject):
        """

        :param storageObject: :class:`.iotcs.client.device.StorageObjectImpl`
        """
        messageDispatcherImpl = MessageDispatcher.getMessageDispatcher(
                self.__directlyConnectedDevice)
        messageDispatcherImpl.removeStorageObjectDependency(storageObject)

    def __str__(self):
        return self.__base.__str__()

    def getValue(self, attributeType, payload, name):
        """
        :param attributeType: `DeviceModelAttribute.Type`
        :param payload: `bytearray`
        :raises ValueError:
        """
        jsonObject = None
        # TODO: Does this ever throw exception ?
        try:
            jsonObject = json.loads(payload)
        except Exception as e:
            raise(ValueError(str(e)))

        if "value" in jsonObject and jsonObject["value"] is not None:
            value = jsonObject["value"]
            return self._getValue(attributeType, value, name)
        else:
            raise(ValueError("bad payload " + str(jsonObject)))

    def _getValue(self, attributeType, jsonValue, name):
        """
        :param attributeType: DeviceModelAttribute.Type
        :param jsonValue: Object
        :param name: String
        :raises: :class:`.ArgumentException` if `jsonValue` is not an
            `attributeType`
        """

        if jsonValue is None:
            return None

        data = None
        castvalue = DeviceModelAttribute.Type.cast(attributeType, jsonValue)
        if attributeType == DeviceModelAttribute.Type.URI:
            storageObjectImpl = None
            try:
                if StorageConnectionBase.isStorageCloudURI(str(jsonValue)):
                    try:
                        #com.oracle.iot.client.StorageObject
                        delegate = \
                            self.__directlyConnectedDevice.createStorageObject(
                                jsonValue)
                        storageObjectImpl = \
                            StorageObjectImpl(self.__directlyConnectedDevice,
                                              delegate)
                        storageObjectImpl.setSyncEventInfo(self, name)
                    except Exception as e:
                        # If SCS object in inaccessable, log it and create
                        # ExternalObject data instead.
                        self._getLogger().warning(
                            f'Storage CS object access failed: {str(e)}')

                data = ExternalObject(str(jsonValue)) \
                    if storageObjectImpl is None else storageObjectImpl
            except Exception as e:
                raise ArgumentException(
                    f'Cannot get value for attribute {name} {str(e)}')

        return castvalue



    class ErrorCallbackBridge(MessageDispatcher.ErrorCallback):
        """
        The default implementation of `iotcs.device.MessageDispatcher.ErrorCallback` that is called when
        there are errors sending messages to the IOT Cloud Service. Used for all virtual device instances
        """

        #Map<String, WeakReference<VirtualDeviceImpl>>
        __slots__ = (
            'deviceSet' # The set of virtual devices to dispatch errors to.
        )

        def __init__(self):
            super(VirtualDeviceImpl.ErrorCallbackBridge, self).__init__()
            self.deviceSet = dict()

        def add(self, virtualDevice):
            """
            Add `virtualDevice` to the list of devices to dispatch errors to.
            :param virtualDevice: `iotcs.device.impl.VirtualDeviceImpl`
            """
            removelist = []
            if len(self.deviceSet) > 0:
                # TODO: this will be a performance problem if the map is very large
                #Map.Entry<String,WeakReference<VirtualDeviceImpl>>>
                for k,v in self.deviceSet.items():
                    removelist.append(k)

            for k in removelist:
                del self.deviceSet[k]

            # JCL assert
            # TODO: Weak references
            if virtualDevice.getEndpointId() != None:
                self.deviceSet[virtualDevice.getEndpointId()] = virtualDevice

        def failed(self, messages, exception):
            """
            Called when `iotcs.device.MessageDispatcher.send` fails to send messages.
            :param messages: The list of messages that could not be sent
            :param exception: The `Exception` that causes the send failure
            """
            #Map<String,List<Message>>
            collatedMessages = dict()
            # Collate the messages by source
            for message in messages:
                source = message.getSource()
                list = None
                # TODO: Does python hand back a refernce or value from a dict access ?
                if source not in collatedMessages or collatedMessages[source] is None:
                    list = []
                    collatedMessages[source] = list
                else:
                    list = collatedMessages[source]

                list.append(message)

            for source,list in collatedMessages.items():
                virtualDevice = None
                if source in self.deviceSet and self.deviceSet[source] is not None:
                    virtualDevice = self.deviceSet[source]
                else:
                    self._getLogger().debug(f'failed(): DEBUG source= '
                        f'{source} deviceSet = {str(self.deviceSet)}')

                # VirtualDeviceImpl
                if virtualDevice is not None:
                    self.invokeErrorCallback(virtualDevice, list, exception)

        def invokeErrorCallback(self, virtualDevice, messages, exception):
            """
            :param virtualDevice: VirtualDeviceImpl
            :param messages: List<Message>
            :param exception: Exception
            """

            #VirtualDeviceBase.NamedValueImpl<?>
            values = None
            for message in messages:
                #List<DataItem<?>>
                dataItems = None
                if isinstance(message,(DataMessage,)):
                    # TODO: How to cast ? is it necessary ?
                    dataMessage = message
                    dataItems = dataMessage.getDataItems()
                #TODO: alert needs error callback
                #elif isinstance(message,(AlertMessage,)):
                #      alertMessage = message
                #      dataItems = alertMessage.getDataItems()

                if dataItems is not None:
                    #VirtualDeviceBase.NamedValueImpl<?>
                    last = None
                    for dataItem in dataItems:
                        #VirtualDeviceBase.NamedValueImpl<?>
                        namedValue = VirtualDeviceBase.NamedValueImpl(dataItem.getKey(), dataItem.getValue())
                        try:
                            #VirtualDeviceAttributeBase<VirtualDevice, ?>
                            attribute = virtualDevice.getAttribute(dataItem.getKey())
                            if attribute.getOnError() is not None:
                                attribute.getOnError().onError(VirtualDeviceBase.ErrorEvent(virtualDevice,
                                                                                            namedValue,
                                                                                            str(exception)))
                        except Exception as e:
                            self._getLogger().debug(str(e))

                        if last is not None:
                            last.setNext(namedValue)
                            last = namedValue
                        else:
                            values = last = namedValue

            if virtualDevice.getErrorCallback() is not None:
                try:
                    errorEvent = VirtualDeviceBase.ErrorEvent(virtualDevice, values, str(exception))
                    virtualDevice.getErrorCallback().onError(errorEvent)
                except Exception as e:
                    # onError could throw an exception
                    self._getLogger().debug(str(e))

        def _getLogger(self):
            return logging.getLogger("iotcs.client.device.VirtualDeviceImpl.ErrorCallbackBridge")

    # NOTE: JCL passes DataMessage.Builder, instead return DataMessage
    # private in JCL, we have 3 processOnCHange, make this one dunderscore
    # TODO: Implement media storage
    # called from _processOnChange(List<Pair>), processOnChange(VirtualDeviceAttribute)
    def __processOnChange(self, attribute, newValue,
                          storageObjectContainer=None,
                          dataMessage=None):
        """
        :param attribute: VirtualDeviceAttributeBase<VirtualDevice, Object>
        :param newValue: Object
        :param storageObjectContainer: a list to return a
            :class:`.iotcs.client.device.StorageObjectImpl` instance
        :raises: :class:`.ArguemntException` if `newValue` does not match
           the `attribute`'s :class:`.DeviceModelAttribute.Type`
        :return: DataMessage
        """
        #DeviceModelAttribute
        deviceModelAttribute = attribute.getDeviceModelAttribute()
        attributeName = deviceModelAttribute.getName()
        if dataMessage is None:
            dataMessage = DataMessage(
                format=self.__base.getDeviceModel().getURN() + ":attributes",
                source=self.__base.getEndpointId())

        dataItem = None
        castvalue = newValue
        dmaType = deviceModelAttribute.getType()
        try:
            if dmaType == DeviceModelAttribute.Type.URI:
                # newValue MUST be a StorageObjectImpl or ExternalObject
                # or will throw. JCL does not check, and forces a cast
                # to ExternalObject
                if isinstance(newValue, (StorageObjectImpl,ExternalObject,)):
                    if isinstance(newValue, StorageObjectImpl):
                        # just for clarity
                        storageObjectImpl = newValue
                        if storageObjectImpl.getSyncStatus() == \
                                    StorageObject.SyncStatus.NOT_IN_SYNC or \
                                storageObjectImpl.getSyncStatus() == \
                                    StorageObject.SyncStatus.SYNC_PENDING:
                            storageObjectContainer.append(storageObjectImpl)
                        storageObjectImpl.setSyncEventInfo(self, attributeName)

                    # newValue isa ExternalObject
                    castvalue = newValue.getURI()
                else:
                    raise ArgumentException(f'invalid type {newValue.__name__} '
                        f'for URI attribute type')
            else:
                castvalue = DeviceModelAttribute.Type.cast(dmaType, newValue)
        except Exception as e:
            self._getLogger().critical(f'{e}')
            raise e

        dataItem = DataItem(attributeName, castvalue)
        dataMessage.add(dataItem)
        return dataMessage

    class errorNotifyThread(Thread):
        __slots__ = (
            'cb',
            'ev',
            'msg'
        )
        def __init__(self, errorCallback, event, msg):
            super(VirtualDeviceImpl.errorNotifyThread, self).__init__(
                name="iotcs.device.impl.errorNotifyThread",
                daemon=True)
            self.cb = errorCallback
            self.ev = event
            self.msg = msg

        def run(self):
            try:
                self.cb(self.ev)
            except:
                self._getLogger().warning("onError raised: " + self.msg)

        def _getLogger(self):
            return logging.getLogger("iotcs.client.device.VirtualDeviceImpl.errorNotifyThread")

    def notifyException(self, attributes, e):

        #AbstractVirtualDevice.ErrorCallback
        errorCallback = self.getErrorCallback()
        if errorCallback is None:
            return

        #VirtualDeviceBase.NamedValueImpl<Object>
        head = None
        tail = None
        for attribute in attributes:
            name = attribute.getDeviceModelAttribute().getName()
            value = attribute.get()
            #VirtualDeviceBase.NamedValueImpl<Object>
            next = VirtualDeviceBase.NamedValueImpl(name, value)
            if head is not None:
                tail.setNext(next)
                tail = next
            else:
                head = tail = next

        #ErrorEvent<VirtualDevice>
        errorEvent = VirtualDeviceBase.ErrorEvent(self, head, str(e))
        msg = str(e)

        # TODO:  Use thread pool
        #errorEventDispatcher.submit(new Runnable() {
        #    @Override
        #    public void run() {
        #        try {
        #            errorCallback.onError(errorEvent);
        #        } catch (Exception ignored) {
        #            self._getLogger().warn("onError threw: " + msg);
        #        }
        #    }
        #});
        VirtualDeviceImpl.errorNotifyThread(errorCallback, errorEvent, msg).run()

    def getPipelineData(self, attribute):
        """
        Return device policy pipeline data for `attribute` else `[]`
        :param attribute: device model attribute name
        :return: device policy pipeline data for `attribute` else `[]`
        """
        devicePolicy = \
            self.__devicePolicyManager.getPolicy(
                self.getDeviceModel().getURN(), self.getEndpointId())

        if devicePolicy is None:
            return []

        pipeline = devicePolicy.getPipeline(attribute)
        if pipeline is None or len(pipeline) == 0:
            return []

        # TODO: Locks for pipelinedatacache
        pipelineData = None
        if attribute in self.__pipelineDataCache:
            pipelineData = self.__pipelineDataCache[attribute]
        if pipelineData is None:
            pipelineData = []
        self.__pipelineDataCache[attribute] = pipelineData


        # create missing function maps
        if len(pipelineData) < len(pipeline):
            # create missing function data maps
            for i in range(len(pipelineData),len(pipeline)):
                pipelineData.append(dict())

        return pipelineData

    def _getLogger(self):
        return logging.getLogger("iotcs.device.impl.VirtualDeviceImpl")

    # Device policy methods
    def addScheduledPolicy(self, window, slide, timeZero, attributeName,
                           pipelineIndex):
        """
        Add a :class:`.ScheduledPolicyData` to :attr:`.__scheduledPolicies`.

        Start the :class;`.TimedPolicyThread` if it is not running.
        :param window:
        :param slide:
        :param timeZero:
        :param attributeName:
        :param pipelineIndex:
        :return:
        """
        with self.__scheduledPoliciesLock:
            key = ScheduledPolicyData.Key(window, slide)
            scheduledPolicyData = None
            if key in self.__scheduledPolicies:
                scheduledPolicyData = self.__scheduledPolicies[key]
            if scheduledPolicyData is None:
                scheduledPolicyData = ScheduledPolicyData(window, slide,
                                                          timeZero)
                self.__scheduledPolicies[key] = scheduledPolicyData
                self.__timedPolicyThread.addTimedPolicyData(scheduledPolicyData)
                if not self.__timedPolicyThread.is_alive() and \
                        not self.__timedPolicyThread.isCancelled():
                    self.__timedPolicyThread.start()

            scheduledPolicyData.addAttribute(attributeName, pipelineIndex)

    def removeScheduledPolicy(self, slide, attributeName, pipelineIndex,
                              window):
        """
        Remove a :class:`.ScheduledPoiicyDate` from :attr:`.__scheduledPolicies`.

        :param slide:
        :param attributeName:
        :param pipelineIndex:
        :param window:
        :return:
        """
        with self.__scheduledPoliciesLock:
            key = ScheduledPolicyData.Key(window, slide)
            scheduledPolicyData = None
            if key in self.__scheduledPolicies:
                scheduledPolicyData = self.__scheduledPolicies[key]
            if scheduledPolicyData is not None:
                scheduledPolicyData.removeAttribute(attributeName,
                                                    pipelineIndex)
                if scheduledPolicyData.isEmpty():
                    del(self.__scheduledPolicies[key])
                    self.__timedPolicyThread.removeTimedPolicyData(
                        scheduledPolicyData)

    #  DevicePolicyManager.ChangeListener interface
    def policyAssigned(self, policy, assignedDevices):
        """

        :param policy:
        :param assignedDevices:
        :return:
        """
        if assignedDevices == None or \
                self.getEndpointId() not in assignedDevices:
            return
        # FINE
        self._getLogger().debug(f'policyAssigned(): {self.getEndpointId()} : '
            f'Policy assigned : {policy.getId()}')
        timeZero = TimeManager.currentTimeMillis()
        for attributeName, newPipeline in policy.getPipelines().items():
            self._policyAssigned(attributeName, newPipeline, timeZero)


    def _policyAssigned(self, attributeName, newPipeline, timeZero):
        """

        :param attributeName:
        :param newPipeline:
        :param timeZero:
        :return:
        """
        if newPipeline is not None and len(newPipeline) > 0:
            index = 0
            indexMax = len(newPipeline)
            while index < indexMax:
                plfunction = newPipeline[index]
                id = plfunction.getId()
                parameters = plfunction.getParameters()
                window = DeviceFunction.getWindow(parameters)
                if window > -1 and not ("eliminateDuplicates" == id or
                                        "detectDuplicates" == id):
                    slide = DeviceFunction.getSlide(parameters, window)
                    self.addScheduledPolicy(window, slide, timeZero,
                                            attributeName, index)
                #  If the first policy in the chain is a computed metric,
                #  see if it refers to other attributes.
                if index == 0 and "computedMetric" == id:
                    formula = None
                    if "formula" in parameters:
                        formula = parameters["formula"]
                    # formula could be None, handled in getAttributeReference
                    triggerAttributes = set()
                    references = Formula.getAttributeReferences(formula)
                    for attr in references:
                        if attr != attributeName:
                            triggerAttributes.add(attr)

                    if len(triggerAttributes) > 0:
                        with self.__computedMetricTriggerMapLock:
                            self.__computedMetricTriggerMap.append(
                                Pair(triggerAttributes, attributeName))
                index += 1

    def policyUnassigned(self, devicePolicy, unassignedDevices):
        """

        :param devicePolicy:
        :param unassignedDevices:
        :return:
        """
        if unassignedDevices is None or \
                self.getEndpointId() not in unassignedDevices:
            return
        # FINE
        self._getLogger().debug(f'policyUnassigned(): {self.getEndpointId()} : '
            f'Policy un-assigned : {devicePolicy.getId()}')
        updatedAttributes = []
        for attributeName, oldPipeline in devicePolicy.getPipelines().items():
            self._policyUnassigned(updatedAttributes, attributeName,
                                  oldPipeline)
        if len(updatedAttributes) > 0:
            #
            #  Call updateFields to ensure the computed metrics get run,
            #  and will put all attributes into one data message.
            #
            self.updateFields(updatedAttributes)

    def _policyUnassigned(self, updatedAttributes, attributeName,
                            oldPipeline):
        """

        :param updatedAttributes:
        :param attributeName:
        :param oldPipeline:
        :return:
        """
        if oldPipeline is None or len(oldPipeline) == 0:
            return

        #
        #  The order in which the oldPipeline is finalized is important.
        #  First, remove any scheduled policies so they don't get executed. Any
        #  pending data will be committed in the next step.
        #  Second, commit any "in process" values. This may cause a computedMetric
        #  to be triggered.
        #  Third, remove any computed metric triggers.
        #  Lastly, remove any data for this pipeline from the policy data cache
        #
        index = 0
        indexMax = len(oldPipeline)
        while index < indexMax:
            plfunction = oldPipeline[index]
            id = plfunction.getId()
            parameters = plfunction.getParameters()
            window = DeviceFunction.getWindow(parameters)
            if window > -1 and not ("eliminateDuplicates" == id or
                                    "detectDuplicates" == id):
                slide = DeviceFunction.getSlide(parameters, window)
                self.removeScheduledPolicy(slide, attributeName, index, window)
            index += 1

        #  commit values from old pipeline
        pipelineData = self.getPipelineData(attributeName)
        if pipelineData is not None and len(pipelineData) > 0:
            if  DevicePolicy.ALL_ATTRIBUTES != attributeName:
                self.processExpiredFunction(updatedAttributes, attributeName,
                                            oldPipeline, pipelineData)
            else:
                self._processExpiredFunction(oldPipeline, pipelineData)
        if attributeName is not None:
            #  remove this attribute from the computedMetricTriggerMap
            with self.__computedMetricTriggerMapLock:
                index = len(self.__computedMetricTriggerMap) - 1
                while 0 <= index:
                    pair = self.__computedMetricTriggerMap[index]
                    if attributeName == pair.getValue():
                        self.__computedMetricTriggerMap.pop(index)
                    index -= 1
        # TODO: Need Locking
        #  remove data from cache
        if attributeName in self.__pipelineDataCache:
            del(self.__pipelineDataCache[attributeName])


    # TODO: See if we can consolidate processExpiredFunction and
    # TODO: _processExpiredFunction

    def processExpiredFunction(self, updatedAttributes, attributeName,
                               pipeline, pipelineData):
        """
        Invoke policy functions on `attributeName` as necessary.

        Routine for handling invocation of a policy function when the window's
        slide expires. This routine gets the value of the function, and then
        processes the remaining functions in the pipeline (if any).

        :param updatedAttributes:
        :param attributeName:
        :param pipeline:
        :param pipelineData:
        :return:
        """
        if pipeline is None or len(pipeline) == 0:
            return
        try:
            attribute = self.getAttribute(attributeName)
            deviceModelAttribute = attribute.getDeviceModelAttribute()
            plfunction = pipeline[0]
            functionId = plfunction.getId()
            config = plfunction.getParameters()
            data = pipelineData[0]
            deviceFunction = DeviceFunction.getDeviceFunction(functionId)
            if deviceFunction is None:
                # SEVERE
                self._getLogger().critical(f'processExpiredFunction(): '
                    f'device function {functionId} not found')
                return
            value = deviceFunction.get(self, attributeName, config, data)
            if value is not None and len(pipeline) > 1:
                #  process remaining policies in the pipeline
                value = self.offer0(deviceModelAttribute, value,
                    pipeline[1: len(pipeline)],
                    pipelineData[1: len(pipelineData)])
            if value is not None:
                policyValue = DeviceModelAttribute.Type.cast(
                    attribute.getDeviceModelAttribute().getType(), value,
                    convert=True)
                if policyValue is not None:
                    # FINE
                    self._getLogger().debug(f'processExpiredFunction(): '
                        f'{self.getEndpointId()} : Set  '
                        f'"{attributeName}"="{policyValue}"')
                    updatedAttributes.append(Pair(attribute, policyValue))
        except Exception as e:
            self._getLogger().warning(f'processExpiredFunction(): '
            f'processing {attributeName}, caught exception: {e}')

    def _processExpiredFunction(self, pipeline, pipelineData):
        """
        Process :attr:`.DevicePolicy.ALL_ATTRIBUTES` functions.

        Routine for handling invocation of a device model policy when the
        window's slide expires. This routine gets the value of the function,
        and then processes the remaining functions in the pipeline (if any).

        :param pipeline:
        :param pipelineData:
        :return:
        """
        if pipeline is None or len(pipeline) == 0:
            return

        if len(pipeline) > 1:
            # TODO: handle more than one function in a pipeline for
            # TODO: all attributes
            # print out a warning message about too many functions
            # for all attributes pipeline
            first = True
            logmessage = ""
            for plfunction in pipeline:
                id = plfunction.getId()
                parameters = plfunction.getParameters()
                deviceFunction = DeviceFunction.getDeviceFunction(id)
                if first:
                    logmessage = f'queueMessage(): ' \
                        f'Only one function allowed for '\
                        f'all-attribute pipeline.' \
                        f'\n\tApplying: '\
                        f'{deviceFunction.getDetails(parameters)}'
                    first = False
                else:
                    logmessage = f'{logmessage}\n\tIgnoring: '\
                        f'{deviceFunction.getDetails(parameters)}'

        try:
            plfunction = pipeline[0]
            functionId = plfunction.getId()
            config = plfunction.getParameters()
            data = pipelineData[0]
            deviceFunction = DeviceFunction.getDeviceFunction(functionId)
            if deviceFunction is None:
                # SEVERE
                self._getLogger().critical(f'_processExpiredFunction() :'
                    f'device function {functionId} not found')
                return
            value = deviceFunction.get(self, None, config, data)

            # TODO: handle more than one function in a pipeline for all-attributes
            # TODO: Leave code here for future reference.
            # if value is not None and len(pipeline) > 1:
            #     # process remaining policies in the pipeline
            #     value = self.offer0(None, value,
            #     pipeline[1:len(pipeline)],
            #     pipelineData[1:len(pipelineData)])

            if value is not None:
                pairs = value
                if len(pairs) == 0:
                    return
                messages = []
                messageDispatcher = MessageDispatcher.getMessageDispatcher(
                    self.__directlyConnectedDevice)
                for messagePair in pairs:
                    message = messagePair.getKey()
                    messages.append(message)
                    solist = messagePair.getValue()
                    if solist is not None:
                        for so in solist:
                            if so is None:
                                continue
                            messageDispatcher.addStorageObjectDependency(so,
                                                        message.getClientId())
                            so.sync()
                messageDispatcher.queue(messages)
        except Exception as e:
            # SEVERE
            self._getLogger().critical(f'_processExpiredFunction(): {e}')

    def checkComputedMetrics(self, updatedAttributes):
        """

        :param updatedAttributes:
        :return:
        """

        if updatedAttributes is None or len(updatedAttributes) == 0:
            return set()

        with self.__computedMetricTriggerMapLock:
            if len(self.__computedMetricTriggerMap) == 0:
                return set()
            # This will be the set of attributes that have computed metrics
            # that are triggered by the set of updated attributes.
            computedAttributes = set()

            for entry in self.__computedMetricTriggerMap:
                # This is the set of attributes that the formula refers to.
                key = entry.getKey()
                # If the set of attributes that the formula refers to
                # is a subset of the updated attributes, then compute
                # the value of the attribute.
                if key.issubset(updatedAttributes):
                    computedAttributes.add(entry.getValue())

        if len(computedAttributes) == 0:
            return computedAttributes

        attributestoremove = set()
        for attributeName in computedAttributes:
            attribute = self.getAttribute(attributeName)
            if not attribute.isSettable():
                self._getLogger().warning(f'checkComputedMetrics(): '
                    f'attempt to modify read-only attribute "{attributeName}"')
                attributestoremove.add(attributeName)
                continue
            devicePolicy = self.__devicePolicyManager.getPolicy(
                self.getDeviceModel().getURN(), self.getEndpointId())
            if devicePolicy is None:
                continue
            pipeline = devicePolicy.getPipeline(attributeName)
            if pipeline is None or len(pipeline) == 0:
                continue
            pipelineData = self.getPipelineData(attributeName)
            # offer0 returns true if the attribute was set. If the attribute
            # was not set, then remove the attribute name from the list of
            # attributesToCompute
            policyValue = self.offer0(attribute.getDeviceModelAttribute(),
                                      attribute.get(), pipeline, pipelineData)
            if policyValue is not None:
                # FINE
                self._getLogger().debug(f'checkComputedMetric(): '
                f'{self.getEndpointId()} : Set : "{attributeName}"="{policyValue}')
                attribute.update(policyValue)
            else:
                attributestoremove.add(attributeName)

        for attributeName in attributestoremove:
            computedAttributes.remove(attributeName)

        return computedAttributes

    # DeviceAnalog implementation

    def setAttributeValue(self, attribute, value):
        """

        :param attribute:
        :param value:
        :return:
        """
        self.set(attribute, value)

    def getAttributeValue(self, attribute):
        """

        :param attribute:
        :return:
        """
        return self.get(attribute)

    def call(self, actionName, args):
        """

        :param actionName:
        :param args: single argument or `list` of args
        """

        self.callImpl(actionName, args)

    # def queueMessage(self, message):
    #     """ generated source for method queueMessage """
    #     self.queueMessage(message, None)

    def _getActionArguments(self, action, argumentValues):
        """
        Return the arguments of the device model action as a :class:`.NamedValue` chain.

        The name/value
        pairs in the chain are in the same order as they appear in the device model action.
        All arguments are returned. If there is no value for an argument in the JSON body,
        the argument's default value is used.

        :param action: the action
        :param argumentValues: the arguments as a `dict`
        :return: the root node of the chain, or `None` if there are no arguments.
        """

        # multiple args for callback are passed as chain of NameValue.
        # 'arguments' is the root of the NamedValue chain.
        arguments = None

        actionName = action.getName()
        deviceModelActionArguments = action.getArguments()
        if deviceModelActionArguments is not None and \
                not len(deviceModelActionArguments) == 0:
            # current argument in the chain
            arg = None
            # For each arg in the action, get the arg value from the JSON body
            # and add it to the value to 'arguments'
            for actionArg in deviceModelActionArguments:
                argName = actionArg.getName()
                if argName in argumentValues:
                    argValue = argumentValues[argName]
                else:
                    argValue = actionArg.getDefaultValue()
                value = None
                try:
                    argType = actionArg.getArgType()
                    value = argType.cast(actionArg.getArgType(), argValue, convert=True)
                    if isinstance(value, numbers.Number):
                        actionArg.checkBounds(value)

                except Exception as e:
                    raise ArgumentException(f'VirtualDeviceImpl.getActionArguments(): '
                        f'{action.getName()}  argument {argName} bad value: {argValue}, {e}')

                namedValue = VirtualDeviceBase.NamedValueImpl(argName, value)
                if arg is not None:
                    arg.setNext(namedValue)
                    arg = namedValue
                else:
                    arg = arguments = namedValue

        return arguments

    def callImpl(self, actionName, argumentValues):
        """
        Invoke an action callback.

        :param actionName: the name of the action from the device model
        :param argumentValues: a map of argument names to values
        :return: a success or failure :class:`.StatusCode` code
         """

        # if actionCallbackMap has the action name or "*" (all actions), or
        # if callableMap (legacy) contains the action name, then handle the action.
        specificActionCallback = None
        allActionsCallback = None
        callable = None
        with self.__ACTIONMAP_LOCK:
            if actionName in self.__actionCallbackMap:
                specificActionCallback = self.__actionCallbackMap[actionName]
            if "*" in self.__actionCallbackMap:
                allActionsCallback = self.__actionCallbackMap["*"]
            if actionName in self.__callableMap:
                callable = self.__callableMap[actionName]

        if ((specificActionCallback is None) and (allActionsCallback is None) and (callable is None)):
            self._getLogger().warning(f'{self.getDeviceModel().getURN()} no handler for "{actionName}"')
            return StatusCode.NOT_IMPLEMENTED

        try:
            deviceModelAction = self.getDeviceModelAction(self.__base.getDeviceModel(), actionName)
            if deviceModelAction is None:
                self._getLogger().warning(f'{self.getDeviceModel().getURN()} does not contain action "{actionName}"')
                return StatusCode.BAD_REQUEST

            actionarguments = self._getActionArguments(deviceModelAction, argumentValues)
            class CallImplActionEvent(VirtualDevice.ActionEvent):
                __slots__ = (
                    'actionName',
                    'virtualDevice',
                    'arguments'
                )

                def __init__(self, actionName, virtualDevice, arguments):
                    super().__init__()
                    self.actionName = actionName
                    self.virtualDevice = virtualDevice
                    self.arguments = arguments

                def getName(self):
                    return self.actionName

                def getVirtualDevice(self):
                    return self.virtualDevice

                def getNamedValue(self):
                    return self.arguments

            actionEvent = CallImplActionEvent(actionName, self, actionarguments)

            if specificActionCallback is not None:
                specificActionCallback.onAction(actionEvent)

            if allActionsCallback is not None:
                allActionsCallback.onAction(actionEvent)

            if callable is not None:
                data = actionarguments.getValue() if actionarguments is not None else None
                callable.call(self, data)

                return StatusCode.ACCEPTED
        except Exception as e:
            # The call method may throw exceptions. Since the CL has no
            # knowledge of what these might be, Exception is caught here.
            #
            # Possible exceptions from CL internals are JSONException,
            # IllegalArgumentException, NullPointerException, and
            # ClassCastException.
            #
            self._getLogger().warning(f'VirtualDeviceImpl.callImpl(): {self.getEndpointId()} {e}')
            return  StatusCode.BAD_REQUEST


# The default `iotcs.device.MessageDispatcher.ErrorCallback` for handling
# message send errods
_ERROR_CALLBACK_BRIDGE = VirtualDeviceImpl.ErrorCallbackBridge()
