# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

from abc import ABC
from abc import abstractmethod

from http.client import RemoteDisconnected
from datetime import datetime
import json
import logging
from queue import PriorityQueue
from queue import Queue
from queue import Empty
import signal
from threading import Thread
import threading
import time
import traceback

from iotcs.client import StorageObject
from iotcs.common import TimeManager
from iotcs.common import StatusCode
from iotcs.common import Util
from iotcs.diagnostics import Diagnostics
from iotcs.diagnostics import TestConnectivity
from iotcs.exception import ArgumentException
from iotcs.exception import NetworkException
from iotcs.exception import TransportException
from iotcs.exception import SocketException
from iotcs.exception import SocketTimeoutException
from iotcs.exception import SecurityException
from iotcs.exception import StateException
from iotcs.message import Message
from iotcs.message import ResponseMessage
from iotcs.message import Resource
from iotcs.message import ResourceMessage
from iotcs.messaging.client.device.persistence import MessagePersistence
from iotcs.messaging.client.device.policy import DevicePolicyManager
from iotcs.messaging.client.device.policy import MessagingPolicy
from iotcs.messaging.client.device.util import RequestHandler
from iotcs.messaging.client.device.util import RequestDispatcher
from iotcs.persistence import PersistenceStoreManager

# com.oracle.iot.client.device.util

class MessageDispatcher(ABC):
    """
    The MessageDispatcher queues messages for automatic dispatching.

    Messages are dispatched to the Oracle IoT Cloud Service.
    The MessageDispatcher prioritizes message dispatching to ensure
    high priority messages are dispatched ahead of lower priority messages.
    There is one MessageDispatcher instance per client.
    """

    __slots__ = ()

    __dispatcherMap = dict()
    __dispatcherMapLock = threading.Lock()

    MESSAGE_DISPATCHER_URN = "urn:oracle:iot:dcd:capability:message_dispatcher"

    @classmethod
    def removeMessageDispatcher(cls, dcd):
        """
        Remove the instance of :class:`.MessageDispatcher` for `dcd`.

        After calling `removeMessageDispatcher`, a call to
        :meth:`.getMessageDispatcher` will return `None`.
        This is an implementation detail and is called from the implementation
        of the virtualization API.

        :param dcd: is a :class:`.DirectlyConnectedDevice` or
        :class:`.Gateway` connected device
        :return: the :class:`.MessageDispatcher` for `dcd` or `None`
        """
        with cls.__dispatcherMapLock:
            return cls.__dispatcherMap.pop(dcd, None)

    @classmethod
    def getMessageDispatcher(cls, dcd):
        """
        Return the instance of a :class:`.MessageDispatcher` for the `dcd`.

        This is an implementation detail and is called from the implementation
        of the virtualization API.

        :param dcd: is a :class:`.DirectlyConnectedDevice` or
        :class:`.Gateway` connected device
        :return: A :class:`.MessageDispatcherImpl` for `dcd`.
        """
        messageDispatcher = None
        with cls.__dispatcherMapLock:
            if dcd in cls.__dispatcherMap.keys():
                messageDispatcher = cls.__dispatcherMap[dcd]
            else:
                messageDispatcher = cls.__dispatcherMap[dcd] = \
                    MessageDispatcherImpl(dcd)
            return messageDispatcher

    @classmethod
    def setMessageDispatcher(cls,  dcd,  messageDispatcher):
        """
        Set the instance of a :class:`.MessageDispatcher` for the `dcd`.

        This is an implementation detail and is called from the implementation
        of the virtualization API.

        :param dcd: is a :class:`.DirectlyConnectedDevice` or
        :class:`.Gateway` connected device
        :param messageDispatcher: is the :class:`.MessageDispatcher` for dcd.
        """

        with cls.__dispatcherMapLock:
            cls.__dispatcherMap[dcd] = messageDispatcher

    @abstractmethod
    def getRequestDispatcher(self):
        """
        Return the :class:`<.RequestDispatcher>` used by instance.

        :return: a :class:`<.RequestDispatcher>`
        """
        pass

    # TODO: Not possible to have both queue methods
    # with single message and array message arg, instead test for "type" Message or
    # [Message] or whatever the syntax is in python. Define the method that it can
    # take both arg types and require the implementation to act appropriately
    @abstractmethod
    def queue(self, messages):
        """
        Add the messages to the outgoing message queue if it is possible to
        do so without violating capacity restrictions.
        :param messages: queue a single or list of `iotcs.shared.Message` sublclass instances
        :raises: :class:`.StateException` if all the messages cannot be added to the queue
                or if `messages` is `None` or empty
        """
        pass


    @abstractmethod
    def offer(self, messages):
        """
        Offer messages to be queued.

        Depending on the policies, if any,
        the messages will be queued if it is possible to do so without
        violating capacity restrictions.

        :param messages: a `list` of messages to be queued
        :raises: :class:`.ArgumentException` if `messages` is `None` or empty
        """
        pass

    @abstractmethod
    def setOnError(self, callback):
        """
        Set a callback to be notified if there is an error in sending messages.
        :param callback: An instance of `iotcs.device.MessageDispatcher.ErrorCallback`.
            If `callback` is `None` the existing callback will be removed.
        """
        pass

    @abstractmethod
    def setOnDelivery(self, callback):
        """
        Set a callback to be notified if message is successfully delivered.
        :param callback: An instance of `iotcs.device.MessageDispatcher.DeliveryCallback`.
            If `callback` is `None` the existing callback will be removed.
        """
        pass

    @abstractmethod
    def close(self):
        pass

    class DeliveryCallback(ABC):
        """
        A callback interface for successful delivery of messages.
        """
        @abstractmethod
        def delivered(self, messages):
            """
            Notify that messages have been successfully dispatched.
            This callback indicates that the messages have been delivered to,
            and accepted by, the server and are no longer queued.
            :param messages: the delivered messages.
            """
            pass

    class ErrorCallback(ABC):
        """
        A callback interface called if errors occur during the delivery of messages.
        """
        @abstractmethod
        def failed(self, messages, exception):
            """
            Notify that an error occurred when attempting to deliver messages
            to the server. This callback indicates that the messages have
            not been delivered to the server and are no longer queued.
            :param messages: the messages that were not delivered
            :param exception: the exception that was raised when attempting
                to deliver the messages.
            """
            pass

class MessageDispatcherImpl(MessageDispatcher):
    """ The MessageDispatcher queues messages for automatic dispatching to the
    for the Oracle IoT Cloud Service.
    The MessageDispatcher prioritizes message dispatching to ensure
    high priority messages are dispatched ahead of lower priority messages.
    There is one MessageDispatcher instance per client.
    """

    class Counters():
        """
        Class to maintain counters
        """
        __slots__ = (
            'counters',
            'countersLock',
            'md'
        )

        def __init__(self, md):
            self.counters = dict()
            self.counters["load"] = 0
            self.counters["totalBytesSent"] = 0
            self.counters["totalBytesReceived"] = 0
            self.counters["totalMessagesReceived"] = 0
            self.counters["totalMessagesRetried"] = 0
            self.counters["totalMessagesSent"] = 0
            self.counters["totalProtocolErrors"] = 0

            self.md = md
            self.countersLock = threading.Lock()

        def __load(self):
            size = float(self.md.outgoingMessageQueue.qsize())
            max = float(self.md.maximumQueueSize)
            # we only want 1 decimal point percision
            load = int(float(size / max) * float(1000))
            self.counters["load"] = load

        def update(self, counter, increment):
            with self.countersLock:
                self.counters[counter] += increment
                if counter == "totalMessagesSent":
                    self.counters["load"] = self.__load()

        def reset(self):
            with self.countersLock:
                for counter in self.counters.keys():
                    self.counters[counter] = 0

        def toJson(self):
            with self.countersLock:
                # update load
                self.counters["load"] = self.__load()
                return json.loads(json.dumps(self.counters))

    class CountersHandler(RequestHandler):
        __slots__ = (
            'counters'
        )
        COUNTERS_URL = "deviceModels/" + MessageDispatcher.MESSAGE_DISPATCHER_URN + "/counters"

        def __init__(self, counters):
            super(MessageDispatcherImpl.CountersHandler, self).__init__()
            self.counters = counters

        def handleRequest(self, request):

            if request is None:
                raise ValueError(f'CountersHandler.handlRequest(): request message is None')

            statusCode = StatusCode.BAD_REQUEST
            response = ResponseMessage(requestMessage=request)
            if request.getMethod() is None or request.getMethod().upper() != \
                    Resource.Method.GET.name.upper():
                statusCode = StatusCode.METHOD_NOT_ALLOWED
            else:
                try:
                    jsonCounters = self.counters.toJson()
                    if jsonCounters is not None:
                        response.setBody(jsonCounters)
                        statusCode = StatusCode.OK
                except Exception as e:
                    statusCode = StatusCode.BAD_REQUEST

            # if we get here then not OK
            response.setStatusCode(statusCode)
            return response


    class ResetHandler(RequestHandler):
        __slots__ = (
            'counters'
        )
        RESET_URL = "deviceModels/" + MessageDispatcher.MESSAGE_DISPATCHER_URN + "/reset"

        def __init__(self, counters):
            super(MessageDispatcherImpl.ResetHandler, self).__init__()
            self.counters = counters

        def handleRequest(self, request):

            if request is None:
                raise ValueError(f'ResetHandler.handlRequest(): request message is None')

            statusCode = StatusCode.BAD_REQUEST
            response = ResponseMessage(requestMessage=request)
            if request.getMethod() is None or request.getMethod().upper() != \
                    Resource.Method.PUT.name.upper():
                statusCode = StatusCode.METHOD_NOT_ALLOWED
            else:
                try:
                    self.counters.reset()
                    statusCode = StatusCode.OK
                except Exception as e:
                    statusCode = StatusCode.BAD_REQUEST

            response.setStatusCode(statusCode)
            return response

    class PollingIntervalHandler(RequestHandler):
        __slots__ = (
            'md'
        )
        POLLING_INTERVAL_URL = \
            f'deviceModels/{MessageDispatcher.MESSAGE_DISPATCHER_URN}/pollingInterval'

        def __init__(self, md):
            super(MessageDispatcherImpl.PollingIntervalHandler, self).__init__()
            self.md = md

        def handleRequest(self, request):

            if request is None:
                raise ArgumentException(
                    f'PollingIntervalHandler.handlRequest(): request message is None')

            body = None
            statusCode = StatusCode.BAD_REQUEST
            response = ResponseMessage(requestMessage=request)
            method = request.getMethod()
            method = method.upper() if method is not None else None
            if method not in [Resource.Method.PUT.name.upper(),
                              Resource.Method.GET.name.upper()]:
                statusCode = StatusCode.METHOD_NOT_ALLOWED
            else:
                if method == Resource.Method.GET.name.upper():
                    try:
                        jsonObject = dict()
                        jsonObject["value"] = self.md.getPollingInterval()
                        body = jsonObject
                        statusCode = StatusCode.OK
                    except Exception as e:
                        statusCode = StatusCode.BAD_REQUEST
                else: # PUT
                    try:
                        jsonObject = json.loads(request.getBody())
                        if jsonObject is not None and "value" in jsonObject:
                            # pollingInterval is only written here and in md constructor
                            pollingInterval =  int(jsonObject["value"])
                            if pollingInterval < 0:
                                body = f'polling interval must be a numeric value greater than or equal to 0'
                            else:
                                self.md.setPollingInterval(pollingInterval)
                                statusCode = StatusCode.OK
                        else:
                            body = f'The "value" parameter must be supplied'
                    except Exception as e:
                        body = f'The "value" parameter must be a number'
                        statusCode = StatusCode.BAD_REQUEST

            response.setBody(body)
            response.setStatusCode(statusCode)
            return response


    # TODO: This and all Thread subclasses, use Thread arguments to pass the
    # TODO: MessageDispatcher instance to the running thread, no need for class member.

    class PendingRequestProcessor(Thread):

        __slots__ = (
            'settleTime', # milliseconds
            'timeZero', # milliseconds
            'averageWaitTime', # float in seconds
            'settled',
            'md',
            'running'
        )
        # TODO: Should be configurable
        # In milliseconds
        MAX_WAIT_TIME = 1000

        def __init__(self, md, settleTime):
            super(MessageDispatcherImpl.PendingRequestProcessor, self).__init__()
            name = 'iotcs.device.MessageDispatcherImpl.PendingRequestProcessor'
            self.daemon = True
            # Just to shorten the reference
            # integer in milliseconds
            maxwaittime = MessageDispatcherImpl.PendingRequestProcessor.MAX_WAIT_TIME
            # integer in milliseconds
            self.settleTime = settleTime
            self.timeZero = TimeManager.currentTimeMillis()
            if settleTime <= maxwaittime:
                # convert to a real number of seconds, settleTime is in milliseconds
                self.averageWaitTime = float(self.settleTime / 1000)
            else:
                quotient = float(self.settleTime / maxwaittime)
                waitTimeMillis = \
                    float(maxwaittime + (self.settleTime - quotient * maxwaittime) / quotient)
                # convert to a real number of seconds, waitTimeMillis is in milliseconds
                self.averageWaitTime = float(waitTimeMillis / 1000)

            self.settled = False
            self.md = md
            self.running = False

        def run(self):
            self.running = True
            while self.running and not self.md.isRequestClose() and not self.settled:

                #List<RequestMessage>
                messageList = None

                with self.md.pendingQueueLock:

                    try:
                        # pendingTrigger.wait returns True if condition was signalled
                        # before timeout expires. If timeout expires returns False.
                        triggersignalled = False
                        # Convert to a real number of seconds
                        # wait at most the amount of time left before settleTime expires
                        waitTime = float((self.settleTime -
                            (TimeManager.currentTimeMillis() - self.timeZero)) / 1000)
                        if waitTime > 0 and self.md.pendingRequestMessages.empty():
                            triggersignalled = self.md.pendingTrigger.wait(waitTime)

                        # waitTime represents how much time is left before settleTime expires.
                        # If we waited in the pendingTrigger.wait call, then waitTime
                        # has been adjusted by how long we actually waited. If waitTime is
                        # greater than zero and we blocked in wait, then a request
                        # was added to the queue. If waitTime is less than or equal to zero,
                        # then we waited the entire time, or waitTime was less than or equal
                        # to zero to begin with because settleTime has expired.

                        # If we waited the entire time and the queue is still empty, bail.
                        # Otherwise, a request was queued, or the queue wasn't empty to begin with.
                        # 'triggersignaled' is true if wait returned before 'waitTime' was exceeded
                        # 'triggersignaled' is false if wait blocked for the entire waitTime
                        if not self.running and not triggersignalled and \
                                self.md.pendingRequestMessages.empty():
                            break

                        # In the case where waitTime is greater than zero,
                        # (triggersignalled == True) we want to make
                        # sure we wait at least averageWaitTime before dispatching the request.
                        # In other words, if we were blocked in the awaitNanos (wait) and a request
                        # was added to the queue, we don't want to process it immediately
                        # rather, wait at least one average wait time so we don't thrash.
                        # 'triggersignalled' is true if wait did not block the full 'waitTime'
                        if triggersignalled:
                            waitTime = self.averageWaitTime
                            waitStart = TimeManager.currentTimeMillis()
                            while waitTime > 0:
                                triggersignalled = self.md.pendingTrigger.wait(waitTime)
                                # TODO: Can we exit here if stopped ?
                                if self.running and triggersignalled:
                                    blockedfor = \
                                        float((TimeManager.currentTimeMillis() - waitStart) / 1000)
                                    waitTime = waitTime - blockedfor
                                else:
                                    # we blocked the entire waitTime
                                    break

                        # Operate on a copy of pendingRequestMessages so that we don't hold
                        # pendingQueueLock while dispatching. At the end of this loop,
                        # pendingRequestMessages will be empty.
                        if self.running and not self.md.pendingRequestMessages.empty():
                            messageList = []
                            requestMessage = None
                            try:
                                requestMessage = self.md.pendingRequestMessages.get(block=False)
                            except:
                                # ignore Empty exception
                                pass

                            while requestMessage is not None:
                                messageList.append(requestMessage)
                                try:
                                    requestMessage = \
                                        self.md.pendingRequestMessages.get(block=False)
                                except:
                                    # Could be the "Empty" exception from queue
                                    requestMessage = None

                    # TODO: Really need to understand potential thread interrupts
                    except InterruptedError as e:
                        continue

                self.settled |= \
                    self.settleTime <= (TimeManager.currentTimeMillis() - self.timeZero)

                if messageList is not None and len(messageList) > 0:
                    for requestMessage in messageList:
                        MessageDispatcherImpl.Dispatcher(self.md, requestMessage,
                                                         self.settled).start()

            self.md.getLogger(fromclass='PendingRequestProcessor').debug(f'run(): Exiting')
            self._stop()

        def _stop(self):
            self.running = False


    class Dispatcher(Thread):
        __slots__ = (
            'md',
            'requestMessage',
            'settled',
        )

        def __init__(self, md, requestMessage, settled=True):
            super(MessageDispatcherImpl.Dispatcher, self).__init__()
            self.name = "iotcs.device.MessageDispatcherImpl.Dispatcher"
            self.daemon = True
            self. md = md
            self.requestMessage = requestMessage
            self.settled = settled

        def run(self):
            responseMessage = \
                self.md.requestDispatcher.dispatch(self.requestMessage)
            if self.settled or \
                    responseMessage.getStatusCode() != StatusCode.NOT_FOUND:
                if self.md.getLogger('Dispatcher').isEnabledFor(logging.DEBUG):
                    self.md.getLogger('Dispatcher').debug(f'Queueing request {self.requestMessage}')
                try:
                    self.md.queue([responseMessage])
                except Exception as e:
                    # SEVERE
                    self.md.getLogger(fromclass='Dispatcher').critical(
                        f'run(): [settled={self.settled}, status={responseMessage.getStatusCode()}'
                            f'Could not queue requestMessage caught exception {e}')
            else:
                # Not settled or status code == not found
                # try again later
                with self.md.pendingQueueLock:
                    try:
                        if self.md.getLogger('Dispatcher').isEnabledFor(logging.DEBUG):
                            self.md.getLogger('Dispatcher').debug(f'Adding pending request {self.requestMessage}')
                        self.md.pendingRequestMessages.put(
                            self.requestMessage, block=False)
                    except Exception as e:
                        # SEVERE
                        # pendingRequestMessages could be full, but should be infinite size.
                        self.md.getLogger(fromclass='Dispatcher').critical(
                            f'run(): Cannot queue pendingRequestMessage for dispatch {e}')
                        responseMessage = \
                            ResponseMessage(requestMessage=self.requestMessage,
                                            statusCode=StatusCode.INTERNAL_SERVER_ERROR)
                        self.md.queue([responseMessage])
                    finally:
                        self.md.pendingTrigger.notify_all()

    class Receiver(Thread):
        __slots__ = (
            'md',
            'counters',
            'running',
            'settled',
            'settleTime',
            'timeZero',
            'pendingRequestProcessor'
        )

        def __init__(self, md):
            super(MessageDispatcherImpl.Receiver, self).__init__()
            self.name = "iotcs.device.MessageDispatcherImpl.Receiver"
            self.daemon = True
            self.md = md
            self.counters = md.counters
            self.running = False

            value = Util.getConfig().getint(
                MessageDispatcherImpl.DEVICE_MESSAGING_SECTION,
                MessageDispatcherImpl.SETTLE_TIME_PROPERTY)
            self.settleTime = value if value >= 0 else \
                MessageDispatcherImpl.DEFAULT_SETTLE_TIME
            self.timeZero = TimeManager.currentTimeMillis()
            self.settled = self.settleTime == 0

            if self.settleTime > 0:
                self.pendingRequestProcessor = \
                    MessageDispatcherImpl.PendingRequestProcessor(
                        md, self.settleTime)
                self.pendingRequestProcessor.start()

        def run(self):
            self.running = True
            while self.running and not self.md.isRequestClose():

                # If there was a SocketException observed before,
                # wait for this flag "waitOnReconnect" to be set to false
                # by the transmitter thread after the next message was sent successfully
                # or close() was called.

                while self.running and self.md.getWaitOnReconnect() and not self.md.isRequestClose():
                    self.md.getLogger(fromclass='Receiver').debug("608: waiting on reconnect")
                    with self.md.waitOnReconnectLock:
                        self.md.getLogger(fromclass='Receiver').debug("610: back from waiting on reconnect")
                        try:
                            # TODO: Should this be timed ?
                            self.md.getLogger(fromclass='Receiver').debug(
                                msg = f'run(): waiting on reconnect')
                            self.md.waitOnReconnectLock.wait()
                            self.md.getLogger(fromclass='Receiver').debug(
                                msg = f'run(): back from wait on reconnect')
                        except:
                            # TODO: What to do on exception ?
                            # TODO: What exceptions can occur ?
                            #self._stop()
                            pass

                        if not self.running or self.md.isRequestClose():
                            self._stop()
                            return

                # If not long polling, block until the pollingInterval has
                # timed out or a message has been sent
                if not self.md.useLongPolling:
                    with self.md.receiveLock:
                        try:
                            # Wait here until the messageSent is signalled, or
                            # until the pollingInverval is expired.
                            pollingInterval = self.md.getPollingInterval()
                            if pollingInterval > 0:
                                self.md.messageSent.wait(float(float(pollingInterval) / 1000.0))
                            else:
                                self.md.messageSent.wait()
                        except Exception as e:
                            continue
                    # checkout for running and closed state
                    if not self.running:
                        return
                    if self.md.isRequestClose():
                        self._stop()
                        return

                try:
                    # Receive will ignore -1 timeout when not long polling
                    self.md.getLogger(fromclass='Receiver').debug(
                        msg = f'run(): Calling receive')
                    requestMessage = self.md.deviceClient.receive(-1)
                    self.md.getLogger(fromclass='Receiver').debug(
                        msg = f'run(): Back from receive')
                    while requestMessage is not None:

                        srimpl = self.md.deviceClient.sendReceiveImpl
                        timetowait = srimpl.sendReceiveTimeLimit
                        #time.sleep(float(timetowait / 1000))

                        self.counters.update("totalMessagesReceived", 1)
                        totalbytesreceived = \
                            len(json.dumps(requestMessage.toJson()).encode(encoding='utf8'))
                        self.counters.update("totalBytesReceived", totalbytesreceived)

                        if self.running and not self.md.isRequestClose():
                            self.settled |= \
                                (self.settleTime <= (TimeManager.currentTimeMillis() - self.timeZero))
                            # TODO: Use thread pool ?
                            dispatcher = self.md.Dispatcher(self.md,
                                                            requestMessage,
                                                            self.settled)
                            dispatcher.start()
                        # Continue to iterate over received messages
                        requestMessage = self.md.deviceClient.receive(-1)
                except Exception as e: # IOException
                # TODO: Are there any exceptions where we don't want to do this ?
                # log the exception, don't quit the receive loop
                # Network connection issues detected.
                # So wait for reconnection signal from transmitter thread.
                    self.md.getLogger(fromclass='Receiver').debug(
                        msg = f'run(): deviceClient.receive raised Exception {e}')
                    self.md.setWaitOnReconnect(True)
                    continue

            self.md.getLogger(fromclass='Receiver').debug(msg = f'run(): Exiting')

        def _stop(self):
            if not self.running:
                return
            if self.pendingRequestProcessor.is_alive():
                self.pendingRequestProcessor._stop()
                with self.md.pendingQueueLock:
                    self.md.pendingTrigger.notify_all()
                try:
                    self.pendingRequestProcessor.join(self.md.JOINTIMEOUT)
                except Exception as e:
                    self.md.getLogger(fromclass='Receiver').debug(
                        f'close(): pendingRequestProcessor.join() raised exception {e}')
                self.md.getLogger(fromclass='Receiver').debug(
                    msg = f'_stop(): Stopping PendingRequestProcessor')
            self.running = False



    class Transmitter(Thread):
        # EEE - day of the week
        # d - day in the month
        # MMM - Month in year
        # yyyy - week year
        # HH - hour in day
        # mm - minute in hour
        # ss - second in minute
        # z - time zone
        # RFC1123_FORMAT = new SimpleDateFormat("EEE, dd MMM yyyy HH:mm:ss z", Locale.ROOT);


        __slots__ = {
            'md',
            'counters',
            'running',
            'backoff',
            'attempt',
            'delayinmillis',
            # For exponential backoff, take the Fibbonaci value times the delay in milliseconds.
            'fib',
            'RFC1123_FORMAT'
        }

        def __init__(self, md):
            super(MessageDispatcherImpl.Transmitter, self).__init__()
            self.name = "iotcs.device.MessageDispatcherImpl.Transmitter"
            self.daemon = True
            self.md = md
            self.counters = md.counters
            self.running = False

            self.backoff = 0
            self.attempt = 0
            self.delayinmillis = self.md.backoffDelay
            self.fib = [0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55, 89, 144]
            self.RFC1123_FORMAT = "%a, %d %b %Y %H:%M:%S %z"

        def calculateBackoff(self, retryafter):
            """
            Calculate the new value of 'backoff'.

            If backoff > 0, then we are already backing off and
            no adjustment is made. Only adjust backoff if backoff
            less than or equal to zero. A backoff value of less
            than or equal to zero means we are either not
            currently backing off, or the backoff period has
            expired. This method should only be called
            from the 'send' method when handling an IOException.
            :return:
            """
            # Calculate backoff from 'Retry-After' header (RFC-7231).
            # If retryAfter is null, then backoff is calculated by the
            # number of attempts.
            if retryafter is not None:
                # retryAfter is the 'Retry-After' header from an HTTP 503 or 429 response.
                # The value is either a number (indicating seconds), or a RFC-1123 date.
                try:
                    backoff = int(retryafter)
                except Exception as e:
                    try:
                        retryafter_datetime = datetime.strptime(retryafter, self.RFC1123_FORMAT)
                        future = retryafter_datetime.timestamp() * 1000
                        now = TimeManager.currentTimeMillis()
                        backoff = future - now
                    except Exception as ee:
                        self.md.getLogger(fromclass='Transmitter').critical(
                            f'calculateBackoff(): Invalid Retry-After value {retryafter} {ee}')

            if self.backoff <= 0:
                self.attempt = min(self.attempt + 1, len(self.fib) - 1)
                self.backoff = self.fib[self.attempt] * self.delayinmillis
            return self.backoff

        def run(self):
            # Messages pending delivery. Messages from the outgoingMessageQueue
            # are placed into this list.
            pendingMessages = list()
            self.running = True

            while self.running:

                # 'newAlert' is set to true if the outgoingMessageQueue
                # contains an alert message. This knowledge is used by
                # the send method to know whether or not to force a send
                # when backoff > 0L. Basically, if we are backing off and
                # a new alert is queued, retry sending immediately.
                newAlert = False
                with self.md.sendLock:
                    # measure how long we wait.
                    t0 = TimeManager.currentTimeMillis()
                    while not self.md.isRequestClose() and \
                            self.md.outgoingMessageQueue.empty():
                        if self.backoff > 0:
                            # If backing off, wait only as long as the backoff time,
                            # then break this loop even if no messages were queued.
                            # This lets the pending messages be retried when there
                            # are no messages being queued.
                            self.md.messageQueued.wait(timeout=self.backoff)
                            break
                        else:
                            self.md.messageQueued.wait()

                    # Adjust backoff by how long we waited for a message to be queued.
                    # If we've waited longer than we need set backoff to 0
                    if self.backoff > 0:
                        waittime = TimeManager.currentTimeMillis() - t0
                        self.backoff = max(self.backoff - waittime, 0)

                    # Add outgoing messages to pendingMessages. Call
                    # send outside of sendLock so that the queue method
                    # is not blocked while sending
                    message = None
                    while True:
                        try:
                            message = self.md.outgoingMessageQueue.get(block=False)
                            if message is not None:
                                newAlert |= message.getType() is Message.Type.ALERT
                                pendingMessages.append(message)
                            else:
                                # message dispatcher most likey closed
                                break
                        except Empty as empty:
                            self.md.getLogger(fromclass='Transmitter').debug(
                                f'run(): outgoingMessageQueue is drained.')
                            break
                        except Exception as e:
                            # TODO: What can cause this ?
                            self.md.getLogger(fromclass='Transmitter').warning(
                                f'run(): exception draining queue: {e}')
                            break


                # sendLock given up
                try:
                    # No op if pendingmessages is zero length
                    self.send(pendingMessages, newAlert)
                except Exception as e:
                    self.md.getLogger(fromclass='Transmitter').warning(f'run(): '
                        f'caught unhandled exception from send(), messages may not have been sent')
                    self.md.dispatchError(pendingMessages, e)

        def getMessagesToSend(self, pendingMessages, receievedNewAlert):
            """
            Get messages to send from the pendingMessages list.

            After this method returns, there may still be messages in the
            pendingMessages list. Left over messages are messages that
            could not be sent because of an in-process storage object
            upload, or messages that could not be sent because of we're
            backing off (due to 'HTTP 503: server busy').

            'receivedNewAlert' is a flag to indicate that at a new alert message
            was queued. If we are backing off, and there aren't any new alert messages,
            then this method returns an empty list.

            :param pendingMessages:
            :param newAlert:
            :return:
            """

            if len(pendingMessages) == 0:
                return list()

            # Flag to indicate that we're in backoff mode. If we're in backoff
            # mode, we'll only try to send messages if one of the messages is
            # an alert. In other words, if isBackoff is true and contains
            # alert is false, then this method should return an empty list.
            isbackoff = self.backoff > 0

            # If we're backking off and we haven't received a new alert,
            # return an empty list
            if isbackoff and not receievedNewAlert:
                return list()

            # errorlist are messages with failed content
            # These are passed to the error callback
            errorlist = list()

            # messagelist are messages that are to be sent.
            # This is the list that is returned from this method.
            messagelist = list()

            # Process pendingMessages.
            # Logically this loop says:
            #    "While the message is not dependent on content,
            #    move the messages from the pendingMessagess to messagelist."
            # As soon as the loop hits a message that is dependent
            # on content (i.e. has a storage object that is still uploading)
            # the loop breaks and there may still be messages in the
            # pendingMessages list.
            messagestoremove = list()
            for message in pendingMessages:
                clientid = message.getClientId()
                if message.getType() is Message.Type.RESPONSE:
                    messagestoremove.append(message)
                    messagelist.append(message)
                elif clientid in self.md.failedContentIds:
                    # if the content fails to upload, move the message
                    # to the errorlist. A messsage will never be sent if
                    # the upload failed.
                    messagestoremove.append(message)
                    # TODO: Use Queue to manage capacity, raises Full if
                    # TODO: queue has fixed size
                    self.md.incrementQueueCapacity(1)
                    errorlist.append(message)
                elif not self.md.isContentDependent(clientid):
                    # If the messages is not dependent on content, move
                    # it to the messagelist
                    messagestoremove.append(message)
                    messagelist.append(message)
                else:
                    # The message is dependent on content and that
                    # content upload is still in progress.
                    # All other messages in the wait list will
                    # remain in the wait list for now
                    break

            if len(errorlist) > 0:
                # TODO: Should be "StorageException"
                self.md.dispatchError(errorlist, NetworkException(
                    f'Transmit.getMessagesToSend(): '
                    f'Content sync failed'))

            for removemessage in messagestoremove:
                pendingMessages.remove(removemessage)

            return messagelist

        def messagesSent(self, messages):
            """
            These messages were successfully delivered to the server.

            Update state and notify callbacks

            :param messages:
            :return:
            """
            # Send is successful, so make sure backoff is reset to zero
            self.backoff = 0
            self.attempt = 0

            # Send is successful, increase the queue capacity
            # by the number of messages sent
            # TODO: Use Queue to manage capacity, raises Full if not infinite size
            self.md.incrementQueueCapacity(len(messages))

            # Send is successful, so the receiver thread was waiting
            # for reconnection, send a wakeup signal
            if self.md.getWaitOnReconnect():
                self.md.setWaitOnReconnectAndNotify(False)

            self.md.dispatchDeliveryCallback(messages)
            self.md.counters.update('totalMessagesSent', len(messages))

            # update total bytes
            jsonmessages = list()
            for message in messages:
                jsonmessages.append(message.toJson())
            msgbytes = json.dumps(jsonmessages).encode('utf-8')
            self.counters.update("totalBytesSent", len(msgbytes))

            with self.md.receiveLock:
                self.md.messageSent.notify_all()

        def send(self, pendingMessages, newAlert):
            """

            :param pendingMessages:
            :param newAlert:
            :return:
            """
            # Note that getMessagesToSend modifies pendingMessages.
            # After this call, pendingMessages will typically be empty.
            # But it might not be empty if there are messages waiting for
            # storage object upload
            messages = self.getMessagesToSend(pendingMessages, newAlert)
            if len(messages) == 0:
                return

            # 'fromIndex' is the first index into the messages array of messages passed
            # to DirectlyConnectedDevice.send. Typically, this will be zero. But in the
            # case where we are trying to send while backing off, this will be the index
            # of the first message passed to the DCD.send method.
            fromindex = 0

            try:
                # Flag to know if we're backing off.
                # Use 'attempt' here, not backoff, because
                # backoff will be zero when the backoff
                # period expires.
                backingoff = self.attempt > 0

                # If we are backing off, then we were trying to resend messages.
                # THere may be persisted messages. C.ear them out now, before trying to
                # send again. If the send fails, then they'll get persisted again.
                if backingoff:
                    messagepersistence = MessagePersistence.getInstance()
                    if messagepersistence is not None:
                        guaranteedmessages = list()
                        for message in messages:
                            if message.getReliability() is Message.Reliability.GUARANTEED_DELIVERY:
                                guaranteedmessages.append(message)
                        if len(guaranteedmessages) != 0:
                            messagepersistence.delete(guaranteedmessages)

                # 'iter' counts how many iterations of this do loop we've done.
                # It is used as an index into 'fib'. The idea is taht, if we are
                # backing off, each iteration of this loop we'll send more messages
                # than the previous loop. In theory, this will help prevent the
                # recovering server from being swamped with messages
                iter = 0

                # 'offset' is how far into the pendingMessages array we have left to send
                # On any iteration, there are len(pendingMessages) - offset
                # left to send
                offset = 0

                while True:
                    # The number of messages in messageArray that we have left to send
                    nummessageslefttosend = len(messages) - offset

                    # The number of messages we want to send on this iteration
                    # depends on whenter or not we are backing off.
                    # If we are backing off, send increasingly more each time.
                    # Otherwise, we want to send all the messages
                    nummessageswewanttosend = nummessageslefttosend
                    if backingoff:
                        iter += 1
                        nummessageswewanttosend = self.fib[min(iter, len(self.fib) - 1)] * 10

                    # 'nummessagestosend' is how many messages to send and is
                    # the lessaer of nummessageslefttosend, nummessageswewanttosend
                    # and maximummessagesperconnection
                    nummessagestosend = min(
                        min(nummessageslefttosend, nummessageswewanttosend),
                        self.md.maximumMessagesPerConnection)

                    # remember where this sublist starts. If the send call
                    # throws an exception, the code will re-queue messages into
                    # the pending queue starting from 'fromindex'
                    fromindex = offset

                    # sublist of messages to send
                    sublist = messages[offset:offset + nummessagestosend]
                    offset = offset + nummessagestosend

                    self.md.deviceClient.send(sublist)
                    self.messagesSent(sublist)

                    if offset >= len(messages):
                        break

            except (TransportException, SocketException,
                    SocketTimeoutException, IOError,) as e:

                # Assume instance is one of TransportException or SocketException
                # or some other IOError
                if isinstance(e, TransportException):
                    status = e.getStatusCode()
                    # 429 means "the user has sent too many requests in a given amount of time."
                    #     See RFC-6585.
                    # 502 means "the server, while acting as a gateway or proxy, received
                    #    an invalid response from an inbound server it accessed while
                    #    attempting to fulfill the request."
                    # 503 means "The server is currently unable to handle the
                    #    request due to a temporary overloading or maintenance of the server."
                    # 504 means "the server, while acting as a gateway or proxy, did not
                    #   receive a timely response from an upstream server it needed to
                    #   access in order to complete the request."
                    #
                    # In all of these cases, this means that the IoT Cloud Service
                    # is temporarily unavailable. So backoff and try again.
                    #
                    if status == StatusCode.SERVICE_UNAVAILABLE or status == StatusCode.BAD_GATEWAY or \
                            status == StatusCode.GATEWAY_TIMEOUT or status == StatusCode.TOO_MANY_REQUESTS:
                        retryafter = e.getResponseHeader("Retry-After")
                        self.calculateBackoff(retryafter)
                        self.md.getLogger(fromclass='Transmitter').warning(f'send(): '
                            f'Server busy. Messages will be tried in {self.backoff} milliseconds.')

                elif isinstance(e, (SocketException, SocketTimeoutException,)):
                    # For network issues, do not do exponential backoff;
                    # rather, use constant delay. Side effect here is that
                    # attempt is set to 1. Attempt needs to be one here
                    # because the code in the try branch checks if (attempt==0).                                                                                          // because the code in the try branch checks if (attempt==0).
                    # If attempt is zero, then the persisted messages won't
                    # get un-persisted in the try block.
                    # Note also that this resets backoff to the full delay,
                    # whereas calculateBackoff returns the remaining delay.
                    self.attempt = 1
                    self.backoff = self.fib[self.attempt] * self.delayinmillis
                    self.md.getLogger(fromclass='Transmitter').warning(f'send(): '
                        f'{e}. Messages will be tried in {self.backoff} milliseconds.')

                self.md.counters.update("totalProtocolErrors", 1)

                failedmessages = list()

                # If attempt > 0, then we're either backing off because of a
                # temporary condition with the server or with the network.
                # In this case, we want to persist guaranteed delivery messages.
                # Note that MessagePersistence.getInstance() might return null.
                messagepersistence = MessagePersistence.getInstance() \
                    if self.attempt > 0 else None

                # The messages we want to persist will be added to this list
                messagestopersist = list() \
                    if messagepersistence is not None else None

                # 'fromIndex' is set in the code above to the index of the first element
                # of the sublist that is being passed to DirectlyConnectedDevice.send.
                # Messages in the 'messages' list with index < fromIndex have been
                # successfully sent. Messages in the 'messages' list with fromIndex => index
                # have not been sent and need to be put back into pendingMessages for
                # reprocessing.
                while fromindex < len(messages):
                    message = messages[fromindex]
                    fromindex += 1
                    retries = message.getRemainingRetries()
                    if retries > 0:
                        # assert message not in pendingmessages
                        pendingMessages.append(message)
                        message.setRemainingRetries(retries - 1)
                        # if message is guaranteed, add it to the list of
                        # messages to persist
                        if messagepersistence is not None and \
                                message.getReliability() is Message.Reliability.GUARANTEED_DELIVERY:
                            messagestopersist.append(message)
                    else:
                        # The message was not added back to pendingMessages, so the
                        # capacity of our queue increased by one.
                        failedmessages.append(message)
                        # TODO: Use Queue to manage overflow raises Full if not infinite size
                        self.md.incrementQueueCapacity(1)
                        self.md.getLogger(fromclass='Transmitter').info(f'send(): '
                            f'Cannot queue message for retry. Message discarded: {message.getClientId()}')

                # Persist messages for retry
                if messagepersistence is not None and len(messagestopersist) > 0:
                    messagepersistence.save(messagestopersist,
                                            self.md.deviceClient.getEndpointId())
                # Tell the client messages failed
                if len(failedmessages) > 0:
                    self.md.dispatchError(failedmessages, e)

            except SecurityException as e:
                # Do not retry messages that failed because of GeneralSecurityException
                self.md.dispatchError(messages, e)
                return

        def updateDiagnosticsData(self, messages):
            """

            :param messages: list of `Message`
            :return:
            """
            payload = []
            for msg in messages:
                payload.append(msg.toJson())

            jsonbytes = json.dumps(payload).encode('utf-8')
            self.counters.update("totalBytesSent", len(jsonbytes))
            self.counters.update("totalMessagesSent", len(messages))

        def _stop(self):
            self.running = False


    class ErrorDispatcher(Thread):
        """
        Dispatch send errors in a thread.
        """
        __slots__ = (
            'errorCallback', # An instalce of `iotcs.device.MessageDispatcher.ErrorCallback`
            'messageList', # The list of messages that could not be delivered.
            'exception' # The `Exception` that occurred dispatchine the messages
        )

        def __init__(self, callback, messageList, exception):
            """

            :param callback: An instance of `iotcs.device.MessageDispatcher.ErrorCallback`
            :param messageList: The list of messages that could not be delivered
            :param exception: The `Exception` that occured dispatching the messages
            """
            super(MessageDispatcherImpl.ErrorDispatcher, self).__init__()
            self.name = "iotcs.device.MessageDispatcherImpl.ErrorDispatcher"
            self.daemon = True
            self.errorCallback = callback
            self.messageList = messageList
            self.exception = exception

        def run(self):
            if self.errorCallback is not None:
                self.errorCallback.failed(self.messageList, self.exception)

    # class MessageDispatcherImpl

    DEVICE_MESSAGING_SECTION = "device_messaging"
    DISPATCHER_THREAD_POOL_SIZE_PROPERTY = "request_dispatcher_thread_pool_size"
    MAXIMUM_MESSAGES_TO_QUEUE_PROPERTY = "dispatcher_max_queue_size"
    MAXIMUM_MESSAGES_PER_CONNECTION_PROPERTY = "dispatcher_max_messages_per_connection"
    POLLING_INTERVAL_PROPERTY = "dispatcher_polling_interval"
    BASIC_NUMBER_OF_RETRIES_PROPERTY = "dispatcher_basic_number_of_retries"
    DISABLE_LONG_POLLING_PROPERTY = "disable_long_polling"
    SETTLE_TIME_PROPERTY = "dispatcher_settle_time"
    MESSAGE_QUEUE_WAITTIME_PROPERTY = "message_queue_waittime"

    # TODO: Implement backoff
    MESSAGE_DISPATCHER_BACKOFF_PROPERTY = "message_dispatcher_backoff"
    # Amount of time in milliseconds to backoff in the face of a 503 from the
    # server. This is the starting value for the amount of time to backoff.
    # The backoff time increases exponentially if the server continues to
    # return a 503. Defaults to 1000 milliseconds. Not less than 100 milliseconds.

    JOINTIMEOUT = 5
    # TODO: make configurable
    # The time to wait for a thread to stop during close.

    DEFAULT_SETTLE_TIME = 10000 # ms

    __slots__ = (
        'deviceClient',
        'requestDispatcher',
        # diagnostics and capabilities handlers
        'diagnosticsHandler',
        'resetHandler',
        'pollingIntervalHandler',
        'testConnectivityHandler',
        'countersHandler',
        # long polling flags
        'useLongPolling',
        'pollingInterval',
        'pollingIntervalLock',
        # message queue support
        'maximumQueueSize',
        'outgoingMessageQueue',
        'maximumMessagesPerConnection',
        'messageQueueWaittime',

        'endpointId',
        'contentLock',
        'contentMap',
        'failedContentIds',

        'receiveLock',
        'messageSent',

        # Transmitter synchronization
        'sendLock',
        'messageQueued',

        'requestClose',
        'closed',
        'closeLock',
        # Callback hooks
        'deliveryCallback',
        'deliveryCallbackLock',
        'errorCallback',
        'errorCallbackLock',
        # Messaging threads
        'receiveThread',
        'receiveThread',
        'transmitThread',
        'transmitThread',
        # Counters
        'counters',
        # This flag is used by receiver thread to wait for reconnection
        # when SocketException exception happens due to network disruptions.
        'waitOnReconnect', #AtomicBoolean False
        'waitOnReconnectLock',
        # RequestMessages that were not handled are stored here by the Dispatcher thread.
        # See `Dispatcher.run()` method.
        'pendingRequestMessages',
        'pendingQueueLock',
        'pendingTrigger',
        'basicNumberOfRetries',
        'backoffDelay',
        'queueCapacity',
        'queueCapacityLock'
    )

    def __init__(self, dcd):

        super(MessageDispatcherImpl, self).__init__()

        self.counters = MessageDispatcherImpl.Counters(self)

        self.deviceClient =  dcd
        # TODO: Is this absolutely necessary ?
        self.endpointId = dcd.getEndpointId()
        # TODO: Need to synchronize these flags
        self.requestClose = False
        self.closed = False
        self.closeLock = threading.RLock()

        self.deliveryCallback = None
        self.deliveryCallbackLock = threading.RLock()
        self.errorCallback = None
        self.errorCallbackLock = threading.RLock()

        config = Util.getConfig()

        self.pollingInterval = config.getint(
            Util.CONFIG_DEVICE_MESSAGING_SECTION,
            self.POLLING_INTERVAL_PROPERTY)
        self.maximumQueueSize = config.getint(
            Util.CONFIG_DEVICE_MESSAGING_SECTION,
            self.MAXIMUM_MESSAGES_TO_QUEUE_PROPERTY)
        # This is used to check for message dispatcher close if blocking in
        # a put or get.
        self.messageQueueWaittime = config.getint(
            Util.CONFIG_DEVICE_MESSAGING_SECTION,
            self.MESSAGE_QUEUE_WAITTIME_PROPERTY)
        self.maximumMessagesPerConnection = config.getint(
            Util.CONFIG_DEVICE_MESSAGING_SECTION,
            self.MAXIMUM_MESSAGES_PER_CONNECTION_PROPERTY)
        self.useLongPolling = not config.getboolean(
            Util.CONFIG_DEVICE_MESSAGING_SECTION,
            self.DISABLE_LONG_POLLING_PROPERTY)
        self.basicNumberOfRetries = config.getint(
            Util.CONFIG_DEVICE_MESSAGING_SECTION,
            self.BASIC_NUMBER_OF_RETRIES_PROPERTY)

        self.pollingIntervalLock = threading.RLock()

        self.backoffDelay = max(config.getint(
            Util.CONFIG_DEVICE_MESSAGING_SECTION,
            self.MESSAGE_DISPATCHER_BACKOFF_PROPERTY, fallback=1000), 100)

        self.contentLock = threading.RLock()
        self.contentMap = dict()
        self.failedContentIds = set()

        self.receiveLock = threading.RLock()
        self.messageSent = threading.Condition(lock=self.receiveLock)

        # Need this since the PriorityQueue blocking is only good for
        # one put/get at a time. Need a way to lock to drain the message
        # and to fill it before getting messages.
        self.sendLock = threading.RLock()
        self.messageQueued = threading.Condition(lock=self.sendLock)

        # Python PriorityQueue is synchronized and blocks on get and put.
        self.outgoingMessageQueue = PriorityQueue(
            maxsize=self.maximumQueueSize)
        # TODO: Do we need queueCapacity ? PriorityQueue maintains "maxSize"
        # # TODO: and either throws if full, or waits for space.
        # This keeps track of how many elements the outgoingMessageQueue can
        # accept without overflowing the queue size. queueCapacity is decremented
        # in the queue method when a message is added to outgoingMessageQueue and
        # is incremented in Transmitter#send method when messages are successfully
        # sent.
        self.queueCapacity = self.maximumQueueSize
        self.queueCapacityLock = threading.RLock()

        # PendingRequestProcessor
        self.pendingQueueLock = threading.RLock()
        self.pendingTrigger = threading.Condition(self.pendingQueueLock)

        # RequestMessages that were not handled are stored here by the Dispatcher thread.
        # See `Dispatcher.run()` method.
        self.pendingRequestMessages = Queue()

        self.waitOnReconnect = False
        self.waitOnReconnectLock = threading.Condition()

        self.requestDispatcher = RequestDispatcher.getInstance()

        # As of 06/11/2018 in order to test these capabilities using curl
        # - The 'iot/webapi/v1/endpoints' api must be used
        # - 'https://<server>:443/iot/webapi/v1/endpoints/<endpointid>/resources/<capabilityurn>/<operation>'
        # - The <capabilityurn> MUST NOT be url encoded, i.e. do not change ':' to '%3A' as is required
        # in other rest api calls using curl

        try:
            resourceMessage = ResourceMessage(source=self.endpointId,
                                              endpointName=self.endpointId)

            testConnectivity = TestConnectivity(self.endpointId, self)
            self.testConnectivityHandler = testConnectivity.getTestConnectivityHandler()
            resource = self.__getResource(self.endpointId, TestConnectivity.TEST_CONNECTIVITY_URL,
                                        self.testConnectivityHandler,
                                        [Resource.Method.GET,
                                         Resource.Method.PUT])
            resourceMessage.register(resource)

            self.countersHandler = MessageDispatcherImpl.CountersHandler(self.counters)
            resource = self.__getResource(self.endpointId, MessageDispatcherImpl.CountersHandler.COUNTERS_URL,
                                          self.countersHandler,
                                          [Resource.Method.GET])
            resourceMessage.register(resource)

            self.pollingIntervalHandler = MessageDispatcherImpl.PollingIntervalHandler(self)
            resource = self.__getResource(self.endpointId, MessageDispatcherImpl.PollingIntervalHandler.POLLING_INTERVAL_URL,
                                          self.pollingIntervalHandler,
                                          [Resource.Method.GET,
                                           Resource.Method.PUT])
            resourceMessage.register(resource)

            self.diagnosticsHandler = Diagnostics()
            resource = self.__getResource(self.endpointId, Diagnostics.DIAGNOSTICS_URL,
                                          self.diagnosticsHandler,
                                          [Resource.Method.GET])
            resourceMessage.register(resource)

            self.resetHandler = MessageDispatcherImpl.ResetHandler(self.counters)
            resource = self.__getResource(self.endpointId, MessageDispatcherImpl.ResetHandler.RESET_URL,
                                          self.resetHandler,
                                          [Resource.Method.PUT])
            resourceMessage.register(resource)

            # TODO: handle error
            self.queue([resourceMessage])
        except Exception as e:
            # SEVERE
            self.getLogger().critical(f'{traceback.format_exc()} {e}')

        # Do this last after everything else is established.
        # Populate outgoing message queue from persisted messages, but leave the
        # messages in persistence. The messages are removed from persistence when
        # they are delivered successfully.
        # TODO: IOT-49043
        messagePersistence = MessagePersistence.getInstance()
        if messagePersistence is not None:
            messages = messagePersistence.load(self.deviceClient.getEndpointId())
            if messages is not None and len(messages) > 0:
                # Delete the persisted messages one by one in case of
                # overflow
                with self.sendLock:
                    for message in messages:
                        # In the constructor there shouldn't be anybody trying
                        # to add to the queue so don't block. But catch the
                        # exception if it overflows.
                        # What should be do with the excess messages ?
                        try:
                            self.outgoingMessageQueue.put(message, block=False)
                            messagePersistence.delete([message])
                            self.incrementQueueCapacity(-1)
                        except Exception as e:
                            self.getLogger().error(f'__init__(): too many persisted messages.')
                            break

        #Start the receive thread first in order to receive any pending
        #requests messages when messages are first transmitted.
        self.receiveThread = self.Receiver(self)
        self.receiveThread.start()

        self.transmitThread = self.Transmitter(self)
        self.transmitThread.start()

    def getQueueSize(self):
        return self.maximumQueueSize

    def getMaximumMessagesPerConnection(self):
        return self.maximumMessagesPerConnection

    def getPollingInterval(self):
        with self.pollingIntervalLock:
            return self.pollingInterval

    def setPollingInterval(self, value):
        with self.pollingIntervalLock:
            self.pollingInterval = value

    def incrementQueueCapacity(self, increment):
        with self.queueCapacityLock:
            self.queueCapacity += increment

    def getQueueCapacity(self):
        with self.queueCapacityLock:
            return self.queueCapacity

    def getWaitOnReconnect(self):
        with self.waitOnReconnectLock:
            return self.waitOnReconnect

    def setWaitOnReconnect(self, value):
        with self.waitOnReconnectLock:
            self.waitOnReconnect = value

    def setWaitOnReconnectAndNotify(self, value):
        with self.waitOnReconnectLock:
            self.waitOnReconnect = value
            self.waitOnReconnectLock.notify_all()

    def getRequestDispatcher(self):
        return self.requestDispatcher

    def addStorageObjectDependency(self, storageObject, clientMsgId):
        with self.contentLock:
            if storageObject not in self.contentMap:
                self.contentMap[storageObject] = set()
            self.contentMap[storageObject].add(clientMsgId)

    def removeStorageObjectDependency(self, storageObject):
        completed = storageObject.getSyncStatus() == StorageObject.SyncStatus.IN_SYNC
        ids = None
        with self.contentLock:
            if storageObject in self.contentMap:
                ids = self.contentMap.pop(storageObject)
            if not completed and ids is not None:
                self.failedContentIds.update(ids)

    def isContentDependent(self, clientId):
        with self.contentLock:
            sets = self.contentMap.values()
            for s in sets:
                if clientId in s:
                    return True

        return False

    def queue(self, messages):
        if messages is None:
            raise ArgumentException(f'MessageDispatcherImpl.queue(): messages is None')
        if len(messages) == 0:
            # TODO: Not sure why an empty message list is an exception.
            # TODO: Maybe log a warning
            raise ArgumentException(f'MessageDispatcherImpl.queue(): messages is empty')

        with self.sendLock:
            qcap = self.getQueueCapacity()
            if qcap < len(messages):
                raise StateException(f'MessageDispatcherImpl.queue(): '
                    f' queue is full need {len(messages)} have {qcap}')
            for message in messages:
                # TODO: Should we block and wait for some timeout length ?
                # TODO: If we let the Queue manage put and get with blocking
                # TODO: it may reduce the amount of bookeeping and locks
                # TODO: to keep in sync
                try:
                    self.outgoingMessageQueue.put(message, block=False)
                except Exception as e:
                    # Could get Full here if queueCapacity is messed up
                    self.getLogger().warning(f'queue(): got Full execption, but should not have.')
            self.messageQueued.notify_all()


    def offer(self, messages):
        """
        Offer messages to be queued.

        Depending on the policies, if any,
        the messages will be queued if it is possible to do so without
        violating capacity restrictions.

        :param messages: a `list` of messages to be queued
        :raises: :class:`.ArgumentException` if `messages` is `None` or empty
        """
        if messages is None or len(messages) == 0:
            raise ArgumentException("messages is None or empty")


        persistenceStore = PersistenceStoreManager.getPersistenceStore(
            self.deviceClient.getEndpointId())
        messagingPolicy = persistenceStore.get(MessagingPolicy.__name__, None)
        if messagingPolicy is None:
            messagingPolicy = MessagingPolicy(self.deviceClient)
            persistenceStore.openTransaction()\
                .put(MessagingPolicy.__name__, messagingPolicy)\
                .commit()
            devicePolicyManager = DevicePolicyManager.getDevicePolicyManager(
                self.deviceClient)
            devicePolicyManager.addChangeListener(messagingPolicy)

        messagesToQueue = []
        try:
            for message in messages:
                messagesFromPolicy = messagingPolicy.applyPolicies(message)
                if messagesFromPolicy is not None:
                    messagesToQueue.extend(messagesFromPolicy)
        except Exception as e:
            # TODO: retry ?
            self.getLogger().warning(f'offer(): {e}')
            self.dispatchError(messages, e)

        if len(messagesToQueue) > 0:
            self.queue(messagesToQueue)

        return

    def setOnDelivery(self, onDelivery):
        with self.deliveryCallbackLock:
            self.deliveryCallback = onDelivery

    def setOnError(self, onError):
        with self.errorCallbackLock:
            self.errorCallback = onError

    def close(self):
        with self.closeLock:
            # For calls to isClosed
            self.closed = True
            # Threads in this file check requestClose
            self.requestClose = True

        try:
            if self.receiveThread is not None:
                self.receiveThread._stop()
                # might be blocked on waitOnReconnect
                with self.waitOnReconnectLock:
                    self.waitOnReconnectLock.notify_all()
                # might also be blocked on messageSent
                with self.messageSent:
                    self.messageSent.notify_all()
                # close the secure connection since it may be in a long
                # polling block
                # TODO: Can this raise exception ?
                sc = self.deviceClient.secureConn
                sc.close()
                # TODO: Is repeating this necessary ?
                # TODO: It shouldn't be.
                # might be blocked on waitOnReconnect
                with self.waitOnReconnectLock:
                    self.waitOnReconnectLock.notify_all()
                # might also be blocked on messageSent
                with self.messageSent:
                    self.messageSent.notify_all()
                try:
                    self.getLogger().debug(msg = f'close(): Joining Receive thread')
                    self.receiveThread.join(timeout=self.JOINTIMEOUT)
                except Exception as e:
                    self.getLogger().debug(f'close(): receiveThread.join() raised exception {e}')
                self.getLogger().debug(msg = f'close(): Receive thread closed')

            # Since we are using the Queue blocking, this causes the
            # Transmit thread to stop waiting on the queue.
            try:
                self.outgoingMessageQueue.put(None, block=False)
            except Exception as e:
                pass

            if self.transmitThread is not None:
                self.transmitThread._stop()
                if self.transmitThread.is_alive():
                    self.getLogger().debug(msg = f'close(): Joining transmit thread')
                    try:
                        self.transmitThread.join(self.JOINTIMEOUT)
                    except Exception as e:
                        self.getLogger().debug(f'close(): transmitThread.join() raised exception {e}')
                self.getLogger().debug(msg = f'close(): Transmitter thread closed')

        # TODO: raised on EINTR why do we care ?
        # TODO: Won't python shut down the main thread itself ?
        # TODO: Do we want to get into OS signal handling ?
        # TODO: It would probably be better to provide the application
        # TODO: with a shutdown handler and let it deal with signals.
        except InterruptedError:
            pass

    def isClosed(self):
        """Return `True` if message dispatcher has been closed."""
        with self.closeLock:
            return self.closed #or self.requestClose

    def isRequestClose(self):
        with self.closeLock:
            return self.requestClose

    def dispatchError(self, messageList, exception):
        """
        Dispatch a messaging error to :class:`.MessageDispatcher.ErrorCallback`.  # noqa: D400,E501

        :param messageList: list of messages with errors
        :param exception: an `Exception` that caused the error.
        """
        errorcallback = None
        with self.errorCallbackLock:
            errorcallback = self.errorCallback

        if errorcallback is None:
            return

        errorDispatcher = MessageDispatcherImpl.ErrorDispatcher(
                    errorcallback, messageList, exception)
        errorDispatcher.run()

    def dispatchDeliveryCallback(self, messages):
        with self.deliveryCallbackLock:
            if self.deliveryCallback is not None:
                self.deliveryCallback.delivered(messages)

    def getLogger(self, fromclass=None):
        """
        Return the *iotcs.device.impl.MessageDispatcherImpl* logger.

        :param fromclass: add `fromclass` to the logger name.
        """
        loggername = 'iotcs.client.device.util.MessageDispatcherImpl'
        if fromclass is not None:
            loggername = f'{loggername}.{fromclass}'
        return logging.getLogger(loggername)

    def __getResource(self, endpointId, name, requestHandler, methods):
        self.requestDispatcher.registerRequestHandler(endpointId, name,
                                                      requestHandler)
        resourceMessage = Resource(name=name, path=name,
                                   endpointName=endpointId, methods=methods)
        return resourceMessage
