# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.


import base64
import collections
import datetime
from enum import IntEnum
import hashlib
import hmac
import logging
import math
import numbers
from random import randrange
import sys
import time


from iotcs.client.model import DeviceModelAttribute
from iotcs.exception import ArgumentException
from iotcs.exception import StateException
from iotcs.message import AlertMessage
from iotcs.messaging.client.device.persistence import BatchByPersistence
from iotcs.messaging.policy import FormulaParser
from iotcs.messaging.policy import Formula
from iotcs.messaging.policy import ValueProvider
from iotcs.common import Pair
from iotcs.common import TimeManager
from iotcs.common import Util

# com.oracle.iot.client.impl.device

class DeviceFunction():
    """DeviceFunction is an abstraction of a policy device function."""

    ApplyBucketData = collections.namedtuple('ApplyBucketData',
        [ 'buckets', 'bucket' ])

    GetBucketData = collections.namedtuple('GetBucketData',
        [ 'bucketsPerWindow', 'bucketsPerSlide',
          'bucketZero', 'buckets'])

    @classmethod
    def getApplyBucketData(cls, id, configuration, data, defaultbucketvalue):
        """
        Return a :class:`.ApplyBucketData` tuple.

        An `ApplyBucketTuple` is::
            [ 'buckets', 'bucket' ]

        :param id: function name prefix
        :param configuration:
        :param data:
        :param defaultbucketvalue:
        :return:
        """
        #  Handling slide:
        #
        # Slide is how much the window moves after the
        # window expires. If there is a window of 5 seconds
        # with a slide of 2 seconds, then at the end of
        # 5 seconds, the window slides over by two seconds.
        # That means that the next window's worth of data
        # would include 3 seconds of data from the previous
        # window, and 2 seconds of new data.
        #
        # To handle this, we divide up the window into buckets.
        # Each bucket represents a period of time, such
        # that the time period is the greatest common factor
        # between the window and the slide. For example, if
        # the window is 60 seconds and the slide is 90
        # seconds, a bucket would span 30 seconds, and
        # there would be three buckets.
        #
        # When the window expires and the get method is called,
        # the return value of the mean policy function will
        # include the value and number of terms of bucket[0]
        # through bucket[n]. Then the buckets that don't
        # contribute to the next window are emptied (so to speak)
        # and the cycle continues.
        #
        # Best case is that the slide equal to the window.
        # In this case, there is only ever one bucket.
        # The worst case is when greatest common factor between
        # slide and window is small. In this case, you end up
        # with a lot of buckets, potentially one bucket per
        # slide time unit (e.g., 90 seconds, 90 buckets).
        # But this is no worse (memory wise) than keeping
        # an array of values and timestamps.
        #
        now = TimeManager.currentTimeMillis()

        # determine if apply has been called for a function
        if f'{id}.applied' not in data:
            data[f'{id}.applied'] = False

        # windowStartTime is the time at which the first
        # call to apply was made for the current window
        # of time. We need to know when this window
        # started so that we can figure out what bucket
        # the data goes into.
        windowStartTime = None
        if f'{id}.windowStartTime' in data:
            windowStartTime = int(data[f'{id}.windowStartTime'])
        if windowStartTime is None:
            windowStartTime = now
            data[f'{id}.windowStartTime'] =  windowStartTime

        # The greatest common factor between the
        # window and the slide represents the greatest
        # amount of time that goes evenly into
        # both window and slide.
        window = cls.getWindow(configuration)
        slide = cls.getSlide(configuration, window)
        span = cls.gcd(window, slide)
        buckets = None
        if f'{id}.buckets' in data:
            buckets = data[f'{id}.buckets']
        if buckets is None:
            # The number of buckets is the window or span
            # (which ever is greater) divided
            # by the amount of time it spans. Thus, if
            # there is a 5 second window with a 2 second slide,
            # the greatest common factor is 1 second and we end
            # up with 5 buckets. But if the slide was 7 seconds,
            # you'd end up with 7 buckets. This lets us fill
            # up buckets without worrying about whether the
            # window is greater than, equal to, or less than
            # the slide.
            # Note: we add 1 so there is a bucket for when
            # a value comes in for the next window, but before
            # the window has been moved.
            numberOfBuckets = int((max(slide, window) / span) + 1)
            buckets = [None for i in range(numberOfBuckets)]
            for i in range(numberOfBuckets):
                if callable(defaultbucketvalue):
                    defaultbucketvalue = defaultbucketvalue()
                buckets[i] = DeviceFunction.Bucket(defaultbucketvalue)

            data[f'{id}.buckets'] =  buckets

        bucketZero = None
        if f'{id}.bucketZero' in data:
            bucketZero = int(data[f'{id}.bucketZero'])
        if bucketZero is None:
            bucketZero = 0
            data[f'{id}.bucketZero'] = bucketZero

        # Which bucket are we working on is calculated
        # by the dividing the amount of time we are into
        # the window by the span of time represented by
        # one bucket. For example, say we have a 2 second
        # slide and a 10 second window giving us 5 buckets.
        # Say our window started at 20 seconds and the
        # value arrives at 25 seconds (5 seconds into the window).
        # The value, then should be added to the
        # third bucket (buckets[2]) since that bucket
        # represents the time from 4 seconds to 6 seconds
        # into the current window.
        bucketIndex = int(((now - windowStartTime) / span))
        bucket = (bucketZero + bucketIndex) % len(buckets)

        return cls.ApplyBucketData(buckets=buckets, bucket = bucket)

    @classmethod
    def getGetBucketData(cls, id, configuration, data):
        """
        Return a :class:`.BucketData` tuple.

        A `GetBucketTuple` is::
            ['bucketsPerWindow', 'bucketsPerSlide',
             'bucketZero', 'buckets']

        :param id: function name prefix
        :param configuration:
        :param data:
        :return:
        """
        buckets = None
        if f'{id}.buckets' in data:
            buckets = data[f'{id}.buckets']
        if buckets is None:
            return None
        bucketZero = None
        if f'{id}.bucketZero' in data:
            bucketZero = int(data[f'{id}.bucketZero'])
        #assert bucketZero is not None
        if bucketZero is None:
            # if buckets is not null, but bucketZero is,
            # something is wrong with our implementation
            return None

        # The greatest common factor between the
        # window and the slide represents the greatest
        # amount of time that goes evenly into
        # both window and slide.
        window = cls.getWindow(configuration)
        slide = cls.getSlide(configuration, window)

        # Each bucket spans this amount of time.
        span = cls.gcd(window, slide)

        # The number of buckets that make up a window
        bucketsPerWindow = int(window / span)

        # The number of buckets that make up the slide
        bucketsPerSlide = int(slide / span)

        # update windowStartTime for the next window.
        # The new windowStartTime is just the current window
        # start time plus the slide.
        windowStartTime = None
        if f'{id}.windowStartTime' in data:
            windowStartTime = int(data[f'{id}.windowStartTime'])
        #assert windowStartTime is not None
        if windowStartTime is None:
            windowStartTime = TimeManager.currentTimeMillis()
        data[f'{id}.windowStartTime'] = \
            windowStartTime + span * bucketsPerSlide

        # update bucketZero index. bucketZero is the index
        # of the zeroth bucket in the circular buckets array.
        data[f'{id}.bucketZero'] = \
            (bucketZero + bucketsPerSlide) % len(buckets)

        return cls.GetBucketData(bucketsPerWindow=bucketsPerWindow,
                             bucketsPerSlide=bucketsPerSlide,
                             bucketZero=bucketZero, buckets=buckets)

    class NetworkCost(IntEnum):
        """In the order of increasing cost."""
        ETHERNET = 0
        CELLULAR = 1
        SATELLITE = 2

        @classmethod
        def getCost(cls, value, property, defaultValue):
            """
            Return the :class:`.NetworkCost` of value.

            :param value: the network cost as a `str`
            :param property: used for logging
            :param defaultValue: the default :class:`.NetworkCost` value
            :return: the :class:`.NetworkCost` of value
            """
            networkCost = None
            if value is not None:
                try:
                    upperValue = value.upper()
                    networkCost = DeviceFunction.NetworkCost[upperValue]
                except Exception as e:
                    DeviceFunction.getLogger().warning(
                        f'NetworkCost.getCost(): '
                        f'invalid "{property}" value: "{value}"')
            if networkCost is None:
                networkCost = defaultValue
                DeviceFunction.getLogger().warning(
                    f'NetworkCost.getCost(): '
                    f'defaulting "{property}" to: "{networkCost}"')
            return networkCost

    class Bucket(object):
        """
        This class is used to implement the *window slide* functionality.

        See :meth:`.mean`
        """
        __slots__ = (
            'value',
            'terms'
        )

        def __init__(self, initialValue):
            """Create an instance of :class:`.Bucket`"""
            self.value = initialValue
            self.terms = 0

        def __str__(self):
            """Return a string representation for this :class:`.Bucket`. """
            return f'{{"value" : {self.value}, "terms" : {self.terms}}}'


    __slots__ = (
        'id'
    )

    ZERO = 0.0
    DEVICE_FUNCTIONS = None
    inProcessValues = dict()
    # A map of attribute-name -> in-process attribute value.
    # The in-process value of an attribute is the value that is being
    # passed along the pipeline; i.e., $(attribute) in a formula or condition

    def __init__(self, id):
        """Create an instance of :class:`.DeviceFunction`."""
        self.id = id

    def apply(self, deviceAnalog, attribute, configuration, data, value):
        """
        Perform the function.

        This method returns `True` if the conditions for the function have
        been met. Only when this function's apply method returns true
        will the next function in the pipeline be applied.

        After this method returns `True`, use
        :meth:`.get` to retrieve
        the value from the function.

        :param deviceAnalog: the :class:`.VirtualDevice`, never `None`
        :param attribute: the :class:`.DeviceModelAttribute`, which may be
            `None` if the function is being applied at the device model level
        :param configuration: the parameters for this function from the device
            policy
        :param data: a place for storing data between invocations of the
            function
        :param value: the value to which the function is being applied
        :return: `True` if the conditions for the function have been satisfied.
        """
        pass

    def get(self, deviceAnalog, attribute, configuration, data):
        """
        Return the value from the function.

        This method should only be called after :meth:`.apply` returns `True`,
        or when a window expires.

        :param deviceAnalog: the :class:`.VirtualDevice`, never `None`
        :param attribute: the :class:`.DeviceModelAttribute`, which may be
            `None` if the function is being applied at the device model level
        :param configuration: the parameters for this function from the device
            policy
        :param data: a place for storing data between invocations of the
            function
        :return: the value from having applied the function
        """
        pass

    @classmethod
    def getWindow(cls, configuration):
        """
        Return a *window* value from a configuration.

        :param configuration: the parameters for this function from the device
            policy
        :return: a window value, or -1 if the configuration is not time based
        """
        for key in ['window', 'delayLimit']:
            criterion = None
            if key in configuration:
                criterion = configuration[key]
            if criterion is not None:
                return int(criterion)

        return -1

    @classmethod
    def getSlide(cls, configuration, window):
        """
        Return a *slide* value from a configuration.

        :param configuration: the parameters for this function from the device
            policy
        :param window the corresponding window for the slide
        :return: the configured slide value, or window if there is no slide or
            slide is zero
        """
        slideParameter = None
        if "slide" in configuration:
            slideParameter = configuration["slide"]
        if slideParameter is not None:
            slide = int(slideParameter)
            return slide if slide > 0 else window
        return window

    def getId(self):
        """
        Return the function id.

        This is the unique id from the function definition.
        :return: the function id
        """
        return self.id

    def __str__(self):
        """
        Return a string representation of this function.

        :return: the function id
        """
        return self.getId()

    def getDetails(self, configuration):
        """
        Return a string representation of this function.

        Useful for logging.
        :param configuration: the parameters for this function from the device
            policy
        :return: a string representation of this function
        """
        return self.getId()

    @classmethod
    def getDeviceFunction(cls, functionId):
        """
        Return the :class:`.DeviceFunction` for this `functionId`.

        :param functionId: the function id
        :return: the :class:`.DeviceFunction` for `functionId`
        """
        if functionId in _POLICY_MAP:
            return _POLICY_MAP[functionId]
        else:
            return None

    @classmethod
    def createInProcessMapKey(cls, endpointId, deviceModelURN, attribute):
        """
        Return the *key* to retrieve a value from `inProcessValues`.

        The format is::

            <endpointId>/deviceModels/<deviceModelURN>:attributes/<attribute>

        :param endpointId: a device endpoint id
        :param deviceModelURN: a device model urn
        :param attribute: a device model attribute name
        :return: a key to return a value from `inProcessValues`
        """
        return f'{endpointId}/deviceModels/{deviceModelURN}:attributes/{attribute}'

    @classmethod
    def putInProcessValue(cls, endpointId, deviceModelURN, attribute, value):
        """
        Add `attribte` and `value` to `inProcessValues`.

        :param endpointId: a device endpoint id
        :param deviceModelURN: a device model urn
        :param attribute: a device model attribute name
        :param value: the attribute value
        """
        cls.inProcessValues[cls.createInProcessMapKey(
            endpointId, deviceModelURN, attribute)] = value

    @classmethod
    def removeInProcessValue(cls, endpointId, deviceModelURN, attribute):
        """
        Return and remove `attribute` value from `inProcessValues`.

        :param endpointId: a device endpoint id
        :param deviceModelURN: a device model urn
        :param attribute: a device model attribute name
        :return: the attribute value
        """
        value = None
        key = cls.createInProcessMapKey(endpointId, deviceModelURN, attribute)
        if key in cls.inProcessValues:
            value = cls.inProcessValues[key]
            del(cls.inProcessValues[key])
            return value
        return value

    @classmethod
    def getInProcessValue(cls, endpointId, deviceModelURN, attribute):
        """
        Return the in process value for `attribute`.

        :param endpointId: a device endpoint id
        :param deviceModelURN: a device model urn
        :param attribute: a device model attribute name
        :return: the in process attribute value or `None`

        """
        key = cls.createInProcessMapKey(endpointId, deviceModelURN, attribute)
        if key in cls.inProcessValues:
            return cls.inProcessValues[key]
        else:
            return None

    @classmethod
    def getActionArgs(cls, deviceAnalog, configuration):
        """
        Return a `list` of action arguments.

        The returned `list` of action arguments maps to the list of
        arguments in the configuration. The list  will contain all `None`
        entries if there is no device model action map, or the action map is
        empty.

        :param deviceAnalog: a representation of a :class:`.VirtualDevice`
        :param configuration: the device function parameters
        :return: a `dict` of action arguments.
        """
        if "arguments" not in configuration:
            return None

        arguments = configuration["arguments"]
        if arguments is None or len(arguments) == 0:
            return None
        # pre fill the action args with `None` for error cases
        #actionargs = [None for i in range(len(arguments))]
        actionargs = dict()

        # TODO: Note JCL does all this work in every iteration over the
        # TODO: arguments list, no need to do that in a loop.

        devicemodel = deviceAnalog.getDeviceModel()
        actionmap = devicemodel.getDeviceModelActions()
        if actionmap is None or len(actionmap) == 0:
            cls.getLogger().warning(f'getActionArgs(): '
                f'no actions in device model {devicemodel.getURN()}')
            # Just return a "None" filled list
            return actionargs

        actionname = None
        if "name" in configuration:
            actionname = str(configuration["name"])
        if actionname is None:
            cls.getLogger().warning(f'getActionArgs(): '
                                    f'no action "name" in configuration')
            return actionargs

        devicemodelaction = None
        if actionname in actionmap:
            devicemodelaction = actionmap[actionname]
        if devicemodelaction is None:
            cls.getLogger().warning(f'getActionArgs(): no action named '
                f'{actionname} in device model {devicemodel.getURN()}')
            return actionargs

        dmaArgumentList = devicemodelaction.getArguments()
        for n, argument in enumerate(arguments):
            # TODO: There is an assumption that the order of "arguments" is the same
            # TODO: as the DeviceModelAction arguments list.
            argtype = dmaArgumentList[n].getArgType()
            argname = dmaArgumentList[n].getName()
            if argname is None:
                # This is for the case where there is a legacy model.
                argname = "value"
            try:
                # if argType is None, convertArg just returns arg
                actionargs[argname] = cls.convertArg(deviceAnalog, argtype, argument)
            except ArgumentException as e:
                cls.getLogger().warning(f'getActionArgs(): '
                    f'bad argument to "{actionname}" in "{devicemodel.getURN()}" : {e}')
                actionargs[argname] = argument

        return actionargs

    @classmethod
    def convertArg(cls, deviceAnalog, argtype, arg):
        """
        Return the argument value.
        arg in an ACTION is a formula
        ACTION type is represented as JSON of form::
            {
                 "name": "name-of-the-action",
                 "arguments": [
                     "formula:argument-1",
                    "formula:argument-2"
                 ]
            }

        field in an ALERT is also a formula
        ALERT type is represented as JSON of form::
            {
                "urn": "URN-of-the-alert",
                "fields": {
                    "name-of-the-field-1": "<formula>",
                    "name-of-the-field-2": "<formula>"
                    ...
                 },
                "severity": "severity level"
            }

        :param deviceAnalog: a representation of a :class:`.VirtualDevice`
        :param argtype: a :class:`.DeviceModelAttribute.Type`
        :param arg: the argument to convert
        :return: the converted argument value or `None`
        :raises: :class:`.ArgumentException` if `arg` cannot be coverted or
            `argtype` is an invalid type.
        """
        if arg is None:
            return None
        conversion = cls.convertFormula(deviceAnalog, arg.__str__())
        if argtype == DeviceModelAttribute.Type.STRING or \
                argtype == DeviceModelAttribute.Type.URI:
            return str(conversion)
        elif argtype == DeviceModelAttribute.Type.BOOLEAN:
            if isinstance(conversion, (bool, )):
                return conversion
            elif isinstance(conversion, (str, )):
                return Util.strtobool(conversion)
            elif isinstance(conversion, (numbers.Number, )):
                return float(conversion) != 0.0
            else:
                raise ArgumentException(
                    f'convertArg(): cannot convert {arg} to {argrtype.name()}')
        elif argtype == DeviceModelAttribute.Type.DATETIME:
            if isinstance(conversion, (numbers.Number, )):
                # TODO: Should this return a pythong datetime ?
                return int(conversion)
            else:
                raise ArgumentException(
                    f'convertArg(): cannot convert {arg} to {argrtype.name()}')
        elif argtype == DeviceModelAttribute.Type.INTEGER:
            if isinstance(conversion, (numbers.Number, )):
                return int(conversion)
            else:
                raise ArgumentException(
                    f'convertArg(): cannot convert {arg} to {argrtype.name()}')
        elif argtype == DeviceModelAttribute.Type.NUMBER:
            if isinstance(conversion, (numbers.Number, )):
                return conversion
            else:
                raise ArgumentException(
                    f'convertArg(): cannot convert {arg} to {argrtype.name()}')
        else:
            cls.getLogger().warning(
                f'convertArg(): unknown type : {argtype.name()}')
            return arg

    @classmethod
    def convertFormula(cls, deviceAnalog, formula):
        """
        Return a value for `formula`.

        :param deviceAnalog: a representation of a :class:`.VirtualDevice`
        :param formula: a formula
        :return: return the formula value else `float('nan')`
        """
        try:
            tokens = FormulaParser.tokenize(formula)
            node = FormulaParser.parseFormula(tokens, formula)
            return cls.compute(node, deviceAnalog)
        except ArgumentException as e:
            cls.getLogger().warning(f'convertFormula(): '
                f'field in formula {formula} not in device model')
        return float('nan')

    @classmethod
    def createAlert(cls, deviceAnalog, configuration):
        """
        Return an :class:`.AlertMessage` instance.

        :param deviceAnalog: a representation of a :class:`.VirtualDevice`
        :param configuration: the parameters for this function from the device
            policy
        :return: a :class:`.AlertMessage` instance
        :raises: :class:`.ArgumentException` if the format is an invalid
            alert format.
        """
        deviceModel = deviceAnalog.getDeviceModel()
        deviceModelFormatMap = deviceModel.getDeviceModelFormats()
        if deviceModelFormatMap is None:
            raise ArgumentException(f'createAlert(): '
                f'{deviceModel.getURN()} does not contain alert formats')
        format = None
        if "urn" in configuration:
            format = str(configuration["urn"])
        if format is None:
            raise ArgumentException(f'createAlert(): '
                f'no "urn" in configuration.')
        deviceModelFormat = None
        if format in deviceModelFormatMap:
            deviceModelFormat = deviceModelFormatMap[format]
        if deviceModelFormat is None:
            raise ArgumentException(f'createAlert(): '
                f'{deviceModel.getURN()} does not contain alert format "{format}"')

        alertSeverity = AlertMessage.Severity.NORMAL
        try:
            severityConfig = None
            if "severity" in configuration:
                severityConfig = str(configuration["severity"])
                if severityConfig is not None:
                    alertSeverity = AlertMessage.Severity[severityConfig.upper()]
        except ArgumentException as e:
            pass

        alertmessage = AlertMessage(source=deviceAnalog.getEndpointId(),
            format=format, description=deviceModelFormat.getName(),
            severity=alertSeverity)

        fieldsFromPolicy = None
        if "fields" in configuration:
            fieldsFromPolicy = configuration.get("fields")
        else:
            return alertmessage

        fields = deviceModelFormat.getFields()
        for field in fields:
            policyValue = None
            if field.getName() in fieldsFromPolicy:
                policyValue = fieldsFromPolicy[field.getName()]
            if policyValue is None:
                continue
            try:
                value = cls.convertArg(deviceAnalog, field.getType(), policyValue)
                cls.addDataItem(alertmessage, field, value)
            except ArgumentException as e:
                cls.getLogger().warning(f'createAlert(): '
                    f'bad value for "{field.getName()}" in "{deviceModel.getURN()}" : {e}')

        return alertmessage

    @classmethod
    def addDataItem(cls, message, field, value):
        """
        Update the `message` with `field` and `value`.

        :param message: message to update
        :param field: data item field
        :param value: data item value
        :raises: :class:`.ArgumentException` if `value` is not a valid type.
        """
        if field.getType() == DeviceModelAttribute.Type.INTEGER or \
                field.getType() == DeviceModelAttribute.Type.NUMBER:
            if isinstance(value, (numbers.Number, )):
                if field.getType() == DeviceModelAttribute.Type.INTEGER:
                    message.dataItem(field.getName(), int(value))
                else:
                    message.dataItem(field.getName(), float(value))
            else:
                raise ArgumentException(f'addDataItem(): '
                    f'value of attribute "{field.getName()}" is not a "{field.getType()}"')
        elif field.getType() == DeviceModelAttribute.Type.URI or \
                field.getType() == DeviceModelAttribute.Type.STRING:
            if isinstance(value, (str, )):
                message.dataItem(field.getName(), str(value))
            else:
                raise ArgumentException(f'addDataItem(): '
                    f'value of attribute "{field.getName()}" is not a "{field.getType()}"')
        elif field.getType() == DeviceModelAttribute.Type.BOOLEAN:
            if isinstance(value, (bool, )):
                message.dataItem(field.getName(), bool(value))
            else:
                raise ArgumentException(f'addDataItem(): '
                    f'value of attribute "{field.getName()}" is not a "{field.getType()}"')
        elif field.getType() == DeviceModelAttribute.Type.DATETIME:
            if isinstance(value, (numbers.Number, )):
                message.dataItem(field.getName(), int(value))
            elif isinstance(value, (datetime.datetime, )):
                message.dataItem(field.getName(), time.mktime(value.timetuple()))
            else:
                raise ArgumentException(f'addDataItem(): '
                    f'value of attribute "{field.getName()}" is not a "{field.getType()}"')
        else:
            raise ArgumentException(f'addDataItem(): '
                f'the type of attribute "{field.getName()}" '
                f'is not a known type ("{field.getType()}")')

    @classmethod
    def compute(cls, node, deviceAnalog):
        """
        Return the computed value of `node`.

        Delegates to :meth:`.Formula.compute`

        :param node: a :class:`.FormulaParserNode` to compute
        :param deviceAnalog: a representation of a :class:`.VirtualDevice`
        :return: the computed value.
        """
        return Formula.compute(ValueProviderImpl(deviceAnalog), node=node)

    @classmethod
    def gcd(cls, x, y):
        """Return the greatest common divisor of `x` and `y`."""
        return x if (y == 0) else math.gcd(y, x % y)

    @classmethod
    def getLogger(cls):
        """Return a logger for *iotcs.messaging.client.device.DeviceFunction. """
        return logging.Logger("iotcs.messaging.client.device.DeviceFunction")

    @classmethod
    def getPersistedBatchedData(cls, batchByPersistence, deviceAnalog):
        """
        Return persisted batch by data.

        :param batchByPersistence: the persisteence engine
        :param deviceAnalog: a representation of a :class:`.VirtualDevice`
        :return: persisted batch by date else `None` if not persisting data
        """
        # TODO: Implement persistence
        if batchByPersistence is None:
            return None

        messages = batchByPersistence.load(deviceAnalog.getEndpointId())
        batchByPersistence.delete(messages)
        pairs = []
        for message in messages:
            pairs.append(Pair(message, None))
        return pairs


    def _windowDetails(self, config):
        """

        :param config:
        :return:
        """
        details = f'['
        comma = ''
        window = None
        if "window" in config:
            window = config["window"]
        if window is not None:
            details = f'window={window}'
            comma = ','
        slide = None
        if "slide" in config:
            slide = config["slide"]
        if slide is not None:
            details = f'{details}{comma}slide={slide}'
        return details

    def _bucketDetails(self, data):
        """

        :param data:
        :return:
        """
        details = ''
        if f'{id}.buckets' in data:
            buckets = data[f'{id}.buckets']
            if buckets is not None:
                bstr = ""
                for bucket in buckets:
                    bstr = f'{bstr}\n{str(bucket)}'
                details = f'{details}{bstr}'
        return details
#
#  Policy definitions
#

class _FilterCondition(DeviceFunction):
    """
    The :class:`.FilterCondition` :class:`.DeviceFunction`

    The outcome of this function determines if the value is processed
    by other functions following this one, in a pipeline of functions.
    """

    def apply(self, deviceAnalog, attribute, configuration, data, value):
        """
        Return `True` if the value should be filtered from the rest of the pipeline.
        :param deviceAnalog: a representation of a :class:`.VirtualDevice`
        :param attribute: the attribute to filter
        :param configuration: the filter configuration
        :param data: the current filter criteria
        :param value: the value to filter
        :return: `True` if `value` should be filtered
        """
        data["filterCondition.value"] = value
        condition = None
        if "filterCondition.condition" in data:
            condition = data.get("filterCondition.condition")
        if condition is None:
            tokenstr = None
            tokens = None
            if "condition" in configuration:
                tokenstr = str(configuration.get("condition"))
                tokens = FormulaParser.tokenize(tokenstr)
            stack = []
            FormulaParser.parseConditionalOrExpression(stack, tokens,
                                                       tokenstr, 0)
            condition = stack.pop()

        data["filterCondition.condition"] = condition
        computedValue = float(self.compute(condition, deviceAnalog))
        return -1.0 < computedValue and computedValue < 1.0

    def get(self, deviceAnalog, attribute, configuration, data):
        """
        Return the function value.

        :param deviceAnalog: a representation of a :class:`.VirtualDevice`
        :param attribute: the attribute to filter
        :param configuration: the filter configuration
        :return: the function value
        """
        value = None
        if "filterCondition.value" in data:
            value = data["filterCondition.value"]
            del(data["filterCondition.value"])

        return value

    def getDetails(self, config):
        """
        Return a string representation of the function.
        :param config: the function parameters
        :return: a string representation of the function
        """
        parentdetails = \
            super(_FilterCondition, self).getDetails(config)
        return f'{parentdetails}[condition="{config.get("condition")}]"'



class _SampleQuality(DeviceFunction):
    """

    """
    #     rand = Random()

    def apply(self, deviceAnalog, attribute, configuration, data, value):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :param value:
        :return:
        """
        data["sample.value"] = value
        terms = None
        if "sample.terms" in data:
            terms = int(data["sample.terms"])
        if terms is None or terms == sys.maxsize:
            terms = 0
        terms += 1
        data["sample.terms"] = terms
        criterion = None
        if "rate" in configuration:
            criterion = int(configuration["rate"])

        # -1 is random, 0 is all (do this for None too)
        if criterion is None or criterion == 0:
            return True
        elif criterion == -1:
            return
        elif criterion == -1:
            return randrange(30) == 0

        return criterion == terms

    def get(self, deviceAnalog, attribute, configuration, data):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :return:
        """
        sample = data.pop("sample.value", None)
        if "sample.terms" in data:
            del(data["sample.terms"])
        return sample

    def getDetails(self, config):
        """

        :param config:
        :return:
        """
        details = super(_SampleQuality, self).getDetails(config)
        rate = None
        if "rate" in config:
            rate = config["rate"]
        if rate is None:
            rate = 0
        details = f'{details}[rate={rate}]'
        return details

class _Mean(DeviceFunction):
    """
    Calculate the simple mean (average) over a sample window.
    """
    def apply(self, deviceAnalog, attribute, configuration, data, value):
        """
        Return `True` if the *mean* value to be retrieved, else `False`

        Apply the *mean* function and if the sample period is expired
        return `True` else `False`
        :param deviceAnalog: a representation of a :class:`.VirtualDevice`
        :param attribute: the device attribute that is being sampled
        :param configuration: the parameters for this function from the device
            policy
        :param data: the accumated data
        :param value: the current value
        :return: `True` if the sample period has expired else `False`
        """
        applybucketdatatuple = self.getApplyBucketData(self.id,
                                                    configuration, data, 0)
        buckets = applybucketdatatuple.buckets
        bucket = applybucketdatatuple.bucket

        number = value
        bvalue = buckets[bucket].value
        buckets[bucket].value = bvalue + float(number)
        buckets[bucket].terms += 1

        # set the applied flag
        data[f'{id}.applied'] = True

        return False

    def get(self, deviceAnalog, attribute, configuration, data):
        """
        Return the *mean* value.

        :param deviceAnalog: a representation of a :class:`.VirtualDevice`
        :param attribute: the device attribute that is being sampled
        :param configuration: the parameters for this function from the device
            policy
        :param data: the accumated data
        :return: the *mean* value
        """
        getbucketdatatuple = self.getGetBucketData(self.id, configuration,
                                                   data)
        if getbucketdatatuple is None:
            return None
        bucketsPerWindow = getbucketdatatuple.bucketsPerWindow
        bucketsPerSlide = getbucketdatatuple.bucketsPerSlide
        bucketZero = getbucketdatatuple.bucketZero
        buckets = getbucketdatatuple.buckets

        # Loop through the number of buckets in the
        # window and sum them up.
        sum = float(0.0)
        terms = int(0)
        for i in range(bucketsPerWindow):
            index = (bucketZero + i) % len(buckets)
            bucket = buckets[index]
            sum = sum + bucket.value
            terms += bucket.terms

        # Now slide the window
        for i in range(bucketsPerSlide):
            bucket = buckets[(bucketZero + i) % len(buckets)]
            bucket.value = float(0.0)
            bucket.terms = 0

        # reset the applied flag
        data[f'{id}.applied'] = False

        if sum == self.ZERO or terms == 0:
            return None

        return sum / terms

    def getDetails(self, config, data=None):
        """
        Return a string representation of the function state.

        :param config: the parameters for this function from the device
            policy
        :return: a string representation of the function state.
        """
        details = super(_Mean, self).getDetails(config)
        details = f'{details}{self._windowDetails(config)}'

        if data is not None:
            details = f'{details}\n{self._bucketDetails(data)}'

        return details

class _Min(DeviceFunction):
    """

    """

    def apply(self, deviceAnalog, attribute, configuration, data, value):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :param value:
        :return:
        """
        applybucketdatatuple = self.getApplyBucketData(self.id,
                                                       configuration,
                                                       data, float('inf'))
        buckets = applybucketdatatuple.buckets
        bucket = applybucketdatatuple.bucket

        newmin = float(value)
        oldmin = buckets[bucket].value
        buckets[bucket].value = min(newmin, oldmin)

        # set the applied flag
        data[f'{id}.applied'] = True

        return False

    def get(self, deviceAnalog, attribute, configuration, data):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :return:
        """

        getbucketdatatuple = self.getGetBucketData(self.id, configuration,
                                                   data)
        if getbucketdatatuple is None:
            return None
        bucketsPerWindow = getbucketdatatuple.bucketsPerWindow
        bucketsPerSlide = getbucketdatatuple.bucketsPerSlide
        bucketZero = getbucketdatatuple.bucketZero
        buckets = getbucketdatatuple.buckets

        # check the applied flag
        if data[f'{id}.applied'] == False:
            return None

        minvalue = float('inf')
        for i in range(bucketsPerWindow):
            index = (bucketZero + i) % len(buckets)
            bucket = buckets[index]
            oldmin = bucket.value
            minvalue = min(oldmin, minvalue)

        # Now slide the window
        for i in range(bucketsPerSlide):
            bucket = buckets[(bucketZero + i) % len(buckets)]
            bucket.value = float('inf')

        # reset the applied flag
        data[f'{id}.applied'] = False

        return minvalue

    def getDetails_(self, config, data=None):
        """

        :param self:
        :param config:
        :return:
        """
        details = super(_Min, self).getDetails(config)
        details = f'{details}{self._windowDetails(config)}'

        if data is not None:
            details = f'{details}\n{self._bucketDetails(data)}'

        return details

class _Max(DeviceFunction):
    """

    """

    def apply(self, deviceAnalog, attribute, configuration, data, value):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :param value:
        :return:
        """
        applybucketdatatuple = self.getApplyBucketData(self.id, configuration,
                                                       data, float('-inf'))
        buckets = applybucketdatatuple.buckets
        bucket = applybucketdatatuple.bucket

        newmax = value
        oldmax = buckets[bucket].value
        buckets[bucket].value = max(oldmax, newmax)

        # set the applied flag
        data[f'{id}.applied'] = True

        return False

    def get(self, deviceAnalog, attribute, configuration, data):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :return:
        """

        getbucketdatatuple = self.getGetBucketData(self.id, configuration,
                                                   data)
        if getbucketdatatuple is None:
            return None
        bucketsPerWindow = getbucketdatatuple.bucketsPerWindow
        bucketsPerSlide = getbucketdatatuple.bucketsPerSlide
        bucketZero = getbucketdatatuple.bucketZero
        buckets = getbucketdatatuple.buckets

        # check the applied flag
        if data[f'{id}.applied'] == False:
            return None

        maxvalue = float('-inf')
        for i in range(bucketsPerWindow):
            index = (bucketZero + i) % len(buckets)
            bucket = buckets[index]
            oldmax = bucket.value
            maxvalue = max(oldmax, maxvalue)

        # Now slide the window
        for i in range(bucketsPerSlide):
            bucket = buckets[(bucketZero + i) % len(buckets)]
            bucket.value = float('-inf')

        # reset the applied flag
        data[f'{id}.applied'] = False

        return maxvalue

    def getDetails(self, config, data=None):
        """

        :param config:
        :param data:
        :return:
        """
        details = super(_Max, self).getDetails(config)
        details = f'{details}{self._windowDetails(config)}'

        if data is not None:
            details = f'{details}\n{self._bucketDetails(data)}'

        return details


class _StandardDeviation(DeviceFunction):
    """

    """
    def apply(self, deviceAnalog, attribute, configuration, data, value):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :param value:
        :return:
        """

        applybucketdatatuple = self.getApplyBucketData(self.id, configuration,
                                                       data, list)
        buckets = applybucketdatatuple.buckets
        bucket = applybucketdatatuple.bucket

        number = float(value)
        buckets[bucket].value.append(number)

        # set the applied flag
        data[f'{id}.applied'] = True

        return False

    def get(self, deviceAnalog, attribute, configuration, data):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :return:
        """
        getbucketdatatuple = self.getGetBucketData(self.id, configuration,
                                                   data)
        if getbucketdatatuple is None:
            return None
        bucketsPerWindow = getbucketdatatuple.bucketsPerWindow
        bucketsPerSlide = getbucketdatatuple.bucketsPerSlide
        bucketZero = getbucketdatatuple.bucketZero
        buckets = getbucketdatatuple.buckets

        # check the applied flag
        if data[f'{id}.applied'] == False:
            return None

        terms = []
        for i in range(bucketsPerWindow):
            index = (bucketZero + i) % len(buckets)
            bucket = buckets[index]
            oldvalues = bucket.value
            terms.extend(oldvalues)

        sum = 0.0
        for n in range(len(terms)):
            d = terms[n]
            sum = sum + d

        mean = sum / len(terms)
        for n in range(len(terms)):
            d = terms[n] - mean
            terms[n] = math.pow(d, 2)

        sum = 0.0
        for n in range(len(terms)):
            d = terms[n]
            sum = sum + d

        mean = sum / len(terms)

        stdDeviation = math.sqrt(mean)

        for i in range(bucketsPerSlide):
            bucket = buckets[(bucketZero + i) % len(buckets)]
            bucket.value.clear()

        # reset the applied flag
        data[f'{id}.applied'] = False

        return stdDeviation

    def getDetails(self, config, data=None):
        """

        :param config:
        :param data:
        :return:
        """

        details = super(_StandardDeviation, self).getDetails(config)
        details = f'{details}{self._windowDetails(config)}'

        if data is not None:
            details = f'{details}\n{self._bucketDetails(data)}'

        return details


class _BatchBySize(DeviceFunction):
    """

    """
    def apply(self, deviceAnalog, attribute, configuration, data, value):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :param value:
        :return:
        """
        # TODO: Implement Persistence
        batchByPersistence = BatchByPersistence.getInstance()
        if batchByPersistence is not None:
            message = (value).getKey()
            batchByPersistence.save([message], deviceAnalog.getEndpointId())
        else:
            values = None
            if "batchBySize.value" in data:
                values = data["batchBySize.value"]
            if values is None:
                values = []
                data["batchBySize.value"] = values
            values.append(value)

        batchCount = None
        if "batchBySize.batchCount" in data:
            batchCount = data["batchBySize.batchCount"]
        if batchCount is None:
            batchCount = 0

        batchCount += 1
        data["batchBySize.batchCount"] = batchCount
        batchSize = None
        if "batchSize" in configuration:
            batchSize = configuration["batchSize"]
        if batchSize is None or batchSize == batchCount:
            return True

        return False

    def get(self, deviceAnalog, attribute, configuration, data):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :return:
        """
        data["batchBySize.batchCount"] = 0
        # TODO: implement BatchByPersistence
        batchByPersistence = BatchByPersistence.getInstance()
        if batchByPersistence is not None:
            value = DeviceFunction.getPersistedBatchedData(batchByPersistence, deviceAnalog)
            return value
        else:
            value = None
            value = data.pop("batchBySize.value", None)
            return value


    def getDetails(self, config, data=None):
        """

        :param config:
        :param data:
        :return:
        """
        details = super(_BatchBySize, self).getDetails(config)
        batchSize = 0
        if "batchSize" in config:
            batchSize = config["batchSize"]

        details = f'{details}[batchSize={batchSize}]'
        return details

class _BatchByTime(DeviceFunction):
    """

    """
    def apply(self, deviceAnalog, attribute, configuration, data, value):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :param value:
        :return:
        """
        # TODO: Implemenet persistence
        batchByPersistence = BatchByPersistence.getInstance()
        if batchByPersistence is not None:
            message = value.getKey()
            batchByPersistence.save([message], deviceAnalog.getEndpointId())
        else:

            if "batchByTime.value" in data:
                values = data["batchByTime.value"]
            if values is None:
                values = []
                data["batchByTime.value"] = values

            values.append(value)
        return False

    def get(self, deviceAnalog, attribute,configuration, data):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :return:
        """
        # TODO: Implement persistence
        batchByPersistence = BatchByPersistence.getInstance()
        if batchByPersistence is not None:
            value = DeviceFunction.getPersistedBatchedData(batchByPersistence, deviceAnalog)
            return value

        else:
            value = None
            if "batchByTime.value" in data:
                value = data.pop("batchByTime.value")
            return value

    def getDetails(self, config, data=None):
        """

        :param config:
        :param data:
        :return:
        """
        details = super(_BatchByTime, self).getDetails(config)
        delayLimit = 0
        if "delayLimit" in config:
            delayLimit = config["delayLimit"]
        details = f'{details}[delayLimit={delayLimit}]'
        return details

class _BatchByCost(DeviceFunction):
    """

    """
    def apply(self, deviceAnalog, attribute, configuration, data, value):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :param value:
        :return:
        """
        batchByPersistence = BatchByPersistence.getInstance()
        if batchByPersistence is not None:
            message = value.getKey()
            batchByPersistence.save([message], deviceAnalog.getEndpointId())
        else:
            values = None
            if "batchByCost.value" in data:
                values = data["batchByCost.value"]
            if values is None:
                values = []
                data["batchByCost.value"] = values
            values.append(value)

        configuredCost = DeviceFunction.NetworkCost.SATELLITE.name
        if "networkCost" in configuration:
            configuredCost = configuration["networkCost"]

        configuredCost = DeviceFunction.NetworkCost.getCost(configuredCost,
                                                            "configured cost",
                                    DeviceFunction.NetworkCost.SATELLITE)

        networkCost = Util.getConfig().get("device_policy", "network_cost",
            fallback=DeviceFunction.NetworkCost.ETHERNET.name)
        networkCost = DeviceFunction.NetworkCost.getCost(networkCost,
                                                         "network cost",
                                        DeviceFunction.NetworkCost.ETHERNET)

        if networkCost > configuredCost:
            return False

        return True

    def get(self, deviceAnalog, attribute,configuration, data):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :return:
        """
        batchByPersistence = BatchByPersistence.getInstance()
        if batchByPersistence is not None:
            value = DeviceFunction.getPersistedBatchedData(batchByPersistence, deviceAnalog)
            return value

        else:
            value = data.pop("batchByCost.value", None)
            return value

    def getDetails(self, config, data=None):
        """

        :param config:
        :param data:
        :return:
        """
        details = super(_BatchByCost, self).getDetails(config)
        networkCost = DeviceFunction.NetworkCost.ETHERNET.name
        if "networkCost" in config:
            networkCost = config["networkCost"]
        details = f'{details}[networkCost={networkCost}]'
        return details

class _EliminateDuplicates(DeviceFunction):
    """

    """
    def apply(self, deviceAnalog, attribute, configuration, data, value):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :param value:
        :return:
        """
        isDuplicate = False
        now = TimeManager.currentTimeMillis()
        lastValue = None
        if "eliminateDuplicates.lastValue" in data:
            lastValue = data["eliminateDuplicates.lastValue"]
        data["eliminateDuplicates.lastValue"] = value

        # If value equals lastValue, then this is a duplicate value.
        # If value is the first duplicate value, then lastValue has already
        # been passed along and we want to filter out the current value and
        # all other sequential duplicates within the window.
        if value == lastValue:
            windowEnd = 0
            # windowEnd is the end time of the current window.
            if "eliminateDuplicates.windowEnd" in data:
                windowEnd = data["eliminateDuplicates.windowEnd"]
            #assert windowEnd is not None
            # If the current window has not expired (i.e., now <= windowEnd),
            # then the value is filtered out.
            isDuplicate = (now <= windowEnd)

            # If the current window has expired (i.e., windowEnd <= now),
            # then update windowEnd.
            if windowEnd <= now:
                window = self.getWindow(configuration)
                data["eliminateDuplicates.windowEnd"] = \
                        now + (window if window > 0 else 0)
        else:
            window = self.getWindow(configuration)
            data["eliminateDuplicates.windowEnd"] = \
                    now + (window if window > 0 else 0)
        return not isDuplicate

    def get(self, deviceAnalog, attribute, configuration, data):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :return:
        """
        lastValue = None
        if "eliminateDuplicates.lastValue" in data:
            lastValue = data["eliminateDuplicates.lastValue"]
        return lastValue

    def getDetails(self, config, data=None):
        """

        :param config:
        :param data:
        :return:
        """
        details = super(_EliminateDuplicates, self).getDetails(config)
        window = 0
        if "window" in config:
            window = config["window"]
        details = f'{details}[window={window}]'
        return details

class _DetectDuplicates(DeviceFunction):
    """

    """
    def apply(self, deviceAnalog, attribute, configuration, data, value):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :param value:
        :return:
        """

        now = TimeManager.currentTimeMillis()
        lastValue = None
        if "detectDuplicates.lastValue" in data:
            lastValue = data["detectDuplicates.lastValue"]
        data["detectDuplicates.lastValue"] = value


        # If value equals lastValue, then this is a duplicate value.
        # If value is the first duplicate value, then lastValue has already
        # been passed along and we want to filter out the current value and
        # all other sequential duplicates within the window.
        if value == lastValue:
            # windowEnd is the end time of the current window.
            windowEnd = 0
            if "windowEnd" in data:
                windowEnd = data["detectDuplicates.windowEnd"]
            # If the current window has expired (i.e., windowEnd <= now),
            # then update windowEnd.
            if windowEnd <= now:
                # windowEnd is the current time plus the window.
                # window is normalized so that window is greater than or equal
                # to zero.
                window = self.getWindow(configuration)
                data["detectDuplicates.windowEnd"] = \
                     now + (window if window > 0 else 0)
                # when the window moves, we need to send an alert.
                data["detectDuplicates.alertSent"] = False
            # The first time we get here, alertSent will be false (because
            # of the "else" part below) and an alert will be sent.
            # alertSent will then be true until the window expires or
            # a non-duplicate value is received.
            alertSent = False
            if "detectDuplicates.alertSent" in data:
                alertSent = data["detectDuplicates.alertSent"]
            if not alertSent:
                data["detectDuplicates.alertSent"] = True
                alertMessage = \
                    DeviceFunction.createAlert(deviceAnalog, configuration)
                deviceAnalog.queueMessage(alertMessage)
        else:
            # Values are not duplicates. Move window.
            # windowEnd is the current time plus the window.
            # window is normalized so that window is greater than or equal to
            # zero.
            window = self.getWindow(configuration)
            data["detectDuplicates.windowEnd"] = \
                    now + (window if window > 0 else 0)
            data["detectDuplicates.alertSent"] = False

        return True

    def get(self, deviceAnalog, attribute, configuration, data):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :return:
        """
        value = None
        if "detectDuplicates.lastValue" in data:
            value = data.get("detectDuplicates.lastValue")
        return value

    def getDetails(self, config, data=None):
        """

        :param config:
        :param data:
        :return:
        """
        details = super(_DetectDuplicates, self).getDetails(config)
        window = 0
        if "window" in config:
            window = config["window"]
        alertFormatURN = None
        if "alertFormatURN" in config:
            alertFormatURN = config["alertFormatURN"]
        if alertFormatURN is None:
            alertFormatURN = ""
        details = \
            f'{details}[window={window}, alertFormatURN="{alertFormatURN}"]'
        return details

class _PrivacyPolicy(DeviceFunction):
    """

    """
    def apply(self, deviceAnalog, attribute, configuration, data, value):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :param value:
        :return:
        """
        data["privacyPolicy.value"] = value
        return True

    def get(self, deviceAnalog, attribute, configuration, data):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :return:
        """
        value = None
        if "privacyPolicy.value" in data:
            value = data["privacyPolicy.value"]
        if value is None:
            return None

        string = value.__str__()
        content = None
        try:
            content = string.encode(encoding='utf-8')
        except Exception as e:
            raise StateException(f'DeviceFunction._PrivacyPolicy.get(): {e}')
        level = None
        if "level" in configuration:
            level = configuration["level"]
        if "one-way" == level:
            try:
                md = hashlib.sha256()
                md.update(content)
                md = md.digest()
                urlEncodedDigest = str(base64.urlsafe_b64encode(md),'utf-8')
                return urlEncodedDigest
            except Exception as e:
                DeviceFunction.getLogger().warning(
                    f'_PrivatePolicy.get(): one-way - '
                    f'Could not create digest, caught exception: {e}')
                return value
        elif "two-way" == level:
            hashingKey = None
            if "hashingKey" in configuration:
                hashingKey = configuration["hashingKey"]
            if hashingKey is None:
                DeviceFunction.getLogger().warning(
                    f'_PrivacyPolicy.get(): two-way - no hashingKey given')
                return value

            key = None
            try:
                key = hashingKey.encode(encoding='utf-8')
            except Exception as e:
                raise StateException(
                    f'DeviceFunction._PrivacyPolicy.get(): two-way - {e}')
            try:
                mac = hmac.new(key, content, digestmod=hashlib.sha256)
                digest = mac.digest()
                urlEncodedDigest = str(base64.b64encode(digest), 'utf-8')
                return urlEncodedDigest
            except Exception as e:
                DeviceFunction.getLogger().warning(
                    f'_PrivacyPolicy().get(): two-way - {e}')
                return value
        elif "random" == level:
            DeviceFunction.getLogger().warning(
                "_PrivacyPolicy.get(): random privacy not supported")
        else:
            # level is "none" or not given (default)
            return value

        return None


    def getDetails(self, config, data=None):
        """

        :param config:
        :param data:
        :return:
        """
        details = super(_PrivacyPolicy, self).getDetails(config)
        level = None
        if "level" in config:
            level = config["level"]
        if level is None:
            level = ""
        details = f'{details}[level="{level}"]'
        return details

class _AlertCondition(DeviceFunction):
    """

    """
    def apply(self, deviceAnalog, attribute, configuration, data, value):
        """
        """
        condition = None
        if "alertCondition.condition" in data:
            condition = data["alertCondition.condition"]
        if condition is None:
            formula = None
            if "condition" in configuration:
                formula = configuration["condition"]
            if formula is not None:
                tokens = FormulaParser.tokenize(formula)
                stack = []
                FormulaParser.parseConditionalOrExpression(stack, tokens,
                                                           formula, 0)
                condition = stack.pop()
                data["alertCondition.condition"] = condition

        computedValue = float(self.compute(condition, deviceAnalog))
        if math.isnan(computedValue) or math.isinf(computedValue) or \
                computedValue == 0.0:
            data["alertCondition.value"] = value
            return True
        try:
            alertMessage = DeviceFunction.createAlert(deviceAnalog,
                                                      configuration)
            # FINE
            DeviceFunction.getLogger().debug(f'_AlertCondition.apply(): '
                f'{deviceAnalog.getEndpointId()} : Alert : '
                f'"{attribute}"="{value}" ({alertMessage.getDescription()})')
            deviceAnalog.queueMessage(alertMessage)
        except Exception as e:
            # SEVERE
            DeviceFunction.getLogger().critical(f'_AlertCondition.apply(): '
                                                f'caught exception {e}')
        filter = None
        if "filter" in configuration:
            filter = bool(configuration["filter"])
        if filter is None or filter == True:
            # if this is a filter, returning false stops the pipeline
            return False
        data["alertCondition.value"] = value
        return True

    def get(self, deviceAnalog, attribute, configuration, data):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :return:
        """
        value = data.pop("alertCondition.value", None)
        return value

    def getDetails(self, config, data=None):
        """

        :param config:
        :param data:
        :return:
        """
        details = super(_AlertCondition, self).getDetails(config)
        dofilter = None
        if "filter" in config:
            dofilter = config["filter"]
        if dofilter is None:
            dofilter = True
        condition = None
        if "condition" in config:
            condition = config["condition"]
        if condition is None:
            condition = ""
        urn = None
        if "urn" in config:
            urn = config["urn"]
        if urn is None:
            urn = ""
        fields = None
        if "fields" in config:
            fields = config["fields"]
        if fields is None:
            fields = ""
        details = f'{details}[condition="{condition}", urn="{urn}", '\
                  f'fields="{fields}", filter={dofilter}'
        return details

class _ComputedMetric(DeviceFunction):
    """

    """
    def apply(self, deviceAnalog, attribute, configuration, data, value):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :param value:
        :return:
        """
        formula = None
        if "computedMetric.formula" in data:
            formula = data["computedMetric.formula"]
        if formula is None:
            formula = str(configuration["formula"])
            if formula is not None:
                tokens = FormulaParser.tokenize(formula)
                formula = FormulaParser.parseFormula(tokens, formula)
                data["computedMetric.formula"] = formula

        computedValue = float(self.compute(formula, deviceAnalog))
        if math.isnan(computedValue) or math.isinf(computedValue):
            return False
        data["computedMetric.value"] = computedValue
        return True

    def get(self, deviceAnalog, attribute, configuration, data):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :return:
        """
        return data.pop("computedMetric.value", None)

    def getDetails(self, config, data=None):
        """"""
        details = super(_ComputedMetric, self).getDetails(config)
        formula = None
        if "formula" in config:
            formula = config["formula"]
        if formula is None:
            formula = ""
        details = f'{details}[formula="{formula}"]'
        return details

class _ActionCondition(DeviceFunction):
    """

    """
    def apply(self, deviceAnalog, attribute, configuration, data, value):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :param value:
        :return:
        """
        condition = None
        if "actionCondition.condition" in data:
            condition = data["actionCondition.condition"]
        if condition is None:
            formula = None
            if "condition" in configuration:
                formula = configuration["condition"]
            if formula is not None:
                tokens = FormulaParser.tokenize(formula)
                stack = []
                FormulaParser.parseConditionalOrExpression(stack, tokens,
                                                           formula, 0)
                condition = stack.pop()
                data["actionCondition.condition"] = condition

        computedValue = float(self.compute(condition, deviceAnalog))
        # zero is false
        if math.isnan(computedValue) or math.isinf(computedValue) or \
                computedValue == 0.0:
            data["actionCondition.value"] = value
            return True
        # getActionArgs may return null
        actionArgs = self.getActionArgs(deviceAnalog, configuration)
        actionName = None
        if "name" in configuration:
            actionName = str(configuration.get("name"))
        try:
            # TODO: temporary until actionCondition allows multiple args
            #argMap = dict()
            #argMap["value"] = actionArgs[0]
            deviceAnalog.call(actionName, actionArgs)
        except Exception as e:
            # SEVERE
            DeviceFunction.getLogger().critical(f'_ActionCondtion.apply(): '
                f'{e}')
        filter = None
        if "filter" in configuration:
            filter = bool(configuration["filter"])
        if filter is None or filter == True:
            #if this is a filter, returning false stops the pipeline
            return False
        data["actionCondition.value"] = value
        return True

    def get(self, deviceAnalog, attribute, configuration, data):
        """

        :param deviceAnalog:
        :param attribute:
        :param configuration:
        :param data:
        :return:
        """
        return data.pop("actionCondition.value", None)

    def getDetails(self, config, data=None):
        """

        :param config:
        :param data:
        :return:
        """
        details = super(_ActionCondition, self).getDetails(config)
        filter = None
        if "filter" in config:
            filter = config["filter"]
        if filter is None:
            filter = True
        condition = None
        if "condition" in config:
            condition = config["condition"]
        if condition is None:
            condition = ""
        action = None
        if "name" in config:
            action = config["name"]
        if action is None:
            action = ""
        arguments = None
        if "arguments" in config:
            arguments = config["arguments"]
        if arguments is None:
            arguments = ""
        details = f'{details}[condition="{condition}", action="{action}", '\
            f'arguments="{arguments}", filter={filter}]'
        return details

class ValueProviderImpl(ValueProvider):
    """This class encapsulates a device function value."""

    __slots__ = (
        'deviceAnalog'
    )

    def __init__(self, deviceAnalog):
        """Create a :class:`.ValueProviderImpl` instance."""
        super(ValueProviderImpl, self).__init__()
        self.deviceAnalog = deviceAnalog

    def getCurrentValue(self, key):
        """
        Return the value of the attribute `key`

        :param key: attribute name
        :return: the value of attribute `key`
        """
        if self.deviceAnalog is None:
            return None
        return self.deviceAnalog.getAttributeValue(key)

    def getInProcessValue(self, key):
        """
        Return the in processs value of attribute `key`

        :param key: attribute name
        :return: the in process value of attribute `key`
        """
        if self.deviceAnalog is None:
            return None
        return DeviceFunction.getInProcessValue(
            self.deviceAnalog.getEndpointId(),
            self.deviceAnalog.getDeviceModel().getURN(), key)

_POLICY_MAP = dict()
MEAN = _Mean("mean")
_POLICY_MAP[MEAN.getId()] = MEAN
SAMPLE = _SampleQuality("sampleQuality")
_POLICY_MAP[SAMPLE.getId()] = SAMPLE
FILTER_CONDITION = _FilterCondition("filterCondition")
_POLICY_MAP[FILTER_CONDITION.getId()] = FILTER_CONDITION
MIN = _Min("min")
_POLICY_MAP[MIN.getId()] = MIN
MAX = _Max("max")
_POLICY_MAP[MAX.getId()] = MAX
STANDARD_DEVIATION = _StandardDeviation("standardDeviation")
_POLICY_MAP[STANDARD_DEVIATION.getId()] = STANDARD_DEVIATION
BATCHBYCOST = _BatchByCost("batchByCost")
_POLICY_MAP[BATCHBYCOST.getId()] = BATCHBYCOST
BATCHBYSIZE = _BatchBySize("batchBySize")
_POLICY_MAP[BATCHBYSIZE.getId()] = BATCHBYSIZE
BATCHBYTIME = _BatchByTime("batchByTime")
_POLICY_MAP[BATCHBYTIME.getId()] = BATCHBYTIME
ELIMINATEDUPLICATES = _EliminateDuplicates("eliminateDuplicates")
_POLICY_MAP[ELIMINATEDUPLICATES.getId()] = ELIMINATEDUPLICATES
DETECTDUPLICATES = _DetectDuplicates("detectDuplicates")
_POLICY_MAP[DETECTDUPLICATES.getId()] = DETECTDUPLICATES
PRIVACY_POLICY = _PrivacyPolicy("privacyPolicy")
_POLICY_MAP[PRIVACY_POLICY.getId()] = PRIVACY_POLICY
ALERT_CONDITION = _AlertCondition("alertCondition")
_POLICY_MAP[ALERT_CONDITION.getId()] = ALERT_CONDITION
COMPUTED_METRIC = _ComputedMetric("computedMetric")
_POLICY_MAP[COMPUTED_METRIC.getId()] = COMPUTED_METRIC
ACTION_CONDITION = _ActionCondition("actionCondition")
_POLICY_MAP[ACTION_CONDITION.getId()] = ACTION_CONDITION
