# Copyright (c) 2018, 2019, Oracle and/or its affiliates.  All rights reserved.
#
# This software is dual-licensed to you under the MIT License (MIT) and
# the Universal Permissive License (UPL).  See the LICENSE file in the root
# directory for license terms.  You may choose either license, or both.

from abc import abstractmethod
import base64
import hashlib
import io
from urllib.parse import urlparse

from Cryptodome.Signature import PKCS1_v1_5
from Cryptodome.Hash import HMAC, SHA256
from Cryptodome import Random
from Cryptodome.Cipher import AES
from Cryptodome.PublicKey import RSA
from Cryptodome.Util.Padding import pad,unpad

from iotcs.common import Closeable


class TrustedAssetsManagerBase(Closeable):

    TA_STORE_PROPERTY = "oracle.iot.client.trustedAssetsStore"
    TA_STORE_PASSWORD_PROPERTY = "oracle.iot.client.trustedAssetsStorePassword"
    DEFAULT_PORT = 443

    __slots__ = (
        'clientId',
        'sharedSecret',
        'serverScheme',
        'serverHost',
        'serverPort',
        'endpointId',
        'privateKey',
        'publicKey',
        'connectedDeviceTag',
        'serverUrl',
        'icdMap'
    )

    def __init__(self):
        self.clientId = None
        self.sharedSecret = None
        self.serverScheme = None
        self.serverHost = None
        self.serverPort = -1;
        self.endpointId = None
        self.privateKey = None
        self.publicKey = None
        self.connectedDeviceTag = None
        self.serverUrl = None
        self.icdMap = dict()

    def setServer(self, uriString):
        self.serverUrl = uriString
        serverUrlData = urlparse(uriString)
        self.serverScheme = serverUrlData.scheme
        self.serverHost = serverUrlData.hostname
        if ((serverUrlData.port != None) and (serverUrlData.port > 0)):
            self.serverPort = serverUrlData.port
        else:
            self.serverPort = TrustedAssetsManagerBase.DEFAULT_PORT

    def setEndpointId(self, clientId):
        self.endpointId = clientId

    def setPrivateKey(self, privateKey):
        self.privateKey = privateKey

    def setPublicKey(self, publicKey):
        self.publicKey = publicKey

    def getServerScheme(self):
        return self.serverScheme

    def getServerHost(self):
        return self.serverHost

    def getServerPort(self):
        return self.serverPort

    def getEndpointId(self):
        return self.endpointId

    def getClientId(self):
        return self.clientId

    # TODO: Does not accept algorithm.
    def signWithPrivateKey(self, message, algorithm=SHA256):
        if algorithm is None:
            algorithm = SHA256
        rsakey = RSA.importKey(bytes(self.privateKey))
        signer = PKCS1_v1_5.new(rsakey)
        digest = algorithm.new()
        digest.update(bytes(message))
        #print(digest.hexdigest())
        cipher_text = signer.sign(digest)
        return cipher_text

    def signWithKey(self, key, message):
        rsakey = RSA.importKey(key)
        signer = PKCS1_v1_5.new(rsakey)
        digest = SHA256.new()
        digest.update(message)
        cipher_text = signer.sign(digest)
        return base64.b64encode(cipher_text)

    def signWithSharedSecret(self, data, algorithm=SHA256, hardwareId=None):
        if algorithm is None:
            algorithm = SHA256

        if data is None:
            raise ValueError("data cannot be None")

        secretKey = None
        if hardwareId is None or hardwareId == self.getClientId():
            secretKey = self.sharedSecret
        elif self.icdMap is not None and len(self.icdMap) > 0:
            secretKey = self.icdMap[hardwareId]

        if secretKey is None:
            raise ValueError("Shared secret not provisioned")

        key=bytes(secretKey,'utf-8')
        digest = HMAC.new(key,digestmod=algorithm)
        digest.update(bytes(data))
        cipher_text = digest.digest()

        return cipher_text

    # def signWithSharedSecretHarwareId(self, message, hardwareId):
    #     key=bytes(self.sharedSecret,'utf-8')
    #     digest = HMAC.new(key,digestmod=SHA256)
    #     digest.update(bytes(hardwareId,"utf-8"))
    #     digest.update(bytes(message))
    #     cipher_text = digest.digest()
    #     return cipher_text

    # TODO: JCL does not return the pair but sets internal members
    # TODO: original code returns pair, and did not set internal members
    def generateKeyPair(self, keytype, keyalgo):
        keypair = self.genRSAKeyPair()
        self.publicKey = keypair[1]
        self.privateKey = keypair[0]
        return  keypair

    def genRSAKeyPair(self,bits=2048):
        new_key = RSA.generate(bits, e=65537)
        public_key = new_key.publickey().exportKey('DER')
        private_key = new_key.exportKey('DER')
        return private_key, public_key

    def getPublicKey(self):
        return self.publicKey

    def getPrivateKey(self):
        return self.privateKey

    @abstractmethod
    def close(self):
        pass


class UnifiedTrustedAssetsManager(TrustedAssetsManagerBase):
    """The TrustedAssetsManager handles trust material used for
    activation and authentication to the IoT CS
    """

    #
    # To have a simple 1 byte version and be readable outside of the blob,
    # the first version number will be first printable ASCII character 33 (!).
    #
    # Since the last printable char is 126 (~), we can have 94 versions.
    #
    FORMAT_VERSION = 33
    MAX_FORMAT_VERSION = 126
    #
    # The URI of the server, e.g., https://iotinst-mydomain.iot.us.oraclecloud.com:443
    #
    SERVER_URI_TAG = 1
    # A client id is either an integration id (for enterprise clients), or an
    # activation id (for device clients). An activation id may also be
    # referred to a hardware id.
    #
    CLIENT_ID_TAG = 2
    #
    # The shared secret as plain text
    #
    SHARED_SECRET_TAG = 3
    #
    # For devices, the endpoint id TLV is omitted from the provisioning file
    # (unless part of a CONNECTED_DEVICE_TAG TLV).
    # For enterpise integrations, the endpoint id is set in the provisioning file
    # by the inclusion of the second ID argument.
    #
    ENDPOINT_ID_TAG = 4
    #
    # The trust anchor is the X509 cert
    #
    TRUST_ANCHOR_TAG = 5
    # Not used#
    PRIVATE_KEY_TAG = 6
    # Not used#
    PUBLIC_KEY_TAG = 7
    #
    # The client id and shared secret of a device that can connect
    # indirectly through the device client
    #
    # Connected device TLV =
    # [CONNECTED_DEVICE_TAG|<length>|[CLIENT_ID_TAG|<icd activation id length>|<icd activation id>][SHARED_SECRET_TAG|<icd shared secrect length>|<icd shared secret>]]
    #
    CONNECTED_DEVICE_TAG = 8
    AES_BLOCK_SIZE = 16
    AES_KEY_SIZE = 128
    PBKDF2_ITERATIONS = 10000


    class TLV:

        __slots__ = (
            'tag',
            'length',
            'value',
            'offsetToNext'
        )

        def __init__(self, buff, offset):
            self.tag = None
            self.length = 0
            self.value = bytes()
            self.offsetToNext = 0
            buffer = bytearray(buff)
            self.tag = buffer[offset] & 0x000000FF
            offset += 1

            self.length = buffer[offset] & 0x000000FF
            self.length = self.length << 8
            offset += 1
            self.length += buffer[offset] & 0x000000FF
            offset += 1

            self.value = buffer[offset:offset + self.length]
            self.offsetToNext = offset + self.length

        @staticmethod
        def writeTLV(bio, tag, value):
            if (isinstance(value,str)):
                valuebuffer = bytearray(value,'utf-8')
            else:
                valuebuffer = value
            bio.write(bytes([tag]));
            valuelen =  len(valuebuffer)
            bio.write(valuelen.to_bytes(2,byteorder="big"))
            bio.write(valuebuffer)

    __slots__ = (
        'path',
        'passphrase',
        'assetsFile',
        'icdMap',
        # TODO: this is 'trustAnchors" in JCL and defined in TrustedAssetsManagerBase
        'trustAnchor',
        'bs'
    )

    def __init__(self, path, passphrase):
        """Construct from encrypted trusted assets.
        path: the path to the trusted assets file
        passphrase: the passphrase for decrypting the file
        """
        if path is None:
            raise ValueError("'path' may not be None")
        if passphrase is None:
            raise ValueError("'passphrase' may not be None")

        self.icdMap = dict()
        self.path = path
        self.passphrase = passphrase
        self.bs = UnifiedTrustedAssetsManager.AES_BLOCK_SIZE
        super(UnifiedTrustedAssetsManager, self).__init__()
        #super().__init__()
        self.load(path, passphrase)
        self.assetsFile = None
        # TODO: What's the difference between trustAnchor and certificate
        # TODO: JCL has DefaultTrustedAssetsManager (probably obsolete ?) has certificate
        self.trustAnchor = None

    def load(self, path, password):
        with open(path, "rb") as self.assetsFile:
            try:
                encrypted_data = bytearray(self.assetsFile.read())
                self.initialize(password, encrypted_data)

            except Exception as e:
                print(e.args)
                raise e
            finally:
                self.assetsFile.close()

    def initialize(self, passphrase, encrypted_data):
        if encrypted_data[0] != UnifiedTrustedAssetsManager.FORMAT_VERSION:
            raise RuntimeError("bad format version")
        try:
            # remove the comments at the end of the data
            base64_len = encrypted_data.find(b"#")
            # base64_len = encrypted_data.index(ord('#'), 1)
            del encrypted_data[base64_len:]

        except ValueError:
            # just means that '#' isn't in the encrypted data
            pass
        # skip the format version, decode the remainder
        cipher_text = base64.b64decode(encrypted_data[1:])
        salt = cipher_text[:UnifiedTrustedAssetsManager.AES_BLOCK_SIZE]
        passbytes = str(passphrase).encode()  # default is UTF-8
        key = self.create_key(passbytes, salt)

        plaintext = self.decrypt(key, cipher_text)
        plainlength = len(plaintext)
        offset = 0
        while offset < plainlength:
            tlv = self.TLV(plaintext, offset)
            if tlv.tag == UnifiedTrustedAssetsManager.SERVER_URI_TAG:
                serverUrl = tlv.value.decode('utf-8')
                self.setServer(serverUrl)
            elif tlv.tag == UnifiedTrustedAssetsManager.CLIENT_ID_TAG:
                self.clientId = tlv.value.decode('utf-8')
            elif tlv.tag == UnifiedTrustedAssetsManager.SHARED_SECRET_TAG:
                self.sharedSecret = tlv.value.decode('utf-8')
            elif tlv.tag == UnifiedTrustedAssetsManager.ENDPOINT_ID_TAG:
                self.endpointId = tlv.value.decode('utf-8')
                self.setEndpointId(self.endpointId)
            elif tlv.tag == UnifiedTrustedAssetsManager.TRUST_ANCHOR_TAG:
                self.trustAnchor = RSA.importKey(bytes(tlv.value))
            elif tlv.tag == UnifiedTrustedAssetsManager.PRIVATE_KEY_TAG:
                self.setPrivateKey(tlv.value)
                length = len(self.privateKey)
            elif tlv.tag == UnifiedTrustedAssetsManager.PUBLIC_KEY_TAG:
                self.setPublicKey(tlv.value)
            elif tlv.tag == UnifiedTrustedAssetsManager.CONNECTED_DEVICE_TAG:
                #print("connected device tag")
                self.icdMap[tlv.value] = None
            else:
                raise RuntimeError(("Unknown value tag in provisioner file" + str(tlv.tag)));
            offset = tlv.offsetToNext

    def create_key(self, password, salt):
        return hashlib.pbkdf2_hmac('sha1', password, salt,
            UnifiedTrustedAssetsManager.PBKDF2_ITERATIONS, 16)

    def isActivated(self):
        return (self.endpointId is not None)

    def decrypt(self, secret_key, cipher_text):
        iv = cipher_text[:16]
        enc = cipher_text[16:]
        cipher = AES.new(secret_key, AES.MODE_CBC, iv)
        decdata = unpad(cipher.decrypt(enc), 16)
        return decdata

    def _pad(self, s):
        return s + (self.bs - len(s) % self.bs) * chr(self.bs - len(s) % self.bs)

    def encrypt(self, secret_key, plain_text):
        plain_text = pad(plain_text,16)
        iv = Random.new().read(AES.block_size)
        passbytes = str(self.passphrase).encode()  # default is UTF-8
        secret_key = self.create_key(passbytes,iv)
        cipher = AES.new(secret_key, AES.MODE_CBC, iv)
        enc_buffer = iv + cipher.encrypt(bytes(plain_text))
        return base64.b64encode(enc_buffer)


    # def setEndpointCredentials(self, endpointId, privateKey, publicKey, certificate):
    #     buffer = self.createTas(self.passphrase, self.serverScheme,
    #                             self.serverHost, self.serverPort,
    #                             self.clientId, self.sharedSecret,
    #                             endpointId, certificate,
    #                             privateKey, publicKey,
    #                             self.icdMap)
    #     #filename = "/home/rajesh/work/new/{}".format(self.clientId)
    #     f = open(self.path, 'wb')
    #     f.write(buffer)
    #     f.close()
    #     self.endpointId = endpointId
    #     self.privateKey = privateKey
    #     self.publicKey = publicKey

    def setEndpointCredentials(self, endpointId, certificate):

        if self.privateKey is None:
            raise  ValueError("Private key not yet generated.")

        self.endpointId = endpointId

        # TODO: What is the anlogous functionality in python ?
        # TODO: certifcate appears to not be used in UnifiedTrustedAssetsManager
        # TODO: but trustAnchors exists, and the trustAnchor passed to createTas
        # TODO: is the "first" trustAnchor
        #self.trustAnchor = certificate

        try:
            self.store()
        except Exception as e:
            ValueError(f'Error storing the trusted assets...{e}')

    def store(self):

        buffer = self.createTas(self.passphrase, self.serverScheme,
                                    self.serverHost, self.serverPort,
                                    self.clientId, self.sharedSecret,
                                    self.endpointId, self.trustAnchor,
                                    self.privateKey, self.publicKey,
                                    self.icdMap)

        f = open(self.path, 'wb')
        f.write(buffer)
        f.close()

    def createTas(self,password, serverScheme, serverHost, serverPort,
                        clientId, sharedSecret, endpointId, trustAnchor,
                        privateKey, publicKey, icdMap):
        bio = io.BytesIO()
        urlbuffer = bytearray(self.serverUrl,'utf-8')
        self.TLV.writeTLV(bio, UnifiedTrustedAssetsManager.SERVER_URI_TAG, urlbuffer)
        self.TLV.writeTLV(bio, UnifiedTrustedAssetsManager.CLIENT_ID_TAG, clientId)
        self.TLV.writeTLV(bio, UnifiedTrustedAssetsManager.SHARED_SECRET_TAG, sharedSecret)
        if endpointId:
            self.TLV.writeTLV(bio, UnifiedTrustedAssetsManager.ENDPOINT_ID_TAG, endpointId)
        if trustAnchor:
            self.TLV.writeTLV(bio, UnifiedTrustedAssetsManager.TRUST_ANCHOR_TAG, trustAnchor)
        if privateKey:
            self.TLV.writeTLV(bio, UnifiedTrustedAssetsManager.PRIVATE_KEY_TAG, privateKey)
        if publicKey:
            self.TLV.writeTLV(bio, UnifiedTrustedAssetsManager.PUBLIC_KEY_TAG, publicKey)
        if icdMap:
            self.TLV.writeTLV(bio, UnifiedTrustedAssetsManager.CONNECTED_DEVICE_TAG, "")
        buffer=bio.getvalue()
        size = len(buffer)

        ciphertext = self.encrypt(self.passphrase,buffer)

        fileBuf = io.BytesIO()
        fileBuf.write(bytes([UnifiedTrustedAssetsManager.FORMAT_VERSION]))
        fileBuf.write(ciphertext)
        fileBuf.write(bytes("\n#serverUri","utf-8"))
        fileBuf.write(urlbuffer)
        fileBuf.write(bytes("\n#clientId","utf-8"))
        fileBuf.write(bytes(self.clientId,"utf-8"))
        fileBuf.write(bytes("\n","utf-8"))
        return fileBuf.getvalue()

    # TODO: JCL does nothing for this, what about here ?
    def close(self):
        """CLose open resources"""
        return
